#include <limits.h>

#include <algorithm>
#include <iterator>
#include <iostream>

#include "index.h"
#include "format.h"
#include "../afuzzy/edist.h"

unsigned int CAhcTree::cnt(int addVal) const
{
    if (!mFirst) {
        return 1;
    }
    return mFirst->cnt(addVal) + addVal + mSecond->cnt(addVal);
}

unsigned int CAhcTree::depth(int addVal) const
{
    if (!mFirst) return mElems.size();
    return addVal + max(mFirst->depth(addVal), mSecond->depth(addVal));
}

void CAhcTree::search(const string & pat, int k, VS & res, int & cnt, bool first) const
{
    if (first) {
        res.clear();
        cnt = 0;
    }
    if (!mFirst) {
        for (unsigned int i = 0; i < mElems.size(); ++i, ++cnt) 
        if (EditDistance(pat, mDict[mElems[i]]) <= k) {
            res.push_back(mDict[mElems[i]]);
        }
        return;
    }
    ++cnt;
    const std::string & s = mDict[mElems[mCenterIndex]];
    int d = EditDistance(pat, s);
#if 0
    std::cout << "Distance to center " <<  s << " = " << d<< " radius = " << mElemRad[mCenterIndex] << endl;
#endif
    if (d > k + mElemRad[mCenterIndex])
        return;
    mFirst->search(pat, k, res, cnt, false);
    mSecond->search(pat, k, res, cnt, false);
    
}

void CAhcTree::fill(VS & str, int row, int col) const
{
    if (!mFirst) {
        for (unsigned int i = 0; i < mElems.size(); ++i) {
            str[row + i] += " ";
            str[row + i] += mDict[mElems[i]];
        }
        return;
    }
    str[row][col] = '*';
    str[row][col + 1] = '-';
    mFirst->fill(str, row, col + 2);
    int maxRow = mSecond->cnt(0) + row;
    for(++row;row < maxRow; ++row) {
        str[row][col] = '*';
    }
    str[row][col] = '*';
    str[row][col + 1] = '-';
    mSecond->fill(str, row, col + 2);
}

CAhcTree * joinClusters(const CAhcTree & lhs, const CAhcTree & rhs)
{
    CAhcTree * newTree= new CAhcTree(lhs.mDict);

    const VS & dict = lhs.mDict;
    if (&rhs.mDict != &dict) {
        std::cerr << "getJointRad - different dictionary pointers" << std::endl;
        throw "getJointRad - different dictionary pointers";
    }

    int M = lhs.mElems.size(), N = rhs.mElems.size();
    newTree->mElems.reserve(M+N);
    newTree->mElems = lhs.mElems;
    copy(rhs.mElems.begin(), rhs.mElems.end(), back_inserter(newTree->mElems));

    newTree->mElemRad = lhs.mElemRad;
    newTree->mElemRad.reserve(M + N);

    copy(rhs.mElemRad.begin(), rhs.mElemRad.end(), back_inserter(newTree->mElemRad));

    for(int i = 0; i < lhs.mElemRad.size(); ++i) {
        for(int k = 0; k < rhs.mElemRad.size(); ++k) {
           int d = EditDistance(dict[lhs.mElems[i]], dict[rhs.mElems[k]]);
           newTree->mElemRad[M + k] = max(newTree->mElemRad[M + k], d);
           newTree->mElemRad[i] = max(newTree->mElemRad[i], d);
        }
    }
    
    newTree->mCenterIndex = std::max_element(newTree->mElemRad.begin(), newTree->mElemRad.end()) - newTree->mElemRad.begin();
    

    newTree->mFirst = &lhs;
    newTree->mSecond = &rhs;
    return newTree;
}

int getJointRad(const CAhcTree & lhs, const CAhcTree &rhs) 
{
    const VS & dict = lhs.mDict;
    if (&rhs.mDict != &dict) {
        std::cerr << "getJointRad - different dictionary pointers" << std::endl;
        throw "getJointRad - different dictionary pointers";
    }

    int M = lhs.mElems.size(), N = rhs.mElems.size();
    VI   tmp = lhs.mElemRad;
    tmp.reserve(M + N);

    copy(rhs.mElemRad.begin(), rhs.mElemRad.end(), back_inserter(tmp));

    for(int i = 0; i < lhs.mElemRad.size(); ++i) {
        for(int k = 0; k < rhs.mElemRad.size(); ++k) {
           int d = EditDistance(dict[lhs.mElems[i]], dict[rhs.mElems[k]]);
           tmp[M + k] = max(tmp[M + k], d);
           tmp[i] = max(tmp[i], d);
        }
    }
    return *max_element(tmp.begin(), tmp.end());
}

void generateTrees(AhcForest & res, const VS & dict, unsigned int maxTreeCnt)
{
    int   N = dict.size();
    VI    bestNghb(N, -1);
    VI    bestNghbDist(N, INT_MAX);
    res.resize(N);
    for (unsigned int i = 0; i < dict.size(); ++i) {
        res[i] = new CAhcTree(dict, i);
    }
    while (res.size() > std::max(maxTreeCnt, 1U)) {
#if 1
        if (!(res.size() % 10)) { 
            std::cout << "res size" << res.size() << std::endl;
        }
#endif
        // first of all we determine the best neighbour for all clusters for which it s undefined
        for (unsigned int i = 0; i < res.size(); ++i)
        if (bestNghbDist[i] == INT_MAX) {
            for (unsigned int k = 0; k < res.size(); ++k) 
            if (i != k) {
                 int d = getJointRad(*res[i], *res[k]);
                 if (bestNghbDist[i] > d) {
                      bestNghbDist[i] = d;
                      bestNghb[i] = k;
                 }
            }
        }
        int bi, bv = INT_MAX;
        for (unsigned int i = 0; i < res.size(); ++i) {
            if (bv > bestNghbDist[i]) {
                  bv = bestNghbDist[i];
                  bi = i;
            }
        }
        int ai = bestNghb[bi];
        // Clusters to be joined bi and ai!!!!
        CAhcTree * t = joinClusters(*res[ai], *res[bi]);
        for (unsigned int i = 0; i < res.size(); ++i) {
            if (i == ai || i == bi || bestNghb[i] == ai || bestNghb[i] == bi) {
                bestNghbDist[i] = INT_MAX;
            } else if (bestNghb[i] > bi) {
                bestNghb[i]--;
            }
        }
        res[ai] = t;
        res.erase(res.begin() + bi);
        bestNghb.erase(bestNghb.begin() + bi);
        bestNghbDist.erase(bestNghbDist.begin() + bi);
    }
}

void print(CAhcTree & tree)
{
    int  N = tree.cnt(1), M = tree.depth(3);
    VS   str(N + 1, std::string(M, ' '));

    tree.fill(str, 0, 0);
    for (unsigned int i = 0; i < str.size(); ++i)
        std::cout << str[i] << std::endl;
    std::cout << "=======================================\n";
}

