#include "index.h"
#include "myagrep.h"
#include "hash.h"
#include "genfullng.h"

long long seekVal;

CSpellIndex::CSpellIndex (bool hasTransp, EAlphabet AlphaType)
              : CInterface(hasTransp, AlphaType), opened (false)
{
    const char * _ABC = GetAlphabet(AlphaType);

    int size = strlen (_ABC);
    ABC = new char[size + 1];
    memcpy (ABC, _ABC, size + 1);
}


CSpellIndex::~CSpellIndex ()
{
    delete[]ABC;
}

bool CSpellIndex::Insert (const char *str)
{
    Dict.insert (string (str));
    return true;
}

struct CSearchReceiver {
    CSearchReceiver(const CSpellIndex::StringSet& Dict, StringList* pResList) : mDict(Dict), mpResList(pResList), mCmpQty(0) {}

    const CSpellIndex::StringSet&   mDict;
    StringList*                     mpResList;
    unsigned                        mCmpQty;

    void operator()(const string& str) {
        ++mCmpQty;
        if (mDict.find(str) != mDict.end()) {
            if (mpResList) { mpResList->push_back(str); }
        }
    }
};


static void ExpandAndSearch(int start,
                        int &numComp, const CSpellIndex::StringSet & dict,
                        string & str, 
                        bool hasTransp, int kerr, 
                        StringList *resList,
                        const char *ABC)
{
    if (1) {
        ++numComp;
#if 0
        if (GetEnvSwitch("SHOW_VARIANTS")) {
            std::cout << "Variant: " << str << std::endl;
        }
#endif
        if (dict.find(str) != dict.end()) {
            if (resList) { resList->push_back (str); }
        }
    }

    int size = str.size ();
    if (!kerr || start > size) return;

    int i = 0;
    // Substitution
    for (i = start; i < size; i++) {
        char origChar = str[i];
        for (const char *ptr = ABC; *ptr; ptr++) {
            if (*ptr != str[i]) {
                str[i] = *ptr;
                ExpandAndSearch(i + 1, numComp, dict, str, hasTransp, kerr - 1, resList, ABC);
            }
        }
        str[i] = origChar;
    }
    // We well try to make as few insertions and erases as possible
    // It is enough to make 1 insertion, 1 deletion per insertion case
    // and 1 insertion, 1 deletion per deletion case
    // Insertion
    str.insert(str.begin() + start, ' ');
    for (i = start; i <= size; i++) {
        for (const char *ptr = ABC; *ptr; ptr++) {
            str[i] = *ptr; 
            ExpandAndSearch(i + 1, numComp, dict, str, hasTransp, kerr - 1, resList, ABC);
        }
        if (i < size) {
            swap(str[i], str[i + 1]);
        }
    }
    str.erase(str.end() - 1);
    if (hasTransp) {
    // Transposition
        for (i = start; i < size - 1; i++) {
            swap(str[i], str[i+1]);
            ExpandAndSearch(i + 2, numComp, dict, str, hasTransp, kerr - 1, resList, ABC);
            swap(str[i], str[i+1]);
        }
    }
    // Deletion
    if (start >= size) return;
    char origChar = str[start];
    str.erase(str.begin() + start);
    for (i = start; i < size; i++) {
        ExpandAndSearch(i, numComp, dict, str, hasTransp, kerr - 1, resList, ABC);
        if (i < size - 1) {
           swap(str[i], origChar);
        }
    }
    str.insert (str.end(), origChar);

}


bool CSpellIndex::Search (const char *str, int kerr, int *numComp,
                          StringList * resList, bool silent)
{

    if (!kerr) {
        if (Dict.find (str) != Dict.end ()) {
            if (resList != NULL) {
                resList->push_back (str);
            }
            if (!silent) {
                cout << str << endl;
            }
        }
        return true;
    }

#if 0
    StringList       lst;
    FillExpansionList (string (str), mHasTransp, kerr, &lst, ABC, 0, false, NULL);

    for (StringList::iterator i = lst.begin (); i != lst.end (); i++) {
        if (numComp)
            (*numComp)++;
        if (Dict.find (*i) != Dict.end ()) {
            if (resList != NULL)
                resList->push_back (string (*i));
            if (!silent) {
                cout << *i << endl;
            }
        }
    }
#else
    string       pat(str);
    int          nc = 0;
    //if (mAlphaType != eDNA) {
    #if 1
        // The naive algorithm is best for all data sets.
        //cout << "Using ExpandAndSearch" << endl;
        ExpandAndSearch (0, nc, Dict, pat, mHasTransp, kerr, resList, ABC);
    #else
        //cout << "Using NFA" << endl;
        CSearchReceiver recv(Dict, resList);

        GenFullNG(recv, pat, ABC, kerr);

        nc = recv.mCmpQty;
    #endif
    if (numComp) *numComp = nc;
#endif
    return true;
}

bool CSpellIndex::CloseDB ()
{
    if (!opened)
        return true;
    unmapData();
    if (close(fdDict)) {
        Perror("Cannot close dict file");
        return false;
    }
    if (close(fdHash)) {
        Perror("Cannot close hash file");
        return false;
    }
    if (close(fdParams)) {
        Perror("Cannot close param file");
        return false;
    }
    opened = false;
    return true;
}

bool CSpellIndex::OpenDB (const char *base_name, bool create)
{
    if (opened)
        CloseDB ();

    string    headerName = string(base_name) + ".hdr";
    string    hashName = string(base_name) + ".hash";
    string    dictName = string(base_name) + ".dict";

    int flags = O_RDWR;
    int access = 0;

    pDict = NULL;
    pHash = NULL;
    pParams = NULL;

    if (create)  {
        flags = O_RDWR | O_TRUNC | O_CREAT;
        access = defaultAccess;
    }
    fdParams = open(headerName.c_str(), flags, access);
    if (fdParams < 0) {
        Perror("Cannot open header file");
        return false;
    }
    fdHash = open(hashName.c_str(), flags, access);
    if (fdHash < 0) {
        Perror("Cannot open hash file");
        close(fdParams);
        return false;
    }
    fdDict = open(dictName.c_str(), flags, access);
    if (fdDict < 0) {
        Perror("Cannot dict file");
        close(fdHash);
        close(fdParams);
        return false;
    }

    if (!create) {
        opened = true;
        if (!mapData()) {
            CloseDB();
            return false;
        }
    }
    
    return opened = true;
}

bool CSpellIndex::SearchInDB (const char *str, int kerr, int *numComp,
                              StringList * resList, bool silent)
{
    if (!opened) {
        cerr << "Database is not opened" << endl;
        return false;
    }
    /* Searching the database */
    StringList lst;
    
    FillExpansionList (string (str), mHasTransp, kerr, &lst, ABC, 0, false, 0, false);

    for (StringList::iterator i = lst.begin (); i != lst.end (); i++) {
        if (numComp)
            (*numComp)++;
        
        if (searchHashDB(*i)) {
            if (!silent)
                cout << *i << endl;
            if (resList != NULL)
                resList->push_back (string (*i));
        }
    }

    return true;
}

#ifndef USE_CHAIN_HASH

bool CSpellIndex::DumpDB (const char *base_name)
{
    if (opened)
        CloseDB ();
    if (!OpenDB (base_name, true))
        return false;           // Create database

    size_t numWords = Dict.size();
    vector<pair<size_t, const string *> >   pos2word;
    
    size_t dictSize = 0;
    // The first set iteration - write 0 and calc size
    for (StringSet::iterator it = Dict.begin (); it != Dict.end (); it++) {
        const string & str = *it;
        size_t         strSize = str.size() + 1;
        char           zeroBuf[strSize + 1];
        memset(zeroBuf, 0, strSize + 1);
        if (write(fdDict, zeroBuf, strSize) != strSize) {
            Perror("Cannot write to dict file");
            CloseDB();
            return false;
        }
        dictSize += strSize;
        pos2word.push_back(make_pair(0U, &str));
    }

    int fillRatio = defaultHashRatio;

    GetEnv("FILL_RATIO", fillRatio);
    if (!fillRatio) fillRatio = defaultHashRatio;

    if (fillRatio < minHashRatio) fillRatio = minHashRatio;
    if (fillRatio > maxHashRatio) fillRatio = maxHashRatio;

    size_t hashSize = from__stl_next_prime((size_t) 100LL * numWords / fillRatio); 
    if (!hashSize) hashSize = 1;

    size_t tmp(0);
    if (lseek(fdHash, getPageAlignedValue((hashSize - 1) * sizeof(size_t)), SEEK_SET) < 0 ||
         write(fdHash, &tmp, sizeof(size_t)) != sizeof(size_t)) {
        Perror("Cannot lseek or write");
        CloseDB();
        return false;
    }
    CHashParams prm(fillRatio, hashSize, dictSize);

    if (write(fdParams, &prm, sizeof prm) != sizeof prm) {
        Perror("Cannot write params");
        CloseDB();
        return false;
    }

    if (!mapData()) {
        CloseDB();
        return false;
    }

    memset(pHash, 0, pParams->hashSize * sizeof(size_t));

    seekVal = 0;

    size_t n = 0;
    int    cntIt = 0;
    // Find out which hash table elements to fill
    for (StringSet::iterator it = Dict.begin (); it != Dict.end (); ++it, ++cntIt) {
        const string & str = *it;
        size_t         strSize = str.size() + 1;
        size_t         buckNum;
        if (!insertHashDB(str, buckNum, *pParams, pHash, &seekVal)) {
            cerr << "Failed to insert string internally" << endl;
            CloseDB();
            return false;
        }
#if 0
        if (cntIt % 10000 == 0) {
            cout << 100 * cntIt / numWords << "% seekVal " << double(seekVal) / max(1, cntIt) << endl;
        }
#endif
        pos2word[n++].first = buckNum;
    }
    // Finally store strings in the order they are in hash table
    // and fill hash positions
    sort(pos2word.begin(), pos2word.end()); // will sort by position

    size_t pos = 0;
    for (n = 0; n < pos2word.size(); ++n) {
        const string & str = *pos2word[n].second;
        size_t         strSize = str.size() + 1;
        memcpy(&pDict[pos], str.data(), strSize);
        pHash[pos2word[n].first] = pos + 1; // do not store zero value
        pos += strSize;
    }

    return CloseDB ();
}

bool CSpellIndex::searchHashDB(const string & str) 
{
    // checks are omitted here for speed
    size_t origPos = hash_func(str) % pParams->hashSize;
    size_t pos = origPos;
    size_t cnt = 0;
    do {
        ++cnt;
        if (!pHash[pos]) { // free cell
#if 0
            cout << str << " cnt in hash (0)" << cnt << endl;
#endif
            return false;
        }
        if (str == &pDict[pHash[pos] - 1]) { // zero addresses are not stored
#if 0
            cout << str << " cnt in hash (fnd) " << cnt << endl;
#endif
            return true;
        }
        pos = (pos + SKIP_CONSTANT) % pParams->hashSize; // hashing with an open address, linear search
        // for a free pos
    } while (pos != origPos);
#if 0
    cout << str << " cnt in hash (end) " << cnt << endl;
#endif
    return false;
}

bool CSpellIndex::mapData() {
    if (!opened) {
        cerr << "DB is not open, cannot map" << endl;
        return false;
    }
    pParams = reinterpret_cast<CHashParams *>(mmap(NULL, getPageAlignedValue(sizeof *pParams), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdParams, 0));
    if (!pParams || pParams == MAP_FAILED) {
        Perror("Cannot map params");
        pParams = NULL;
        return false;
    }
    pHash = reinterpret_cast<size_t *>(mmap(NULL, getPageAlignedValue(pParams->hashSize * sizeof(size_t)), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdHash, 0));
    if (!pHash || pHash == MAP_FAILED) {
        Perror("Cannot map hash");
        pHash = NULL;
        unmapData();
        return false;
    }

    pDict = reinterpret_cast<char *>(mmap(NULL, getPageAlignedValue(pParams->dictSize), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdDict, 0));
    if (!pDict || pDict == MAP_FAILED) {
        Perror("Cannot map dict");
        pDict = NULL;
        unmapData();
        return false;
    }
    return true;
}

void CSpellIndex::unmapData() {
    if (pParams) {
        if (pHash && munmap(pHash, getPageAlignedValue(pParams->hashSize * sizeof(size_t)))) {
            Perror("Cannot unmap hash");
        }
        pHash = NULL;
        if (pDict && munmap(pDict, getPageAlignedValue(pParams->dictSize))) {
            Perror("Cannot unmap dict");
        }
        pDict = NULL;
        if (munmap(pParams, getPageAlignedValue(sizeof *pParams))) {
            Perror("Cannot unmap params");
        }
        pParams = NULL;
    }
}

#else 

bool CSpellIndex::DumpDB (const char *base_name)
{
    if (opened)
        CloseDB ();
    if (!OpenDB (base_name, true))
        return false;           // Create database

    size_t numWords = Dict.size();
    size_t tmpBucketNum = 1 + numWords / 1000 ;

    size_t bucketNum = from__stl_next_prime(tmpBucketNum);

    size_t tmp(0);

    if (lseek(fdHash, getPageAlignedValue(bucketNum * sizeof(Bucket)), SEEK_SET) < 0 ||
         write(fdHash, &tmp, sizeof(size_t)) != sizeof(size_t)) {
        Perror("Cannot lseek or write");
        CloseDB();
        return false;
    }
    // we need dict size
    size_t dictSize = 0;
    for (StringSet::iterator it = Dict.begin (); it != Dict.end (); it++) {
        dictSize += it->size() + 1;
    }
    Params prm(bucketNum, dictSize);

    if (write(fdParams, &prm, sizeof prm) != sizeof prm) {
        Perror("Cannot write params");
        CloseDB();
        return false;
    }

    // wee need to reorganize strings by buckets
    vector<vector<const string *> > tmpBuckets(bucketNum);
    
    for (StringSet::iterator it = Dict.begin (); it != Dict.end (); it++) {
        int pos = hash_func(*it) % bucketNum;
        tmpBuckets[pos].push_back(&(*it));
    }
    // dump dict
    for (unsigned int i = 0; i < bucketNum; ++i) {
        for (unsigned int k = 0; k < tmpBuckets[i].size(); ++k) {
            const string  & str = *tmpBuckets[i][k];
            size_t strSize = str.size() + 1;
            if (write(fdDict, str.c_str(), strSize) != strSize) {
                Perror("Cannot write to dict file");
                CloseDB();
                return false;
            }
        }
    }


    if (!mapData()) {
        CloseDB();
        return false;
    }

    // fill hash bucket values
    size_t pos = 0;
    for (unsigned int i = 0; i < bucketNum; ++i) {
        pHash[i] = Bucket(pos, tmpBuckets[i].size());
        for (unsigned int k = 0; k < tmpBuckets[i].size(); ++k) {
            pos += tmpBuckets[i][k]->size() + 1;
        }
    }

    return CloseDB ();
}

bool CSpellIndex::mapData() {
    if (!opened) {
        cerr << "DB is not open, cannot map" << endl;
        return false;
    }
    pParams = reinterpret_cast<Params *>(mmap(NULL, getPageAlignedValue(sizeof *pParams), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdParams, 0));
    if (!pParams || pParams == MAP_FAILED) {
        Perror("Cannot map params");
        pParams = NULL;
        return false;
    }
    pHash = reinterpret_cast<Bucket *>(mmap(NULL, getPageAlignedValue(pParams->bucketNum * sizeof(Bucket)), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdHash, 0));
    if (!pHash || pHash == MAP_FAILED) {
        Perror("Cannot map hash");
        pHash = NULL;
        unmapData();
        return false;
    }

    pDict = reinterpret_cast<char *>(mmap(NULL, getPageAlignedValue(pParams->dictSize), 
                                        PROT_WRITE | PROT_READ, MAP_SHARED, fdDict, 0));
    if (!pDict || pDict == MAP_FAILED) {
        Perror("Cannot map dict");
        pDict = NULL;
        unmapData();
        return false;
    }

    return true;
}

void CSpellIndex::unmapData() {
    if (pParams) {
        if (pHash && munmap(pHash, getPageAlignedValue(pParams->bucketNum * sizeof(Bucket)))) {
            Perror("Cannot unmap hash");
        }
        pHash = NULL;
        if (pDict && munmap(pDict, getPageAlignedValue(pParams->dictSize))) {
            Perror("Cannot unmap dict");
        }
        pDict = NULL;
        if (munmap(pParams, getPageAlignedValue(sizeof *pParams))) {
            Perror("Cannot unmap params");
        }
        pParams = NULL;
    }
}

bool CSpellIndex::searchHashDB(const string & str) 
{
    // checks are ommitted here for speed
    size_t bucket = hash_func(str) % pParams->bucketNum;

    char * start = &pDict[pHash[bucket].offset];

#if 0
    cout << str << " num in hash " << pHash[bucket].num << endl;
#endif

    for (unsigned int i = 0; i < pHash[bucket].num; ++i, start += strlen(start) + 1) {
        if (str == start) {
            return true;
        }
    }
        
    return false;
}
#endif
