#include <fstream>
#include <limits>
#include <iomanip>

#include "index.h"
#include "../Common/hash.h"
#include "../Common/perf_hash.h"

unsigned    aNumCheck[MAX_KERR + 1];
unsigned    aNumSearchInBucket[MAX_KERR + 1];

#define SKIP_LOG   5


#define MAX_ALPHA_SIZE_LOG          2
#define MAX_STRING_SIZE_LOG         (8 - MAX_ALPHA_SIZE_LOG)
#define MAX_SUPPORTED_ALPHA_SIZE    (1<<MAX_ALPHA_SIZE_LOG) 
#define MAX_STRING_LEN              ((1<<MAX_STRING_SIZE_LOG) - 2)

class CCharPosDNA {
public:
    CCharPosDNA(EAlphabet Type = eAlphaUnknown, unsigned char Char = 0, unsigned Pos = 0) {
        if (Pos == 255) {
            mPos = MAX_STRING_LEN + 1;
        } else {
            if (Pos > MAX_STRING_LEN) {
                THROW_AND_PRINT_RUNTIME_ERROR("Does not support pos: " << Pos)
            }
            mPos = Pos;
        }
        if (Type != eAlphaUnknown) {
            mCharIndx = ConvertChar(Type, Char);
        }
    }
    unsigned char   GetChar(const char* pABC) const { return pABC[mCharIndx]; }
    unsigned char   mPos:MAX_STRING_SIZE_LOG;
private:
    unsigned char   mCharIndx:MAX_ALPHA_SIZE_LOG;
}
#ifdef __GNUC__
 __attribute__ ((aligned(1)))
#endif
;



CAlternGenMorFrankPerfHashCompDNA::CAlternGenMorFrankPerfHashCompDNA(bool hasTransp, EAlphabet AlphaType) : 
                                                        CInterface(hasTransp, AlphaType), 
                                                        mpDictPerfectHash(NULL),
                                                        mbOpened(false)
{
    memset(maapHashContent,  0, sizeof maapHashContent);
    memset(maapCompactHashTable, 0, sizeof maapCompactHashTable);
    memset(maapPerfectHash,  0, sizeof maapPerfectHash);
}

struct CUniqueStrCollectorShard {
    ofstream*           mpTmpDictShard;
    bool*               mpHasString;

    CUniqueStrCollectorShard(ofstream* pTmpDictShard, bool* pHasString) : 
                            mpTmpDictShard(pTmpDictShard), 
                            mpHasString(pHasString) {}

    void operator()(const string& s, unsigned char* pPos, char* pCurrChar, unsigned k) {
        unsigned iShardId = hash_func(s) % SHARD_QTY;
        mpHasString[iShardId] = true;

        mpTmpDictShard[iShardId] << s << endl;
    }
};

struct CInsertStr {
    cmph_t**    mppPerfectHash; 
    ofstream*   mpTmpChunk;
    unsigned    maMaxHashVal[SHARD_QTY];
    EAlphabet   mAlphaType;

    CInsertStr(cmph_t** ppPerfectHash, ofstream* pTmpChunk,
               EAlphabet AlphaType) : mppPerfectHash(ppPerfectHash), mpTmpChunk(pTmpChunk), mAlphaType(AlphaType) {
        fill(maMaxHashVal, maMaxHashVal + SHARD_QTY, 0);
    }

    void operator()(const string& s, unsigned char* pPos, char* pCurrChar, unsigned k) {
        unsigned    NonPerfHashVal = hash_func(s);
        size_t      iShardId = NonPerfHashVal % SHARD_QTY;
        unsigned    HashVal = cmph_search(mppPerfectHash[iShardId], s.data(), s.size());

        maMaxHashVal[iShardId] = max(maMaxHashVal[iShardId], HashVal);

        size_t sz = s.size();

        mpTmpChunk[iShardId].write((const char*)&HashVal, sizeof(HashVal));
        for (unsigned i = 0; i < k; ++i) {
            CCharPosDNA    Elem(mAlphaType, pCurrChar[i], pPos[i]);
            mpTmpChunk[iShardId].write((const char*)&Elem, sizeof Elem);
        }
    }
};

struct CSearchStr {
    const CAlternGenMorFrankPerfHashCompDNA&   mIndex;
    const char*             mpABC;
    cmph_t**                mppPerfectHash; 
    size_t*                 mpTableSize;
    CCompactTable**         mppCompactHashTable;
    unsigned char**         mppHashContent;
    unsigned                mKerr;
    StringList*             mpResList;
    int                     mNumComp;
    unsigned                mNumCheck;
    unsigned                mNumSearchInBucket;

    CSearchStr(CAlternGenMorFrankPerfHashCompDNA& Index,
               const char* pABC,
               cmph_t** ppPerfectHash, 
               size_t* pTableSize, CCompactTable** ppCompactHashTable, unsigned char** ppHashContent, 
               unsigned kErr, StringList* pResList) :
                   mIndex(Index),
                   mpABC(pABC),
                   mppPerfectHash(ppPerfectHash), mpTableSize(pTableSize),
                   mppCompactHashTable(ppCompactHashTable), mppHashContent(ppHashContent), 
                   mKerr(kErr), mpResList(pResList) {
                        mNumCheck = 0;
                        mNumSearchInBucket = 0;
                        mNumComp = 0;
                   }

    void operator()(const string& s, unsigned char* pPatternPos, char* pCurrChar, unsigned kErrPattern) {
/*
            i1 i2 ... ik   IndexWord 
            j1 j2 ... jk   QueryWord (represented by pPos)
            ____________________________________________________________
            Transposition-unaware case.
            The comparing function produces indices:
            i_a1 ... i_al (in the index word, but not in the query word)
            j_b1 ... j_bm (in the query word, but not in the index word)
            k_1  ... k_n  (common indices)
            ____________________________________________________________
            
*/
        while (kErrPattern && pPatternPos[kErrPattern - 1] == 255) {
            --kErrPattern;
        }

        if (!mKerr) {
            ++mNumCheck;
            ++mNumSearchInBucket;

            ++mNumComp;

            if (mIndex.FindInDict(s.c_str(), s.size())) {
                if (mpResList) mpResList->push_back(s);
            }

            return;
        }
        unsigned ShardId = hash_func(s) % SHARD_QTY;
        if (!mppPerfectHash[ShardId]) return; // Shard is empty
        size_t HashVal = cmph_search(mppPerfectHash[ShardId], s.data(), s.size()) % mpTableSize[ShardId];

        size_t              DataSize = mKerr * sizeof(CCharPosDNA);
        size_t              OrigPatSize = kErrPattern + s.size();
        unsigned            PatSize = s.size();

        bool                bFirst = true;

        for (const CCharPosDNA* pData = (const CCharPosDNA*)(mppHashContent[ShardId] + DataSize * mppCompactHashTable[ShardId]->GetVal(HashVal)),
                            *pDataEnd = (const CCharPosDNA*)(mppHashContent[ShardId] + DataSize * mppCompactHashTable[ShardId]->GetVal(HashVal + 1));
                                    pData < pDataEnd;
                                    pData += mKerr) {
            int NumIndex = 0, NumPattern = 0; 
            unsigned kEstim = 0;

            ++mNumCheck;

            unsigned kErrIndex = mKerr;

            while (kErrIndex && pData[kErrIndex - 1].mPos == (MAX_STRING_LEN + 1)) {

                --kErrIndex;
            }

            if (kErrIndex && pData[kErrIndex - 1].mPos > PatSize) {
                continue;
            }

            unsigned kDel = 0;

            while (NumIndex < kErrIndex && NumPattern < kErrPattern) {
                unsigned char IndexPos = pData[NumIndex].mPos;
                unsigned char PatternPos = pPatternPos[NumPattern];

                if (IndexPos < PatternPos) {
                    ++NumIndex;
                } else if (IndexPos > PatternPos) {
                    ++NumPattern;
                } else {
                    ++NumIndex;
                    ++NumPattern;
                }
                ++kEstim;
            }
            kEstim += kErrIndex - NumIndex;
            kEstim += kErrPattern - NumPattern;

            //continue;

            if (kEstim <= mKerr) {
                ++mNumSearchInBucket;
                //continue;

                const char* pOrigPat  = s.data();

                unsigned MaxPossibSize = s.size() + kErrPattern + kErrIndex;
                char aModPat[MaxPossibSize + 1];

                unsigned NumIndex = 0;
                unsigned NumPattern = 0;

                unsigned PrevPatPos = 0;
                unsigned PatAccountForDel = 0;
                unsigned PatAccountForIns = 0;

                while (NumIndex < kErrIndex && NumPattern < kErrPattern ) {
                    unsigned char IndexPos = pData[NumIndex].mPos;
                    unsigned char PatternPos = pPatternPos[NumPattern];

                    if (IndexPos < PatternPos) {
                        // Extra letter in the index => 
                        // 1. insert a wildcard in the original pattern
                        // 2. insert a character from an indexed string

                        unsigned qty = PatAccountForIns + IndexPos - PrevPatPos;
                        
                        memcpy(aModPat + PrevPatPos, pOrigPat + PrevPatPos - PatAccountForIns, qty);

                        PrevPatPos += qty;

                        aModPat[PrevPatPos] = pData[NumIndex].GetChar(mpABC);

                        ++PrevPatPos;
                        ++PatAccountForIns;
                        ++NumIndex;
                    } else if (IndexPos > PatternPos) {
                        // Skip extra letters in the pattern
                        unsigned qty = PatAccountForIns + PatternPos - PrevPatPos;

                        memcpy(aModPat + PrevPatPos, pOrigPat + PrevPatPos - PatAccountForIns, qty);

                        PrevPatPos += qty;
                        ++PatAccountForDel;
                        
                        ++NumPattern;
                    } else {
                        // Substitution: 
                        // 1. In the original pattern: will be handled through a Hamming-distance computation (insert a wildcard)
                        // 2. In the reduced pattern replace a character with a character from an indexed string
                        unsigned qty = PatAccountForIns + PatternPos - PrevPatPos;

                        memcpy(aModPat + PrevPatPos, pOrigPat + PrevPatPos - PatAccountForIns, qty);

                        PrevPatPos += qty; 

                        aModPat[PrevPatPos] = pData[NumIndex].GetChar(mpABC);

                        ++PrevPatPos;

                        ++PatAccountForDel;
                        ++PatAccountForIns;
                        
                        ++NumPattern;
                        ++NumIndex;
                    }
                }

                while (NumIndex < kErrIndex) {
                    unsigned char IndexPos = pData[NumIndex].mPos;
 
                    unsigned qty = PatAccountForIns + IndexPos - PrevPatPos;
                        
                    memcpy(aModPat + PrevPatPos, pOrigPat + PrevPatPos - PatAccountForIns, qty);

                    PrevPatPos += qty;

                    aModPat[PrevPatPos] = pData[NumIndex].GetChar(mpABC);

                    ++PrevPatPos;
                    ++PatAccountForIns;
                    ++NumIndex;
                }

                while (NumPattern < kErrPattern) {
                    unsigned char PatternPos = pPatternPos[NumPattern];
                    unsigned qty = PatAccountForIns + PatternPos - PrevPatPos;

                    memcpy(aModPat + PrevPatPos, pOrigPat + PrevPatPos - PatAccountForIns, qty);

                    PrevPatPos += qty;
                    ++PatAccountForDel;
                        
                    ++NumPattern;
                }

                unsigned qty = OrigPatSize - PatAccountForDel + PatAccountForIns - PrevPatPos;

                memcpy(aModPat + PrevPatPos, pOrigPat + PrevPatPos - PatAccountForIns, qty);

                PrevPatPos += qty;

                if (PrevPatPos > MaxPossibSize) {
                    THROW_AND_PRINT_RUNTIME_ERROR("Internal error, PrevPatPos is too big: " << PrevPatPos);
                }

                aModPat[PrevPatPos] = 0;

                
                if (bFirst) {
                    ++mNumComp;
                    bFirst = false;
                    if (!mIndex.FindInDict(aModPat, PrevPatPos)) {
                        return; // The indexed string is different than a string obtained from the pattern by deletions
                    }
                }

                if (mpResList) mpResList->push_back(aModPat);
            }
        }
    }
};

void
CAlternGenMorFrankPerfHashCompDNA::LoadDictPerfectHash()
{
    const string&   sHashName = PerfectHashDictName();

    FILE* fdHash = fopen(sHashName.c_str(), "r");

    if (!fdHash) {
        THROW_AND_PRINT_RUNTIME_ERROR("Cannot open file: '" << sHashName << "'");
    }

    if (!(mpDictPerfectHash = cmph_load(fdHash))) {
        THROW_AND_PRINT_RUNTIME_ERROR("Cannot load dictionary hash");
    }
}

void
CAlternGenMorFrankPerfHashCompDNA::LoadPerfectHash(unsigned kerr)
{
    if (kerr > MAX_KERR) {
        THROW_AND_PRINT_RUNTIME_ERROR("Kerr is too big: " << kerr);
    }

    for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
        if (!maaHasString[kerr][ShardId]) {
            continue;
        }
        if (maapPerfectHash[kerr][ShardId]) {
            THROW_AND_PRINT_RUNTIME_ERROR("The hash is already loaded, kerr: " << kerr);
        }

        const string&   sHashName = PerfectHashName(kerr, ShardId);

        FILE* fdHash = fopen(sHashName.c_str(), "r");

        if (!fdHash) {
            THROW_AND_PRINT_RUNTIME_ERROR("Cannot open file: '" << sHashName << "'");
        }

        if (!(maapPerfectHash[kerr][ShardId] = cmph_load(fdHash))) {
            THROW_AND_PRINT_RUNTIME_ERROR("Cannot load hash for kerr = " << kerr);
        }
    }
}

void 
CAlternGenMorFrankPerfHashCompDNA::PrintAddStat(SearchType type, int testNum, int kerr, ostream &out)
{
    out << " NumCheck:" << aNumCheck[kerr]/float(testNum) << " NumSearchInBucket: " << aNumSearchInBucket[kerr]/float(testNum) << endl; 
}

bool
CAlternGenMorFrankPerfHashCompDNA::CreateInMemIndex(const char *inputFile,
                 int insertCount, int *dcnt, 
                 bool silent)
{
    const char*     pABC = GetAlphabet(mAlphaType);

    bool bRetStat = false;

    try {
        // First create a dictionary perfect hash
        {
            ifstream    inp(inputFile);

            if (!inp) {
                Perror("Cannot open file: " +string(inputFile));
                return false;
            }

            inp.exceptions(ios::badbit);

            string w;

            vector<string>        vTmpDict;
            vector<const char*>   vpStrPtr;

            while (getline(inp, w)) {
                vTmpDict.push_back(w);
                vpStrPtr.push_back(vTmpDict.back().c_str());
            }

            if (dcnt) *dcnt = vTmpDict.size();

            const string&   sHashName = PerfectHashDictName();

            FILE* fdHash = fopen(sHashName.c_str(), "w");

            if (!fdHash) {
                Perror("Cannot open file: '" + sHashName + "'");
                return false;
            }

            // Source of keys
            cmph_io_adapter_t *pSource = cmph_io_vector_adapter((char **)&vpStrPtr[0], vpStrPtr.size());

            cmph_config_t* pConfig = cmph_config_new(pSource);
            cmph_config_set_algo(pConfig, CMPH_BDZ);

            cmph_config_set_mphf_fd(pConfig, fdHash);
            cmph_t* pHash = cmph_new(pConfig);
            if (!pHash) {
                cerr << "Failed to create a perfect hash!" << endl;
                return false;
            }
            if (!cmph_dump(pHash, fdHash)) {
                cerr << "Failed to dump the perfect hash (for the original dictionary)!" << endl;
                return false;
            }

            // Store string keys in the order specified by the hash
            
            for (unsigned i = 0; i < vTmpDict.size(); ++i) {
                size_t HashVal = cmph_search(pHash, vTmpDict[i].data(), vTmpDict[i].size());

                if (HashVal >= mvDictStr.size()) {
                    mvDictStr.resize(HashVal + 1);
                }
                mvDictStr[HashVal] = vTmpDict[i];
            }
            
            cmph_destroy(pHash);       
            cmph_config_destroy(pConfig);
            cmph_io_vector_adapter_destroy(pSource);
            fclose(fdHash);

            cout << "Dictionary hash is created!" << endl;
        }

        // Then create an additional hash over reduced-alphabet strings in several steps
        for (unsigned k = 1; k <= MaxTestKerr<CAlternGenMorFrankPerfHashCompDNA>(); ++k) {
            // Write unsorted data into several files/chunks/shards.
            
            {
                ofstream         aTmpDict[SHARD_QTY];

                for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
                    string fn = TmpDictShardName(k, ShardId);

                    aTmpDict[ShardId].open(fn.c_str(), ios_base::out | ios_base::trunc);

                    if (!aTmpDict[ShardId]) {
                        Perror("Cannot open file: " + fn);
                        return false; 
                    }
                    aTmpDict[ShardId].exceptions(ios::failbit | ios::badbit | ios::eofbit);
                }

           
                cout << "K = " << k << " Preparing strings to generate perfect hash" << endl;

                CUniqueStrCollectorShard    ShardWriter(aTmpDict, maaHasString[k]);

                for (size_t i = 0; i < mvDictStr.size(); ++i) {
                    GenAllDeletions(mvDictStr[i], 0, k, ShardWriter);
                }

                for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
                    aTmpDict[ShardId].close();

                    if (!maaHasString[k][ShardId]) continue;

                    string sTmpDictShardName = TmpDictShardName(k, ShardId);

                    if (!GeneratePerfectHash(sTmpDictShardName, PerfectHashName(k, ShardId), CMPH_BMZ)) {
                        cerr << "Failed to generate a perfect hash!" << endl;
                        return false;
                    }

                    if (unlink(sTmpDictShardName.c_str())) {
                        Perror("Cannot delete file: '" + sTmpDictShardName + "'");
                        return false;
                    }

                    cout << "Created a perfect hash for shard # " << ShardId << endl;
                }

                LoadPerfectHash(k);
            }

            LoadDictPerfectHash();

            ofstream         aTmpChunk[SHARD_QTY];

            for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
                if (!maaHasString[k][ShardId]) continue;

                string fn = TmpChunkName(k, ShardId);

                aTmpChunk[ShardId].open(fn.c_str(), ios_base::out | ios_base::trunc);

                if (!aTmpChunk[ShardId]) {
                    Perror("Cannot open file: " + fn);
                    return false; 
                }
                aTmpChunk[ShardId].exceptions(ios::failbit | ios::badbit | ios::eofbit);
            }

            

            cout << "K = " << k << " The number of unique dictionary entries in the reduced alphabet: " << mvDictStr.size() << endl;

            CInsertStr      Inserter(maapPerfectHash[k], aTmpChunk, mAlphaType);

            // 2. Produce strings obtained by up to k deletions, save this data to the disk (in chunks)

            for (size_t i = 0; i < mvDictStr.size(); ++i) {
                GenAllDeletions(mvDictStr[i], 0, k, Inserter);
            }

            cout << "Closing temp chunks" << endl;

            for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
                if (!maaHasString[k][ShardId]) continue;

                aTmpChunk[ShardId].close();
            }


            long long TotalQty = 0;
            long long TotalTableDataSize = 0;
            long long TotalTableSize = 0;

            // 3. Read all the chunks, sort the data, and remove duplicates.
            cout << "Reading temp chunks." << endl;

            CCharPosDNA         aData[k];
            size_t              ElemSize = k * sizeof(CCharPosDNA);

            for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
                if (!maaHasString[k][ShardId]) continue;

                size_t TableSize = maaTableSize[k][ShardId] = Inserter.maMaxHashVal[ShardId] + 1;
                cout << "K = " << k << " ShardId: " << ShardId << " Table size: " << TableSize << endl;


                vector<size_t>      vHashTable(TableSize + 1);
                size_t*             pTmpHashTable =  &vHashTable[0];

                maapCompactHashTable[k][ShardId] = new CCompactTable(TableSize + 1, SKIP_LOG);

                CCompactTable&      CompactHashTable = *maapCompactHashTable[k][ShardId]; 

                fill(pTmpHashTable, pTmpHashTable + TableSize + 1, 0);

                ifstream    inp;
                string      fn = TmpChunkName(k, ShardId);

                inp.open(fn.c_str());

                if (!inp) {
                    Perror("Cannot open file '" + fn + "' for reading");
                    return false;
                }

                inp.exceptions(ios::badbit);

                Unsigned2StringList hTmp;

                char c;
                inp.read(&c, 0);

                while (!inp.eof()) {
                    size_t          HashVal;

                    if (!inp.read((char*)&HashVal, sizeof HashVal) && inp.eof()) break;
                    inp.read((char*)aData, ElemSize);

                    hTmp[HashVal].push_back(string((const char*)aData, ElemSize));
                };

                inp.close();

                size_t ShardQty = 0;
        
                for (Unsigned2StringList::iterator it = hTmp.begin(); it != hTmp.end(); ++it) {
                    size_t HashVal = it->first;

                    if (HashVal >= TableSize) {
                        cerr << "Internal error: a HashVal = " << HashVal << " is >= than TableSize = " << TableSize << " Shard id: " << ShardId << " k = " << k << endl;
                        return false;
                    }
                    StringList& lst = it->second;

                    sort(lst.begin(), lst.end());
                    unsigned qty = unique(lst.begin(), lst.end()) - lst.begin();
                    lst.erase(lst.begin() + qty, lst.end());
                    ShardQty += qty;
                    pTmpHashTable[HashVal + 1] += qty;
                }


                TotalQty += ShardQty;

                size_t ShardSize = ShardQty * ElemSize;
                maapHashContent[k][ShardId] = new unsigned char [ShardSize];
                unsigned char* pHashContent = maapHashContent[k][ShardId];

                fill(pHashContent, pHashContent + ShardSize, 0);

                unsigned Skip = CompactHashTable.GetSkip();

                for (size_t i = 0; i <= TableSize; i += Skip) {
                    size_t iNext = min(TableSize + 1, i + Skip);
                    size_t aBlock[Skip];
                    unsigned BlockQty = iNext - i;
                    for (unsigned j = 0; j < BlockQty; ++j) {
                        aBlock[j] = pTmpHashTable[i + j];
                    }
                    CompactHashTable.AddBlock(aBlock, BlockQty);
                }

                if (CompactHashTable.GetVal(0)) {
                    THROW_AND_PRINT_RUNTIME_ERROR("Internal error: the first hash table element is non-zero and is equal to: " << CompactHashTable.GetVal(0));
                }

                size_t cumul = 0;

                // Verify the correctness
                for (size_t i = 0; i <= TableSize; i++) {
                    size_t val = CompactHashTable.GetVal(i);

                    cumul += pTmpHashTable[i];

                    if (cumul != val ) {
                        THROW_AND_PRINT_RUNTIME_ERROR("Internal error, table verification failed for element index: " << i
                                                  << " expected value: " << cumul * ElemSize << " obtained value: " << val
                                                  << " TableSize: " << TableSize);
                    }
                }

                TotalTableDataSize += CompactHashTable.GetDataSize();
                TotalTableSize += TableSize + 1;


                for (Unsigned2StringList::iterator it = hTmp.begin(); it != hTmp.end(); ++it) {
                    size_t HashVal = it->first;

                    if (HashVal >= TableSize) {
                        cerr << "Internal error: a HashVal = " << HashVal << " is > than TableSize = " << TableSize << endl;
                        return false;
                    }
                    StringList& lst = it->second;
                    unsigned qty = lst.size();

                    size_t RealSize = qty * ElemSize;
                    size_t ExpSize =  ElemSize * (CompactHashTable.GetVal(HashVal + 1) - CompactHashTable.GetVal(HashVal));

                    if (RealSize != ExpSize) {
                        THROW_AND_PRINT_RUNTIME_ERROR("Internal error, expected size: " << ExpSize << 
                                                      " Obtained: " << RealSize << 
                                                      " The number of elements: " << qty << 
                                                      " Element size: " << ElemSize << 
                                                      " HashVal: " << HashVal);
                    }
                    size_t Start = CompactHashTable.GetVal(HashVal) * ElemSize;

                    for (unsigned i = 0; i < qty; ++i) {
                        if (lst[i].size() != ElemSize) {
                            THROW_AND_PRINT_RUNTIME_ERROR("Internal error, element size: " << lst[i].size() << " is not equal to the expected value: " << ElemSize << " HashVal = " << HashVal << " k = " << k << " i = " << i);
                        }
                        memcpy(&pHashContent[Start], lst[i].data(), ElemSize);
                        Start += ElemSize;
                    }

                    if (Start != ElemSize * CompactHashTable.GetVal(HashVal + 1)) {
                        THROW_AND_PRINT_RUNTIME_ERROR("Internal error: Start = " << Start << " does not match the next offset: " << (ElemSize * CompactHashTable.GetVal(HashVal + 1)) << " HashVal: " << HashVal);
                    }
                }

                string TmpName = TmpChunkName(k, ShardId);

                if (unlink(TmpName.c_str())) {
                    Perror("Cannot delete file: '" + TmpName + "'");
                    return false;
                }
            }

            long long TotalSize = ((long long)TotalQty) * ElemSize;
            cout << "Total # of entries: " << TotalQty  << " kerr = " << k << endl
             << " Total # of bytes for the hash content: " << TotalSize  << endl
             << " Total # of bytes for the hash content + hash table: " << (TotalSize + TotalTableDataSize) << endl
             << " The average size of table entry: " << TotalTableDataSize / float(TotalTableSize) << endl;
        
        }

        for (unsigned k = 1; k <= MaxTestKerr<CAlternGenMorFrankPerfHashCompDNA>(); ++k) {
            ClearPerfectHash(k);
        }
        for (unsigned k = 1; k <= MaxTestKerr<CAlternGenMorFrankPerfHashCompDNA>(); ++k) {
            LoadPerfectHash(k);
        }

        bRetStat = true;
    } catch (const exception& e) {
        cerr << "Exception during indexing: " << e.what() << endl;
    }

    return bRetStat;
}


bool
CAlternGenMorFrankPerfHashCompDNA::Insert(const char *str)
{
    return true;
}

bool
CAlternGenMorFrankPerfHashCompDNA::Search(const char *str, int kerr, int *numComp,
       StringList * resList, bool silent)
{
    string sOrig(str);

    CSearchStr Searcher(*this,
                        GetAlphabet(mAlphaType),
                        maapPerfectHash[kerr], 
                        maaTableSize[kerr], 
                        maapCompactHashTable[kerr], 
                        maapHashContent[kerr], 
                        kerr, resList);

    GenAllDeletions(sOrig, 0, kerr, Searcher);

    if (numComp) *numComp += Searcher.mNumComp;
    aNumCheck[kerr] += Searcher.mNumCheck;
    aNumSearchInBucket[kerr] += Searcher.mNumSearchInBucket;

    return true;
}

bool
CAlternGenMorFrankPerfHashCompDNA::DumpDB(const char *base_name)
{
    // All the perfect hash tables are already dumped in the form of mph files
    for (unsigned kerr = 1; kerr <= MaxTestKerr<CAlternGenMorFrankPerfHashCompDNA>(); ++kerr) {
        for (unsigned ShardId = 0; ShardId < SHARD_QTY; ++ShardId) {
            if (!maaHasString[kerr][ShardId]) continue;

            string sHashContent = HashContentName(kerr, ShardId);
            string sCompactHashTable = CompactHashTableName(kerr, ShardId);
        
            ofstream OutHashContent(sHashContent.c_str(), ios_base::out | ios_base::trunc);
            if (!OutHashContent) {
                Perror("Cannot open: '" + sHashContent + "' for writing");
                return false;
            }
            ofstream OutCompactHashTable(sCompactHashTable.c_str(), ios_base::out | ios_base::trunc);
            if (!OutCompactHashTable) {
                Perror("Cannot open: '" + sCompactHashTable + "' for writing");
                return false;
            }

            maapCompactHashTable[kerr][ShardId]->Dump(OutCompactHashTable);

            OutHashContent.exceptions(ios::failbit | ios::badbit | ios::eofbit);

            size_t  DataSize = kerr * sizeof(CCharPosDNA);

            OutHashContent.write(reinterpret_cast<const char*>(maapHashContent[kerr][ShardId]), 
                                 DataSize * maapCompactHashTable[kerr][ShardId]->GetVal(maaTableSize[kerr][ShardId]));
        }
    }

    return true;
}

bool
CAlternGenMorFrankPerfHashCompDNA::OpenDB(const char *base_name, bool create)
{
    return true;
}

bool
CAlternGenMorFrankPerfHashCompDNA::CloseDB()
{
    return true;
}

#ifdef _DEBUG
void 
CAlternGenMorFrankPerfHashCompDNA::Print()
{
}
#endif
