#include <cassert>

#include "myexcept.h"
#include "alphabet.h"
#include "pack.h"

inline unsigned PackPos(unsigned char Pos) {
    if (255 == Pos) { 
        return MAX_STRING_LEN + 1;  // A special marker
    } else if (Pos > MAX_STRING_LEN) {
        THROW_AND_PRINT_RUNTIME_ERROR("Position is too large: " << Pos);
    }

    return Pos;
}

inline unsigned char UnpackPos(unsigned Pos) {
    return (Pos >= MAX_STRING_LEN + 1) ? 255 : Pos;
};

size_t PackedSize(size_t Qty, unsigned k) {
    switch (k) {
        case 1: return 2 * Qty;
        case 2: return sizeof(CCharPos2) * Qty;
        case 3: return sizeof(CCharPos3_2) * (Qty / 2) + sizeof(CCharPos3_1) * (Qty & 1);
        case 4: return sizeof(CCharPos4) * Qty; 
        default: THROW_AND_PRINT_RUNTIME_ERROR("Unsupported k = " << k);
    }
};

size_t PackCharPos1(EAlphabet kAlphaType, 
               const unsigned char pCharPosArr[], 
               size_t Qty, 
               char* pDst, size_t DstSize) 
{
    size_t ReqSize = 2 * Qty;

    if (DstSize < ReqSize) {
        THROW_AND_PRINT_RUNTIME_ERROR("The destination array is too small (k=1), required: " << ReqSize << " bytes, available: " << DstSize << " bytes");
    }

    memcpy(pDst, pCharPosArr, ReqSize);

    return ReqSize;
}

size_t PackCharPos2(EAlphabet kAlphaType, 
               const unsigned char pCharPosArr[], 
               size_t Qty, 
               char   pDstArr[], size_t DstSize) 
{
    size_t ReqSize = sizeof(CCharPos2) * Qty;

    if (DstSize < ReqSize) {
        THROW_AND_PRINT_RUNTIME_ERROR("The destination array is too small (k=2), required: " << ReqSize << " bytes, available: " << DstSize << " bytes");
    }
    
    const unsigned char*    pSrc = pCharPosArr;
    char*                   pDst = pDstArr;

    CCharPos2               CurrElem;

    for (size_t i = 0; i < Qty; ++i) {
        CurrElem.mCharIndex1    = ConvertChar(kAlphaType, *pSrc++);
        CurrElem.mPos1          = PackPos(*pSrc++);
        CurrElem.mCharIndex2    = ConvertChar(kAlphaType, *pSrc++);
        CurrElem.mPos2          = PackPos(*pSrc++);

        memcpy(pDst, &CurrElem, sizeof CurrElem);
        pDst += sizeof CurrElem;
    }

    assert(pDst - pDstArr == ReqSize);

    return ReqSize;
}

size_t PackCharPos3(EAlphabet kAlphaType, 
               const unsigned char pCharPosArr[], 
               size_t Qty,
               char pDstArr[], size_t DstSize) 
{
    size_t QtyHalf = Qty / 2;
    size_t ReqSize = sizeof(CCharPos3_2) * QtyHalf + (Qty & 1) * sizeof(CCharPos3_1);

    if (DstSize < ReqSize) {
        THROW_AND_PRINT_RUNTIME_ERROR("The destination array is too small (k=3), required: " << ReqSize << " bytes, available: " << DstSize << " bytes");
    }
    
    const unsigned char*    pSrc = pCharPosArr;
    char*                   pDst = pDstArr;

    CCharPos3_2             TwoCurrElem;

    // Packed two elements together
    for (size_t i = 0; i < QtyHalf; ++i) {
        TwoCurrElem.mCharIndex1_1    = ConvertChar(kAlphaType, *pSrc++);
        TwoCurrElem.mPos1_1          = PackPos(*pSrc++);
        TwoCurrElem.mCharIndex1_2    = ConvertChar(kAlphaType, *pSrc++);
        TwoCurrElem.mPos1_2          = PackPos(*pSrc++);
        TwoCurrElem.mCharIndex1_3    = ConvertChar(kAlphaType, *pSrc++);
        TwoCurrElem.mPos1_3          = PackPos(*pSrc++);

        TwoCurrElem.mCharIndex2_1    = ConvertChar(kAlphaType, *pSrc++);
        TwoCurrElem.mPos2_1          = PackPos(*pSrc++);
        TwoCurrElem.mCharIndex2_2    = ConvertChar(kAlphaType, *pSrc++);
        TwoCurrElem.mPos2_2          = PackPos(*pSrc++);
        TwoCurrElem.mCharIndex2_3    = ConvertChar(kAlphaType, *pSrc++);
        TwoCurrElem.mPos2_3          = PackPos(*pSrc++);

        memcpy(pDst, &TwoCurrElem, sizeof TwoCurrElem);
        pDst += sizeof TwoCurrElem;
    }

    if (Qty & 1) {
        CCharPos3_1             CurrElem;

        CurrElem.mCharIndex1    = ConvertChar(kAlphaType, *pSrc++);
        CurrElem.mPos1          = PackPos(*pSrc++);
        CurrElem.mCharIndex2    = ConvertChar(kAlphaType, *pSrc++);
        CurrElem.mPos2          = PackPos(*pSrc++);
        CurrElem.mCharIndex3    = ConvertChar(kAlphaType, *pSrc++);
        CurrElem.mPos3          = PackPos(*pSrc++);

        memcpy(pDst, &CurrElem, sizeof CurrElem);
        pDst += sizeof CurrElem;
    }

    return ReqSize;
}

size_t PackCharPos4(EAlphabet kAlphaType, 
               const unsigned char pCharPosArr[], 
               size_t Qty,
               char pDstArr[], size_t DstSize) 
{
    size_t ReqSize = sizeof(CCharPos4) * Qty;

    if (DstSize < ReqSize) {
        THROW_AND_PRINT_RUNTIME_ERROR("The destination array is too small (k=4), required: " << ReqSize << " bytes, available: " << DstSize << " bytes");
    }
    
    const unsigned char*    pSrc = pCharPosArr;
    char*                   pDst = pDstArr;

    CCharPos4               CurrElem;

    for (size_t i = 0; i < Qty; ++i) {
        CurrElem.mCharIndex1    = ConvertChar(kAlphaType, *pSrc++);
        CurrElem.mPos1          = PackPos(*pSrc++);
        CurrElem.mCharIndex2    = ConvertChar(kAlphaType, *pSrc++);
        CurrElem.mPos2          = PackPos(*pSrc++);
        CurrElem.mCharIndex3    = ConvertChar(kAlphaType, *pSrc++);
        CurrElem.mPos3          = PackPos(*pSrc++);
        CurrElem.mCharIndex4    = ConvertChar(kAlphaType, *pSrc++);
        CurrElem.mPos4          = PackPos(*pSrc++);

        memcpy(pDst, &CurrElem, sizeof CurrElem);
        pDst += sizeof CurrElem;
    }

    assert(pDst - pDstArr == ReqSize);

    return ReqSize;
}

size_t PackCharPos(EAlphabet kAlphaType, 
                   const unsigned char pCharPosArr[], 
                   size_t SrcQty, unsigned k,
                   char* pDst, size_t DstSize) {
    if (1 == k) {
        return PackCharPos1(kAlphaType, pCharPosArr, SrcQty, pDst, DstSize);
    } else if (2 == k) {
        return PackCharPos2(kAlphaType, pCharPosArr, SrcQty, pDst, DstSize);
    } else if (3 == k) {
        return PackCharPos3(kAlphaType, pCharPosArr, SrcQty, pDst, DstSize);
    } else if (4 == k) {
        return PackCharPos4(kAlphaType, pCharPosArr, SrcQty, pDst, DstSize);
    }

    THROW_AND_PRINT_RUNTIME_ERROR("Invalid k = " << k << " Acceptable values are in the range from 1 to 4");
}

size_t UnpackCharPos1(const unsigned char* pSrc, size_t SrcSize,
                    unsigned char* pDstChar, unsigned char* pDstPos, size_t DstSize)
{
    if (SrcSize & 1) {
        THROW_AND_PRINT_RUNTIME_ERROR("Expecting even number of bytes (k=1)");
    }

    size_t  ReqSize = 1 * SrcSize / 2;

    if (ReqSize > DstSize) {
        THROW_AND_PRINT_RUNTIME_ERROR("The destination array is too small (k=1), required: " << ReqSize << " bytes, available: " << DstSize << " bytes");
    }

    const unsigned char* pSrcEnd = pSrc + SrcSize; 

    while (pSrc < pSrcEnd) {
        *pDstChar++ = *pSrc++;
        *pDstPos++  = *pSrc++;
    }

    return ReqSize;
}

size_t UnpackCharPos2(const char* pABC,
                    const unsigned char* pSrc, size_t SrcSize,
                    unsigned char* pDstChar, unsigned char* pDstPos, size_t DstSize)
{
    if (SrcSize % sizeof(CCharPos2)) {
        THROW_AND_PRINT_RUNTIME_ERROR("The number of bytes is not a multiple of " << sizeof(CCharPos2) << " (k=2)");
    }
    size_t  ReqSize = 2 * SrcSize / sizeof(CCharPos2);

    if (ReqSize > DstSize) {
        THROW_AND_PRINT_RUNTIME_ERROR("The destination array is too small (k=2), required: " << ReqSize << " bytes, available: " << DstSize << " bytes");
    }

    const unsigned char* pSrcEnd = pSrc + SrcSize; 

    const CCharPos2*  pCurrElem;

    while (pSrc < pSrcEnd) {
        pCurrElem = reinterpret_cast<const CCharPos2*>(pSrc);
        *pDstChar++ = pABC[pCurrElem->mCharIndex1];
        *pDstPos++  = UnpackPos(pCurrElem->mPos1);
        *pDstChar++ = pABC[pCurrElem->mCharIndex2];
        *pDstPos++  = UnpackPos(pCurrElem->mPos2);

        pSrc += sizeof *pCurrElem; 
    }

    return ReqSize;
}

size_t UnpackCharPos3(const char* pABC,
                    const unsigned char* pSrc, size_t SrcSize,
                    unsigned char* pDstChar, unsigned char* pDstPos, size_t DstSize)
{

    size_t  Qty1        = SrcSize / sizeof(CCharPos3_2);
    size_t  Remainder   = SrcSize % sizeof(CCharPos3_2); 
    size_t  ReqSize     = 3 * 2 * Qty1;

    if (!Remainder) {
    } else if (Remainder == sizeof(CCharPos3_1)) {
        ReqSize += 3;
    } else {
        THROW_AND_PRINT_RUNTIME_ERROR("Invalid number of bytes: " << SrcSize << " remainder: " << Remainder << " (k=3)");
    }

    if (ReqSize > DstSize) {
        THROW_AND_PRINT_RUNTIME_ERROR("The destination array is too small (k=3), required: " << ReqSize << " bytes, available: " << DstSize << " bytes");
    }

    const CCharPos3_2*  pCurrTwoElem;

    while (Qty1--) {
        pCurrTwoElem = reinterpret_cast<const CCharPos3_2*>(pSrc);

        *pDstChar++ = pABC[pCurrTwoElem->mCharIndex1_1];
        *pDstPos++  = UnpackPos(pCurrTwoElem->mPos1_1);
        *pDstChar++ = pABC[pCurrTwoElem->mCharIndex1_2];
        *pDstPos++  = UnpackPos(pCurrTwoElem->mPos1_2);
        *pDstChar++ = pABC[pCurrTwoElem->mCharIndex1_3];
        *pDstPos++  = UnpackPos(pCurrTwoElem->mPos1_3);

        *pDstChar++ = pABC[pCurrTwoElem->mCharIndex2_1];
        *pDstPos++  = UnpackPos(pCurrTwoElem->mPos2_1);
        *pDstChar++ = pABC[pCurrTwoElem->mCharIndex2_2];
        *pDstPos++  = UnpackPos(pCurrTwoElem->mPos2_2);
        *pDstChar++ = pABC[pCurrTwoElem->mCharIndex2_3];
        *pDstPos++  = UnpackPos(pCurrTwoElem->mPos2_3);

        pSrc += sizeof *pCurrTwoElem; 
    }

    if (Remainder) {
        const CCharPos3_1*  pCurrElem = reinterpret_cast<const CCharPos3_1*>(pSrc);

        *pDstChar++ = pABC[pCurrElem->mCharIndex1];
        *pDstPos++  = UnpackPos(pCurrElem->mPos1);
        *pDstChar++ = pABC[pCurrElem->mCharIndex2];
        *pDstPos++  = UnpackPos(pCurrElem->mPos2);
        *pDstChar++ = pABC[pCurrElem->mCharIndex3];
        *pDstPos++  = UnpackPos(pCurrElem->mPos3);
    }

    return ReqSize;
}

size_t UnpackCharPos4(const char* pABC,
                    const unsigned char* pSrc, size_t SrcSize,
                    unsigned char* pDstChar, unsigned char* pDstPos, size_t DstSize)
{
    if (SrcSize % sizeof(CCharPos4)) {
        THROW_AND_PRINT_RUNTIME_ERROR("The number of bytes is not a multiple of " << sizeof(CCharPos4) << " (k=2)");
    }
    size_t  ReqSize = 4 * SrcSize / sizeof(CCharPos4);

    if (ReqSize > DstSize) {
        THROW_AND_PRINT_RUNTIME_ERROR("The destination array is too small (k=4), required: " << ReqSize << " bytes, available: " << DstSize << " bytes");
    }

    const unsigned char* pSrcEnd = pSrc + SrcSize; 

    const CCharPos4*  pCurrElem;

    while (pSrc < pSrcEnd) {
        pCurrElem = reinterpret_cast<const CCharPos4*>(pSrc);

        *pDstChar++ = pABC[pCurrElem->mCharIndex1];
        *pDstPos++  = UnpackPos(pCurrElem->mPos1);
        *pDstChar++ = pABC[pCurrElem->mCharIndex2];
        *pDstPos++  = UnpackPos(pCurrElem->mPos2);
        *pDstChar++ = pABC[pCurrElem->mCharIndex3];
        *pDstPos++  = UnpackPos(pCurrElem->mPos3);
        *pDstChar++ = pABC[pCurrElem->mCharIndex4];
        *pDstPos++  = UnpackPos(pCurrElem->mPos4);

        pSrc += sizeof *pCurrElem; 
    }

    return ReqSize;
}

size_t UnpackCharPos(const char* pABC,
                    const unsigned char* pSrc, size_t SrcSize, unsigned k,
                    unsigned char* pDstChar, unsigned char* pDstPos, size_t DstSize)
{
    if (1 == k) {
        return UnpackCharPos1(pSrc, SrcSize, pDstChar, pDstPos, DstSize);
    } else if (2 == k) {
        return UnpackCharPos2(pABC, pSrc, SrcSize, pDstChar, pDstPos, DstSize);
    } else if (3 == k) {
        return UnpackCharPos3(pABC, pSrc, SrcSize, pDstChar, pDstPos, DstSize);
    } else if (4 == k) {
        return UnpackCharPos4(pABC, pSrc, SrcSize, pDstChar, pDstPos, DstSize);
    }

    THROW_AND_PRINT_RUNTIME_ERROR("Invalid k = " << k << " Acceptable values are in the range from 1 to 4");
}
