#include <fstream>
#include <limits>
#include <iomanip>

#include "index.h"
#include "../Common/hash.h"
#include "../Common/size_mult.h"  

#define SKIP_LOG   5
#define WRITE_CHUNK_QTY  20

unsigned    aNumCheck[MAX_KERR + 1];
unsigned    aNumSearchInBucket[MAX_KERR + 1];
unsigned    aNumSearchInBucketMiss[MAX_KERR + 1];

CAlternGenRedAlphaMorFrankCompact::CAlternGenRedAlphaMorFrankCompact(bool hasTransp, EAlphabet AlphaType) : 
                                                        CInterface(hasTransp, AlphaType), 
                                                        mRedHash(hasTransp, AlphaType), 
                                                        mbOpened(false)
{
    memset(mapHashContent, 0, sizeof mapHashContent);
    memset(mapCompactHashTable, 0, sizeof mapCompactHashTable);
}

struct CInsertStr {
    size_t      mTableSize; 
    ofstream*   mpTmpChunk;

    CInsertStr(size_t TableSize, ofstream* pTmpChunk) : mTableSize(TableSize), mpTmpChunk(pTmpChunk){}

    void operator()(const string& s, unsigned char* pPos, char* pCurrChar, unsigned k) {
        size_t HashVal = hash_func(s) % mTableSize;
        size_t iChunkNum = HashVal % WRITE_CHUNK_QTY;

        mpTmpChunk[iChunkNum].write((const char*)&HashVal, sizeof(HashVal));
        mpTmpChunk[iChunkNum].write((const char*)pPos, k);
        mpTmpChunk[iChunkNum].write((const char*)pCurrChar, k);
    }
};

struct CSearchStr {
    CAlternGenRedAlpha&     mRedHash;
    size_t                  mTableSize; 
    const CCompactTable&    mCompactHashTable;
    const unsigned char*    mpHashContent;
    string                  mOrigPat;
    unsigned                mKerr;
    StringList*             mpResList;
    int                     mNumComp;
    unsigned                mNumCheck;
    unsigned                mNumSearchInBucket;
    unsigned                mNumSearchInBucketMiss;

    CSearchStr(CAlternGenRedAlpha& RedHash,
               size_t TableSize, const CCompactTable& CompactHashTable, const unsigned char* pHashContent, 
               const string& sOrigPat, unsigned kErr, StringList*   pResList) :
                   mRedHash(RedHash),
                   mTableSize(TableSize), mCompactHashTable(CompactHashTable), mpHashContent(pHashContent), 
                   mOrigPat(sOrigPat), mKerr(kErr), mpResList(pResList) {
                        mNumCheck = 0;
                        mNumSearchInBucket = 0;
                        mNumSearchInBucketMiss = 0;
                        mNumComp = 0;
                   }

    void operator()(const string& s, unsigned char* pPatternPos, char* pCurrChar, unsigned kErrPattern) {
/*
            i1 i2 ... ik   IndexWord 
            j1 j2 ... jk   QueryWord (represented by pPos)
            ____________________________________________________________
            Transposition-unaware case.
            The comparing function produces indices:
            i_a1 ... i_al (in the index word, but not in the query word)
            j_b1 ... j_bm (in the query word, but not in the index word)
            k_1  ... k_n  (common indices)
            ____________________________________________________________
            
*/
        while (kErrPattern && pPatternPos[kErrPattern - 1] == 255) {
            --kErrPattern;
        }

        if (!mKerr) {
            string        sPat = mOrigPat;
            const string& sRedPat = s;

            // Apply the same deletions to the original pattern as to the reduced pattern
            for (unsigned i = 0; i < kErrPattern; ++i) {
                sPat.erase(pPatternPos[i], 1);
            }

            if (sPat.size() != sRedPat.size()) {
                THROW_AND_PRINT_RUNTIME_ERROR("Internal error reduced and original patterns have different length (after deletions are applied to the original pattern)")        ;
           }
        
            ++mNumCheck;
            ++mNumSearchInBucket;

            unsigned tmp = mNumComp;

            mRedHash.SearchInBucket(sPat.c_str(), sRedPat.c_str(), mKerr, mpResList, mNumComp);

            if (tmp == mNumComp) {
                mNumSearchInBucketMiss++;
            }
            return;
        }
        size_t HashVal = hash_func(s) % mTableSize;

        size_t              DataSize = mKerr * 2;
        unsigned            RedPatSize = s.size();

        char                aData[2*mKerr];
        size_t              ElemSize = sizeof(aData);

        for (const unsigned char*   pIndexPos = mpHashContent + mCompactHashTable.GetVal(HashVal) * ElemSize,
                                *   pIndexPosEnd = mpHashContent + mCompactHashTable.GetVal(HashVal + 1) * ElemSize;
                                    pIndexPos < pIndexPosEnd;
                                    pIndexPos += DataSize) {
            int NumIndex = 0, NumPattern = 0; 
            unsigned kEstim = 0;

            ++mNumCheck;

            unsigned kErrIndex = mKerr;

            while (kErrIndex && pIndexPos[kErrIndex - 1] == 255) {
                --kErrIndex;
            }

            if (kErrIndex && pIndexPos[kErrIndex - 1] > RedPatSize) {
                continue;
            }

            unsigned kDel = 0;

            while (NumIndex < kErrIndex && NumPattern < kErrPattern) {
                unsigned char IndexPos = pIndexPos[NumIndex];
                unsigned char PatternPos = pPatternPos[NumPattern];

                if (IndexPos < PatternPos) {
                    ++NumIndex;
                } else if (IndexPos > PatternPos) {
                    ++NumPattern;
                } else {
                    ++NumIndex;
                    ++NumPattern;
                }
                ++kEstim;
            }
            kEstim += kErrIndex - NumIndex;
            kEstim += kErrPattern - NumPattern;

            //continue;

            if (kEstim <= mKerr) {
                ++mNumSearchInBucket;
                // Generate updated reduced-alphabet pattern as well as updated original pattern
                // Update includes deletion of certain characters as well as insertion of wildcards.
                const char *  pIndexChar = reinterpret_cast<const char *>(pIndexPos) + mKerr;


                //continue;

                const char* pOrigPat = mOrigPat.data();
                const char* pOrigRedPat  = s.data();

                unsigned MaxPossibSize = mOrigPat.size() + kErrIndex;
                char aPat[MaxPossibSize + 1];
                char aRedPat[MaxPossibSize + 1];

                unsigned NumIndex = 0;
                unsigned NumPattern = 0;

                unsigned PrevPatPos = 0;
                unsigned PatAccountForDel = 0;
                unsigned PatAccountForIns = 0;

                while (NumIndex < kErrIndex && NumPattern < kErrPattern ) {
                    unsigned char IndexPos = pIndexPos[NumIndex];
                    unsigned char PatternPos = pPatternPos[NumPattern];

                    if (IndexPos < PatternPos) {
                        // Extra letter in the index => 
                        // 1. insert a wildcard in the original pattern
                        // 2. insert a character from an indexed string

                        unsigned qty = PatAccountForIns + IndexPos - PrevPatPos;
                        
                        memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                        memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                        PrevPatPos += qty;

                        aRedPat[PrevPatPos] = pIndexChar[NumIndex];
                        aPat[PrevPatPos]    = REDALPHA_WILDCARD_CHAR;

                        ++PrevPatPos;
                        ++PatAccountForIns;
                        ++NumIndex;
                    } else if (IndexPos > PatternPos) {
                        // Skip extra letters in the pattern
                        unsigned qty = PatAccountForIns + PatternPos - PrevPatPos;

                        memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                        memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                        PrevPatPos += qty;
                        ++PatAccountForDel;
                        
                        ++NumPattern;
                    } else {
                        // Substitution: 
                        // 1. In the original pattern: will be handled through a Hamming-distance computation (insert a wildcard)
                        // 2. In the reduced pattern replace a character with a character from an indexed string
                        unsigned qty = PatAccountForIns + PatternPos - PrevPatPos;

                        memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                        memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                        PrevPatPos += qty; 

                        aRedPat[PrevPatPos] = pIndexChar[NumIndex];
                        aPat[PrevPatPos]    = REDALPHA_WILDCARD_CHAR;

                        ++PrevPatPos;

                        ++PatAccountForDel;
                        ++PatAccountForIns;
                        
                        ++NumPattern;
                        ++NumIndex;
                    }
                }

                while (NumIndex < kErrIndex) {
                    unsigned char IndexPos = pIndexPos[NumIndex];
 
                    unsigned qty = PatAccountForIns + IndexPos - PrevPatPos;
                        
                    memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                    memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                    PrevPatPos += qty;

                    aRedPat[PrevPatPos] = pIndexChar[NumIndex];
                    aPat[PrevPatPos]    = REDALPHA_WILDCARD_CHAR;

                    ++PrevPatPos;
                    ++PatAccountForIns;
                    ++NumIndex;
                }

                while (NumPattern < kErrPattern) {
                    unsigned char PatternPos = pPatternPos[NumPattern];
                    unsigned qty = PatAccountForIns + PatternPos - PrevPatPos;

                    memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                    memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                    PrevPatPos += qty;
                    ++PatAccountForDel;
                        
                    ++NumPattern;
                }

                unsigned qty = mOrigPat.size() - PatAccountForDel + PatAccountForIns - PrevPatPos;

                memcpy(aPat    + PrevPatPos, pOrigPat    + PrevPatPos - PatAccountForIns + PatAccountForDel, qty);
                memcpy(aRedPat + PrevPatPos, pOrigRedPat + PrevPatPos - PatAccountForIns, qty);

                PrevPatPos += qty;

                if (PrevPatPos > MaxPossibSize) {
                    THROW_AND_PRINT_RUNTIME_ERROR("Internal error, PrevPatPos is too big: " << PrevPatPos);
                }

                aPat[PrevPatPos] = 0;
                aRedPat[PrevPatPos] = 0;

                unsigned tmp = mNumComp;

                mRedHash.SearchInBucket(aPat, aRedPat, mKerr, mpResList, mNumComp);

                if (tmp == mNumComp) {
                    mNumSearchInBucketMiss++;
                }
            }
        }
    }
};

void 
CAlternGenRedAlphaMorFrankCompact::PrintAddStat(SearchType type, int testNum, int kerr, ostream &out)
{
    out << " NumCheck:" << aNumCheck[kerr]/float(testNum) << " NumSearchInBucket: " << aNumSearchInBucket[kerr]/float(testNum) << " NumSearchInBucketMiss:" << aNumSearchInBucketMiss[kerr]/float(testNum) << endl; 
}

bool
CAlternGenRedAlphaMorFrankCompact::CreateInMemIndex(const char *inputFile,
                 int insertCount, int *dcnt, 
                 bool silent)
{

    bool bRetStat = false;

    try {
        // First create a regular reduced-alphabet index
        if (!mRedHash.CreateInMemIndex(inputFile, insertCount, dcnt, silent)) {
            return false;
        }

        // Then create an additional hash over reduced-alphabet strings in three steps
        for (unsigned k = 1; k <= MaxTestKerr<CAlternGenRedAlphaMorFrankCompact>(); ++k) {
            // 1. Write unsorted data into several files/chunks.
            const THashData& HashData = mRedHash.GetHashData();
            ofstream         aTmpChunk[WRITE_CHUNK_QTY];

            for (unsigned i = 0; i < WRITE_CHUNK_QTY; ++i) {
                string fn = TmpChunkName(k, i);

                aTmpChunk[i].open(fn.c_str(), ios_base::out | ios_base::trunc);

                if (!aTmpChunk[i]) {
                    Perror("Cannot open file: " + fn);
                    return false; 
                }
                aTmpChunk[i].exceptions(ios::failbit | ios::badbit | ios::eofbit);
            }

            size_t TableSize = maTableSize[k] = from__stl_next_prime(HashData.size() * aSizeMultiplier[k]);
            

            cout << "K = " << k << " Hash size: " << HashData.size() << 
                    " Size multiplier: " << aSizeMultiplier[k] << 
                    " TableQty: "   << TableSize << endl;

            CInsertStr  Inserter(TableSize, aTmpChunk);

            for (THashData::const_iterator it = HashData.begin(); it != HashData.end(); ++it) {
                const string& s = it->first;
#if 0
                cout << it->first << " : " << HashVal << endl;
#endif
                
                GenAllDeletions(s, 0, k, Inserter);
            }
            cout << "Closing temp chunks" << endl;
            for (unsigned i = 0; i < WRITE_CHUNK_QTY; ++i) {
                aTmpChunk[i].close();
            }

            long long TotalQty = 0;

            // 1. Read all the chunks, sort the data, remove duplicates, and memorize the number of entries.
            cout << "Reading temp chunks (first pass)" << endl;

            mapCompactHashTable[k] = new CCompactTable(TableSize + 1, SKIP_LOG);

            vector<size_t>          vTmpHashTable(TableSize + 1);
            size_t*                 pTmpHashTable = &vTmpHashTable[0];

            fill(pTmpHashTable, pTmpHashTable + TableSize + 1, 0);

            CCompactTable&          CompactHashTable = *mapCompactHashTable[k];

            char                aData[2*k];
            size_t              ElemSize = sizeof(aData);

            for (unsigned i = 0; i < WRITE_CHUNK_QTY; ++i) {
                cout << "Chunk id: " << i << endl;
                ifstream    inp;
                string      fn = TmpChunkName(k, i);

                inp.open(fn.c_str());

                if (!inp) {
                    Perror("Cannot open file '" + fn + "' for reading");
                    return false;
                }
                inp.exceptions(ios::badbit);

                Unsigned2StringList hTmp;

                char c;
                inp.read(&c, 0);

                while (!inp.eof()) {
                    size_t          HashVal;

                    if (!inp.read((char*)&HashVal, sizeof HashVal) && inp.eof()) break;
                    inp.read((char*)aData, ElemSize);

                    hTmp[HashVal].push_back(string(aData, ElemSize));
                };

                inp.close();
        
                for (Unsigned2StringList::iterator it = hTmp.begin(); it != hTmp.end(); ++it) {
                    size_t HashVal = it->first;

                    if (HashVal >= TableSize) {
                        cerr << "Internal error: a HashVal = " << HashVal << " is >= than TableSize = " << TableSize << " Chunk id: " << i << " k = " << k << endl;
                        return false;
                    }
                    StringList& lst = it->second;

                    sort(lst.begin(), lst.end());
                    unsigned qty = unique(lst.begin(), lst.end()) - lst.begin();
                    TotalQty += qty;
                    pTmpHashTable[HashVal + 1] += qty;
#if 0
                    cout << "###: " << HashVal << " -> " << qty << endl;
#endif
                }
            }

            long long TotalSize = ((long long)TotalQty) * ElemSize;

            cout << "Total # of entries: " << TotalQty  << " kerr = " << k << endl
                 << " Total # of bytes for the hash content: " << TotalSize  << endl
                 << " Total # of bytes for the hash content + hash table: " << (TotalSize + CompactHashTable.GetDataSize()) << endl
                 << " The average size of table entry: " << CompactHashTable.GetDataSize() / float(TableSize + 1) << endl;

            if (TotalSize > numeric_limits<size_t>::max()) {
                THROW_AND_PRINT_RUNTIME_ERROR("The data size is too big and won't fit into the memory!");
            }

            mapHashContent[k] = new unsigned char [TotalSize];

            unsigned char* pHashContent = mapHashContent[k];

            fill(pHashContent, pHashContent + TotalSize, 0);

            unsigned Skip = CompactHashTable.GetSkip();

            for (size_t i = 0; i <= TableSize; i += Skip) {
                size_t iNext = min(TableSize + 1, i + Skip);
                size_t aBlock[Skip];
                unsigned BlockQty = iNext - i;
                for (unsigned j = 0; j < BlockQty; ++j) {
                    aBlock[j] = pTmpHashTable[i + j];
                }
                CompactHashTable.AddBlock(aBlock, BlockQty);
            }

            if (CompactHashTable.GetVal(0)) {
                THROW_AND_PRINT_RUNTIME_ERROR("Internal error: the first hash table element is non-zero and is equal to: " << CompactHashTable.GetVal(0));
            }

            size_t cumul = 0;

            // Verify the correctness
            for (size_t i = 0; i <= TableSize; i++) {
                size_t val = CompactHashTable.GetVal(i);

                cumul += pTmpHashTable[i];

                if (cumul != val ) {
                    THROW_AND_PRINT_RUNTIME_ERROR("Internal error, table verification failed for element index: " << i
                                                  << " expected value: " << cumul * ElemSize << " obtained value: " << val
                                                  << " TableSize: " << TableSize);
                }
            }

            // Convert quantities into offsets
            for (size_t i = 1; i <= TableSize; ++i) {
                pTmpHashTable[i] += pTmpHashTable[i - 1];
            }

            vTmpHashTable.clear();
            pTmpHashTable = NULL;

            // The final pass: fill the contents of the hash
            cout << "Reading temp chunks (second pass)" << endl;

            for (unsigned i = 0; i < WRITE_CHUNK_QTY; ++i) {
                cout << "Chunk id: " << i << endl;
                ifstream    inp;
                string      fn = TmpChunkName(k, i);

                inp.open(fn.c_str());

                if (!inp) {
                    Perror("Cannot open file '" + fn + "' for reading");
                    return false;
                }
                inp.exceptions(ios::badbit);

                Unsigned2StringList hTmp;

                while (!inp.eof()) {
                    size_t          HashVal;

                    if (!inp.read((char*)&HashVal, sizeof HashVal) && inp.eof()) break;
                    inp.read((char*)aData, ElemSize);

                    hTmp[HashVal].push_back(string(aData, ElemSize));
                }

                inp.close();
        
                for (Unsigned2StringList::iterator it = hTmp.begin(); it != hTmp.end(); ++it) {
                    size_t HashVal = it->first;

                    if (HashVal >= TableSize) {
                        cerr << "Internal error: a HashVal = " << HashVal << " is > than TableSize = " << TableSize << endl;
                        return false;
                    }
                    StringList& lst = it->second;

                    sort(lst.begin(), lst.end());
                    unsigned qty = unique(lst.begin(), lst.end()) - lst.begin();

                    size_t RealSize = qty * ElemSize;
                    size_t ExpSize =  ElemSize * (CompactHashTable.GetVal(HashVal + 1) - CompactHashTable.GetVal(HashVal));

                    if (RealSize != ExpSize) {
                        THROW_AND_PRINT_RUNTIME_ERROR("Internal error, expected size: " << ExpSize << 
                                                      " Obtained: " << RealSize << 
                                                      " The number of elements: " << qty << 
                                                      " Element size: " << ElemSize << 
                                                      " HashVal: " << HashVal);
                    }
                    size_t Start = ElemSize *CompactHashTable.GetVal(HashVal);

                    for (unsigned i = 0; i < qty; ++i) {
                        if (lst[i].size() != ElemSize) {
                            THROW_AND_PRINT_RUNTIME_ERROR("Internal error, element size: " << lst[i].size() << " is not equal to the expected value: " << ElemSize << " HashVal = " << HashVal << " k = " << k << " i = " << i);
                        }
                        memcpy(&pHashContent[Start], lst[i].data(), ElemSize);
                        Start += ElemSize;
                    }

                    if (Start != ElemSize * CompactHashTable.GetVal(HashVal + 1)) {
                        THROW_AND_PRINT_RUNTIME_ERROR("Internal error: Start = " << Start << " does not match the next offset: " << CompactHashTable.GetVal(HashVal + 1) << " HashVal: " << HashVal);
                    }
                }

                if (unlink(fn.c_str())) {
                    THROW_AND_PRINT_RUNTIME_ERROR("Cannot delete file: '" << fn << "', reason: " << strerror(errno));
                }
            }
        }
        bRetStat = true;
    } catch (const exception& e) {
        cerr << "Exception during indexing: " << e.what() << endl;
    }

    return bRetStat;
}


bool
CAlternGenRedAlphaMorFrankCompact::Insert(const char *str)
{
    return true;
}

bool
CAlternGenRedAlphaMorFrankCompact::Search(const char *str, int kerr, int *numComp,
       StringList * resList, bool silent)
{
    string sOrig(str);
    string sReduced(str);

    mRedHash.ToRedStr(sReduced);

    CSearchStr Searcher(mRedHash,
                        maTableSize[kerr], *mapCompactHashTable[kerr], mapHashContent[kerr], 
                        sOrig, kerr, resList);

    GenAllDeletions(sReduced, 0, kerr, Searcher);

    if (numComp) *numComp += Searcher.mNumComp;
    aNumCheck[kerr] += Searcher.mNumCheck;
    aNumSearchInBucket[kerr] += Searcher.mNumSearchInBucket;
    aNumSearchInBucketMiss[kerr] += Searcher.mNumSearchInBucketMiss;

    return true;
}

bool
CAlternGenRedAlphaMorFrankCompact::DumpDB(const char *base_name)
{
    if (!mRedHash.DumpDB(base_name)) {
        return false;
    }

    for (unsigned kerr = 1; kerr <= MaxTestKerr<CAlternGenRedAlphaMorFrankCompact>(); ++kerr) {
        string sHashContent = HashContentName(kerr);
        string sCompactHashTable = CompactHashTableName(kerr);
        
        ofstream OutHashContent(sHashContent.c_str(), ios_base::out | ios_base::trunc);

        if (!OutHashContent) {
            Perror("Cannot open: '" + sHashContent + "' for writing");
            return false;
        }

        ofstream OutCompactHashTable(sCompactHashTable.c_str(), ios_base::out | ios_base::trunc);

        if (!OutCompactHashTable) {
            Perror("Cannot open: '" + sCompactHashTable + "' for writing");
            return false;
        }

        mapCompactHashTable[kerr]->Dump(OutCompactHashTable);

        OutHashContent.exceptions(ios::failbit | ios::badbit | ios::eofbit);

        size_t              DataSize = kerr * 2;

        OutHashContent.write(reinterpret_cast<const char*>(mapHashContent[kerr]), 
                             DataSize * mapCompactHashTable[kerr]->GetVal(maTableSize[kerr]));
    }

    return true;
}

bool
CAlternGenRedAlphaMorFrankCompact::OpenDB(const char *base_name, bool create)
{
    return true;
}

bool
CAlternGenRedAlphaMorFrankCompact::CloseDB()
{
    return true;
}

#ifdef _DEBUG
void 
CAlternGenRedAlphaMorFrankCompact::Print()
{
}
#endif
