#include <sys/types.h>
#include <sys/mman.h>

#include <sstream>
#include <algorithm>
#include <../Common/myexcept.h>
#include <../Common/utils.h>
#include "testutil.h"

#include "erratatrie.h"

using namespace std;

void
CErrataTree::PrintErrataTrie(const TTrie& trie, int kerr) const
{ 
    if (trie.Empty()) {
        return;
    }
    ExtendedPrintVisitor    Prnt(mNodeInfo, mCombinedKetsInfo, cout);

    cout << " ==== Main trie: ====" << endl;
    PrintTrieStartNode<NODE_ALLOCATOR>(trie, Prnt, trie.GetRoot());
    cout << " ====================" << endl;

    for (size_t i = 0; i < Prnt.mKETs.size(); ++i) {
        cout << " === KET trie: " << Prnt.mKETs[i] << " Parent is: " << trie.GetRoot() << endl;
        TTrie   ChildTrie(Prnt.mKETs[i]);

        if (kerr > 1) {
            PrintErrataTrie(ChildTrie, kerr - 1);
        } else {
            PrintTrie(ChildTrie, cout);
        }
        cout << " ====================" << endl;
    }

    for (size_t i = 0; i < Prnt.mKETPrimes.size(); ++i) {
        cout << " === KET' trie: " << Prnt.mKETPrimes[i] << " Parent is: " << trie.GetRoot() << endl;
        TTrie   ChildTrie(Prnt.mKETPrimes[i]);

        if (kerr > 1) {
            PrintErrataTrie(ChildTrie, kerr - 1);
        } else {
            PrintTrie(ChildTrie, cout);
        }
        cout << " ====================" << endl;
    }
}

void
CErrataTree::Create(const vs& StrList, int kerr, bool bHasTransp)
{
    mHasTransp  = bHasTransp;
    mKerr       = kerr;

    if (StrList.empty()) {
        return;
    }

    cout << "Total memory before creation of trie: " << GetTotalMemUsed() << endl;

    for (size_t i = 0; i < StrList.size(); ++i) {
        if (TTrie::IsLongString(StrList[i])) {
            THROW_AND_PRINT_RUNTIME_ERROR("String: '" << StrList[i] << "' is too long");
        }
        mTrie.Insert(StrList[i]);
    }

    cout << "Total memory before errata tries creation: " << GetTotalMemUsed() << endl;

    CreateErrataTrees(mTrie, kerr, bHasTransp);

    cout << "Total memory after errata tries creation: " << GetTotalMemUsed() << endl;
}

bool 
CErrataTree::DumpErrataTree(int FDTrieOffsets, int FDTrie,
                  int FDAddNodeInfo, int FDCombKetOffsets, int FDCombKets)
{
    try {
        // First create an empty file long enough to keep the mapping info
        // from node index to node offset
        off_t  dummy = -1;
        size_t OffsetSize = 0;

        errno = 0;

        for (size_t i = 0; i <= CNodeAllocator::GetMaxNodeId(); ++i) {
            if (write(FDTrieOffsets, &dummy, sizeof dummy) != sizeof dummy) {
                Perror("Cannot write to trie offset file");
                return false;
            }
            OffsetSize += sizeof dummy;
        }
        // Then map this file to memory in write mode
        void *pBuff = mmap (NULL, OffsetSize, PROT_WRITE | PROT_READ, MAP_SHARED, FDTrieOffsets, 0);

        if (MAP_FAILED == pBuff) {
            Perror("Cannot map trie offset file in write mode");
            return false;
        }
        off_t*  pOffsets = reinterpret_cast<off_t *>(pBuff);
        // Dump main trie 
        if (!mTrie.Empty()) {
            DumpTrie(FDTrie, mTrie.GetRoot(), pOffsets);
        }

        // Dump auxilliary tries and mNodeInfo itself
        for (size_t i = 0; i < mNodeInfo.size(); ++i) {
            CAddNodeInfo&   NodeInfo = mNodeInfo[i];

            if (write(FDAddNodeInfo, &NodeInfo, sizeof NodeInfo) != sizeof NodeInfo) {
                Perror("Cannot write to add node info file");
                return false;
            }
            if (!NodeInfo.IsKetSubstInsEmpty()) {
                DumpTrie(FDTrie, NodeInfo.mKetSubstInsTrie, pOffsets);
            }
            if (!NodeInfo.IsKetDelEmpty()) {
                DumpTrie(FDTrie, NodeInfo.mKetDelTrie, pOffsets);
            }
        }

        // Dump combined ket tries

        for (size_t i = 0; i < mCombinedKetsInfo.size(); ++i) {
            off_t   CombKetOff = -1;

            if (mCombinedKetsInfo[i]) {
                CombKetOff = lseek(FDCombKets, 0, SEEK_CUR);

                if (-1 == CombKetOff) {
                    Perror("Cannot lseek in comb ket inf file");
                    return false;
                }
                const CAddCombinedKets&   CombKet = *mCombinedKetsInfo[i];
                SmallCounter              Qty = CombKet.GetQty();
                const COneKetPrime*       pKetPrime = CombKet.GetKetArr();        
                if (write(FDCombKets, &Qty, sizeof Qty) != sizeof Qty) {
                    Perror("Cannot write Qty to combined ket info file");
                    return false;
                }
                for (SmallCounter k = 0; k < Qty; ++k) {
                    if (write(FDCombKets, &pKetPrime[k], sizeof pKetPrime[k]) != sizeof pKetPrime[k]) {
                        Perror("Cannot write ket prime to combined ket info file");
                        return false;
                    }
                    if (pKetPrime[k].mKetPrime == NILL_NODE_INDEX) {
                        THROW_AND_PRINT_RUNTIME_ERROR("Internal error: nill KetPrime index");
                    }
                    DumpTrie(FDTrie, pKetPrime[k].mKetPrime, pOffsets);
                }
            }
            if (write(FDCombKetOffsets, &CombKetOff, sizeof CombKetOff) != sizeof CombKetOff) {
                Perror("Cannot write to comb ket offset file");
                return false;
            }
        }

        // Finally unmap pOffsets
        if (munmap(0, OffsetSize)) {
            Perror("Cannot unmap trie offset file");
            return false;
        }

    } catch (const exception& e) {
        cerr << "Cannot construct tree, exception: " << e.what() << endl;
        return false;
    } catch (...) {
        cerr << "Cannot construct tree, unknown exception" << endl;
        return false;
    }

    return true;
}

void
CErrataTree::DumpTrie(int fd, NodeIndex StartIndex, off_t *pOffsets)
{
    // Record current node offset
    pOffsets[StartIndex] = LseekThrowOnErr(fd, 0, SEEK_CUR);

    const NodeIndex*    pChildren;
    SmallCounter        LabelLen;
    SmallCounter        ChildQty;
    const char*         pLabel;

    mTrie.AccessNode(StartIndex, LabelLen, pLabel, ChildQty, pChildren);

    errno = 0;

    if (write(fd, &LabelLen, sizeof LabelLen) != sizeof LabelLen) {
        THROW_AND_PRINT_RUNTIME_ERROR("Cannot write label len: " << strerror(errno));
    }
    if (write(fd, pLabel, LabelLen) != LabelLen) {
        THROW_AND_PRINT_RUNTIME_ERROR("Cannot write label len: " << strerror(errno));
    }
    if (write(fd, &ChildQty, sizeof ChildQty) != sizeof ChildQty) {
        THROW_AND_PRINT_RUNTIME_ERROR("Cannot write child qty: " << strerror(errno));
    }
    if (ChildQty) {
        size_t ArrSize = ChildQty * sizeof pChildren[0];

        if (write(fd, &pChildren[0], ArrSize) != ArrSize) {
            THROW_AND_PRINT_RUNTIME_ERROR("Cannot write dummy child offsets: " << strerror(errno));
        }
    }

    for (unsigned i = 0; i < ChildQty; ++i) {
        DumpTrie(fd, pChildren[i], pOffsets);
    }
}

void
CErrataTree::SearchErrataTreesStartWith(SmallCounter Level, 
                                NodeIndex* pVisitedNodes, 
                                SmallCounter StartCentroidLevel, 
                                NodeIndex StartIndex, SmallCounter LabelStartPoint,
                                char *pPrefix,  SmallCounter PrefixStartPoint, SmallCounter* pPrefixStart,
                                char* pPattern, SmallCounter PatternStartPoint, SmallCounter* pPatternStart,
                                int kerr, vs* pvResList) const
{
    if (Level >= MAX_SMALL_COUNTER) {
        THROW_AND_PRINT_RUNTIME_ERROR("Internal error: level is too deep");
    }
    const NodeIndex*    pChildren;
    SmallCounter        LabelLen;
    SmallCounter        ChildQty;
    const char*         pLabel;
    const CAddNodeInfo&       NodeInfo = mNodeInfo[StartIndex];

    mTrie.AccessNode(StartIndex, LabelLen, pLabel, ChildQty, pChildren);

    if (LabelStartPoint > LabelLen) {
        THROW_AND_PRINT_RUNTIME_ERROR("Internal error: LabelStartPoint = " << int(LabelStartPoint) << " is greater than label len = " << int(LabelLen));
    }

    if (!LabelStartPoint) {
        pVisitedNodes[Level] = StartIndex;
        /* The following two are used only for searching in KET' */
        pPatternStart[Level] = PatternStartPoint + LabelLen;
        pPrefixStart[Level]  = PrefixStartPoint + LabelLen;
    }

    /*
     *  LabelStartPoint reaches the point one character after the end of the label,
     *  or it points to the label character, which is not equal to the respective character of the pattern.
     *  PS: If, at some point we reach the end of the pattern, then the following holds:
     *     pPattern[PatternStartPoint] == 0 and pLabel[LabelStartPoint] != 0
     *     Therefore, the loop will terminate.
    */
    for (;LabelStartPoint < LabelLen && WildCardEqual(pLabel[LabelStartPoint],pPattern[PatternStartPoint]); 
            ++PatternStartPoint,
            ++PrefixStartPoint,
            ++LabelStartPoint){ 
        pPrefix[PrefixStartPoint] = pLabel[LabelStartPoint];
    }

    if (!kerr) { /* 
                  *  If we search with zero errors in a KET or a KET' trees, just follow the path.
                  *  Note that a corresponding KET does not undergo a centroid-path decomposition,
                  *  it is just a regular trie
                  */
        if (LabelStartPoint < LabelLen) {
            return; // No match
        }

        if (!ChildQty) {
            if (!pPattern[PatternStartPoint]) { // i.e. a match
                if (pvResList) {
                    pPrefix[PrefixStartPoint] = 0;
                    pvResList->push_back(pPrefix);
                }
            }
        } else {
            for (int i = 0; i < ChildQty; ++i) {
                const NodeIndex*      pChildrenChild;
                SmallCounter          LabelLenChild;
                SmallCounter          ChildQtyChild;
                const char*           pLabelChild;

                mTrie.AccessNode(pChildren[i], LabelLenChild, pLabelChild, ChildQtyChild, pChildrenChild);

                /* 
                 * Note the following:
                 * 1) wildcard would not match a zero-code character, i.e., the end of the string.
                 * 2) The first character of the pattern matches at most one first character of a child label.
                 */

                if (LabelLenChild && WildCardEqual(pLabelChild[0], pPattern[PatternStartPoint]) ||
                    !LabelLenChild && !pPattern[PatternStartPoint]) {
                    SearchErrataTreesStartWith(Level + 1, pVisitedNodes, StartCentroidLevel,
                                            pChildren[i], 0, 
                                            pPrefix, PrefixStartPoint, pPrefixStart,
                                            pPattern, PatternStartPoint, pPatternStart, 
                                            0, pvResList);
  
                    return;
                }
            }
   
        }

        return;
    }

    if (ChildQty && !NodeInfo.mHasCentroid) {
        THROW_AND_PRINT_RUNTIME_ERROR("Internal error: node " << StartIndex << " ChildQty = " << int(ChildQty) << " is in a trie that did not undergo a centroid path decomposition");
    }

    assert(sizeof(int) > sizeof(SmallCounter)); // Otherwise, we have problem with signed numbers
    int  KetPrimeStartLevel = 0;
    bool bSearchPrevious = false; // If this flag is set, then we search all k-errata' subtrees hanging from the current 
    // centroid path on the way to the current node.

    if (LabelStartPoint < LabelLen) {
        assert(pLabel[LabelStartPoint] != pPattern[PatternStartPoint]);
        bSearchPrevious = true;
        KetPrimeStartLevel = Level - int(!ChildQty);

        // If kerr > 0, three cases are to be considered: insertion (in the text), deletion (from the text), substitution. 
        // For transposition-aware searching, a possible transposition should also be considered.
        
        if (kerr > 0) {
            // 1. Insertion (into the text).
            if (pPattern[PatternStartPoint]) { // Make sure we have not reached the end of the pattern yet
                // Skip filling pPrefix (because there is no text character that corresponds to the prefix)

                SearchErrataTreesStartWith(Level, pVisitedNodes, Level, /* StartCentroidLevel changed */
                                            StartIndex, LabelStartPoint, 
                                            pPrefix,  PrefixStartPoint,  pPrefixStart,
                                            pPattern, PatternStartPoint + 1, pPatternStart, 
                                            kerr - 1, pvResList);
            }

            // 2. Deletion (from the text).

            pPrefix[PrefixStartPoint] = pLabel[LabelStartPoint];

            SearchErrataTreesStartWith(Level, pVisitedNodes, Level, /* StartCentroidLevel changed */
                                            StartIndex, LabelStartPoint   + 1, 
                                            pPrefix,    PrefixStartPoint  + 1, pPrefixStart,
                                            pPattern, PatternStartPoint, pPatternStart, 
                                            kerr - 1, pvResList);

            // 3. Substitution
            if (pPattern[PatternStartPoint]) { // Make sure we have not reached the end of the pattern yet
                pPrefix[PrefixStartPoint] = pLabel[LabelStartPoint];

                SearchErrataTreesStartWith(Level, pVisitedNodes, Level, /* StartCentroidLevel changed */
                                            StartIndex, LabelStartPoint   + 1, 
                                            pPrefix,    PrefixStartPoint  + 1,  pPrefixStart,
                                            pPattern,   PatternStartPoint + 1, pPatternStart, 
                                            kerr - 1, pvResList);
            }
            // 4. Transposition
            if (mHasTransp && 
                pPattern[PatternStartPoint] && pPattern[PatternStartPoint + 1] && 
                pPattern[PatternStartPoint + 1] == pLabel[LabelStartPoint]) {
                
                swap(pPattern[PatternStartPoint], pPattern[PatternStartPoint + 1]);

                SearchErrataTreesStartWith(Level, pVisitedNodes, Level, /* StartCentroidLevel changed */
                                            StartIndex, LabelStartPoint   + 1, 
                                            pPrefix,    PrefixStartPoint  + 1,  pPrefixStart,
                                            pPattern,   PatternStartPoint + 1, pPatternStart, 
                                            kerr - 1, pvResList);

                swap(pPattern[PatternStartPoint], pPattern[PatternStartPoint + 1]);
            } 
        }
    } else {
        /* 
         * At this point, we try to move along one of the child labels.
         * 1. If none matches the next pattern character: pPattern[PatternStartPoint], then we search in the k-errata subtree,
         *    which is attached to the current node. In this case, we set bSearchPrevious = true;
         * 2. If the match is achieved, two sub-cases are possible:
         *    2.1 This is a match along centroid label. Just continue searching recursively with the same value of StartCentroidLevel.
         *    2.2 The label is not centroid. Then, search recursively, but change StartCentroidLevel to be equal to Level + 1.
         * 3. We have reached a terminal node with no children. Set bSearchPrevious = true.
         *    If pPattern[PatternStartPoint] == 0, report a match.
         */

         /*
          * Note that additionally we might also need to consider deletion, insertion, substitution and transposition along 
          * the centroid path:
          */
        bool bCheckCentroidEdits = false;

        if (ChildQty) {
            SmallCounter MatchingPath = 0;
            bool         bHasMatch = false;
            for (int i = 0; i < ChildQty; ++i) {
                const NodeIndex*      pChildrenChild;
                SmallCounter          LabelLenChild;
                SmallCounter          ChildQtyChild;
                const char*           pLabelChild;

                mTrie.AccessNode(pChildren[i], LabelLenChild, pLabelChild, ChildQtyChild, pChildrenChild);

                /* 
                 * Note the following:
                 * 1) wildcard would not match a zero-code character, i.e., the end of the string.
                 * 2) The first character of the pattern matches at most one first character of a child label.
                 */
                if (LabelLenChild && WildCardEqual(pLabelChild[0], pPattern[PatternStartPoint]) ||
                    !LabelLenChild && !pPattern[PatternStartPoint]) {
                    bHasMatch = true;
                    MatchingPath = i;
                    break;
                }
            }
            if (!bHasMatch) {
                bSearchPrevious = true;
                KetPrimeStartLevel = Level - 1; /* 
                                                 * Don't search current level KET', because it is searched via
                                                 * KET and considering possible edits along the centroid path.
                                                 */

                /* 
                 * If no label matches a prefix, trie k-errata trees (insertion/substitution and deletion)
                 */

                if (kerr > 0) {
                    NodeIndex    aVisitedKETNodes[MAX_SMALL_COUNTER];
                    SmallCounter aPatternStartKET[MAX_SMALL_COUNTER];
                    SmallCounter aPrefixStartKET [MAX_SMALL_COUNTER];

                    if (!NodeInfo.IsKetSubstInsEmpty()) {

                        SearchErrataTreesStartWith(0, aVisitedKETNodes, 0, 
                                          NodeInfo.mKetSubstInsTrie, 0, /* Starting from the beginning of the label */
                                          pPrefix, PrefixStartPoint, aPrefixStartKET,
                                          pPattern, PatternStartPoint, aPatternStartKET, 
                                          kerr - 1, pvResList);
                    }

                    if (!NodeInfo.IsKetDelEmpty()) {
                        SearchErrataTreesStartWith(0, aVisitedKETNodes, 0, 
                                          NodeInfo.mKetDelTrie, 0, /* Starting from the beginning of the label */
                                          pPrefix, PrefixStartPoint, aPrefixStartKET,
                                          pPattern, PatternStartPoint, aPatternStartKET, 
                                          kerr - 1, pvResList);
                    }
                }

                /* 
                 * KET trie does not cover a centroid path.
                 * Therefore, if kerr > 0, we should also try to advance along the main path.
                 * Three cases are to be considered: insertion (in the text), deletion (from the text), substitution. Transposition is TODO/TBD.
                 */
                if (kerr > 0) {
                    bCheckCentroidEdits = true;
                }


            } else {
                if (MatchingPath == NodeInfo.mCentroid) {
                    SearchErrataTreesStartWith(Level + 1, pVisitedNodes, StartCentroidLevel, /* StartCentroidLevel HAS NOT changed */
                                            pChildren[MatchingPath], 0, /* Starting from the beginning of the label */
                                            pPrefix, PrefixStartPoint, pPrefixStart,
                                            pPattern, PatternStartPoint, pPatternStart, 
                                            kerr /* The same kerr, no mismatche occurred */, pvResList);
                } else {
                    bSearchPrevious = true;
                    bCheckCentroidEdits = true; // Also consider edits along the centroid path

                    KetPrimeStartLevel = Level - 1; // Don't include this level KET', we are going to search it using the following recursive call:

                    SearchErrataTreesStartWith(Level + 1, pVisitedNodes, Level + 1, /* StartCentroidLevel changed */
                                            pChildren[MatchingPath], 0, /* Starting from the beginning of the label */
                                            pPrefix, PrefixStartPoint, pPrefixStart,
                                            pPattern, PatternStartPoint, pPatternStart, 
                                            kerr /* The same kerr, no mismatche occurred */, pvResList);
                }
            }

            if (bCheckCentroidEdits) {
                const NodeIndex*      pChildrenChild;
                SmallCounter          LabelLenChild;
                SmallCounter          ChildQtyChild;
                const char*           pLabelChild;

                NodeIndex             NewStartIndex = pChildren[NodeInfo.mCentroid];

                mTrie.AccessNode(NewStartIndex, LabelLenChild, pLabelChild, ChildQtyChild, pChildrenChild);

                // 1. Insertion (into the text).
                if (pPattern[PatternStartPoint]) { // Make sure we have not reached the end of the pattern yet
                    // Skip filling pPrefix (because there is no text character that corresponds to the prefix)

                    SearchErrataTreesStartWith(Level + 1, pVisitedNodes, Level + 1, /* StartCentroidLevel changed */
                                                NewStartIndex,              0, 
                                                pPrefix,    PrefixStartPoint,  pPrefixStart,
                                                pPattern, PatternStartPoint + 1, pPatternStart, 
                                                kerr - 1, pvResList);
                }

                if (LabelLenChild) { // Deletion (from the text) and substitution (unlike insertion) are not possible for an empty label.

                    // 2. Deletion (from the text).
                    pPrefix[PrefixStartPoint] = pLabelChild[0];

                    SearchErrataTreesStartWith(Level + 1, pVisitedNodes, Level + 1, /* StartCentroidLevel changed */
                                                    NewStartIndex,               1, 
                                                    pPrefix,  PrefixStartPoint + 1, pPrefixStart,
                                                    pPattern, PatternStartPoint, pPatternStart, 
                                                    kerr - 1, pvResList);
                    // 3. Substitution
                   
                    pPrefix[PrefixStartPoint] = pLabelChild[0];
                    if (pPattern[PatternStartPoint]) { // Make sure we have not reached the end of the pattern yet
                        SearchErrataTreesStartWith(Level + 1, pVisitedNodes, Level + 1, /* StartCentroidLevel changed */
                                                    NewStartIndex,                1, 
                                                    pPrefix,  PrefixStartPoint  + 1, pPrefixStart,
                                                    pPattern, PatternStartPoint + 1, pPatternStart, 
                                                    kerr - 1, pvResList);
                    }
                }
            }
        } else {
            // 3. Terminal node with no children
            //    Does the rest of the pattern has length <= kerr
            if (!pPattern[PatternStartPoint] || strlen(pPattern + PatternStartPoint) <= kerr) {
                bSearchPrevious = true;
                KetPrimeStartLevel = Level - 1; // There is not KET' tree at this level, because there are no children

                if (pvResList) {
                    pPrefix[PrefixStartPoint] = 0;
                    pvResList->push_back(pPrefix);
                }
            }
        }
    }

    // If kerr > 0, and ChildQty > 1, then try searching k-errata tries hanging from the centroid path
    if (!StartCentroidLevel && !mCombinedKetsInfo[pVisitedNodes[0]]) {
        ++StartCentroidLevel; // Some root nodes (which have only one child) do not have KET' info
    }

    if (kerr > 0 && bSearchPrevious && Level >= StartCentroidLevel) {
        // Choose, which KET' trees, should be searched.
        int CurrLevel = KetPrimeStartLevel;

        while (CurrLevel >= (int)StartCentroidLevel) {
            const CAddCombinedKets* pCombKetsInfo = mCombinedKetsInfo[pVisitedNodes[CurrLevel]];

            if (!pCombKetsInfo) {
                THROW_AND_PRINT_RUNTIME_ERROR("Internal error: no combined k-errata tree for NodeIndex = " << pVisitedNodes[CurrLevel] << ", Level = " << int(CurrLevel) << " Child qty: " << int(ChildQty));
            }

            /* 
             * Find a combined KET that has the largest span and covers most levels
             * Note that the first KET covers exactly one level. Therefore, for i == 0
             * pKetArra[i]->mStartLevel == CurLevel and pKetArra[i]->mStartLevel >= StartCentroidLevel.
             * Therefore, we always execute the code inside if () in the loop below.
             */
            const COneKetPrime* pKetArr = pCombKetsInfo->GetKetArr();
            SmallCounter        i = pCombKetsInfo->GetQty();
            char                pKETPrimePrefix[MAX_SMALL_COUNTER];

            NodeIndex    aVisitedKETPrimeNodes[MAX_SMALL_COUNTER];
            SmallCounter aPatternStartKETPrime[MAX_SMALL_COUNTER];
            SmallCounter aPrefixStartKETPrime [MAX_SMALL_COUNTER];

            memcpy(pKETPrimePrefix, pPrefix, PrefixStartPoint + 1);

            /* 
             *  Note that here we use ever shortening pieces of pKETPrimePrefix.
             *  Therefore, we don't have to create a separate copy of pPrefix for each recursive call,
             *  because the prefix remains unchanged. We, however, cannot use pPrefix here,
             *  because its elements with indices from 0 .. pPatternStart[Level] - 1 may be modified.
             *  This elements, however, may be used for other recursive calls (when we exit this function
             *  and return to an upper level).
             */

            for (;i; --i) {
                SmallCounter NewStartLevel = pKetArr[i - 1].mStartLevel; 

                if (NewStartLevel >= StartCentroidLevel) {
                    SearchErrataTreesStartWith(0, aVisitedKETPrimeNodes, 0, 
                                          pKetArr[i -1].mKetPrime, 0, /* Starting from the beginning of the label */
                                          pKETPrimePrefix,  pPrefixStart[NewStartLevel], aPrefixStartKETPrime,
                                          pPattern, pPatternStart[NewStartLevel], aPatternStartKETPrime, 
                                          kerr - 1, pvResList);
            
                    CurrLevel = NewStartLevel - 1;
                    break; 
                }
            }
            if (!i) {
                THROW_AND_PRINT_RUNTIME_ERROR("Internal error: zero index reached for NodeIndex = " << pVisitedNodes[CurrLevel] << ", Level = " << int(CurrLevel) << " StartCentroidLevel = " << int(StartCentroidLevel));
            }
        }
    }
}

void
CErrataTree::CreateErrataTrees(TTrie trie, int kerr, bool bHasTransp)
{
    if (trie.Empty()) {
        return;
    }

    ResizeAddInfo();

    try {
        // Perform a centroid path decomposition, but skip the very first root node
        NodeIndex RootIndex = trie.GetRoot();

        FillChildQty(RootIndex);

        RecordCentroidPaths(RootIndex);

        NodeIndex    aVisitedNodes[MAX_SMALL_COUNTER];

        CreateErrataTreesStartWith(0, aVisitedNodes, 0, RootIndex, kerr, bHasTransp, kerr < 2);
    } catch (const runtime_error &err) {
        cerr << err.what() << endl;
        throw;
    }
}

void
CErrataTree::CreateErrataTreesStartWith(SmallCounter Level, NodeIndex* pVisitedNodes, SmallCounter StartCentroidLevel, 
                                        NodeIndex StartIndex, int kerr, bool bHasTransp, bool bReuseExistingNodes)
{
    if (StartIndex >= mNodeInfo.size()) {
        THROW_AND_PRINT_RUNTIME_ERROR(" Invalid index: " << StartIndex);
    }
    pVisitedNodes[Level] = StartIndex;

    NodeIndex*      pChildren;
    SmallCounter    LabelLen;
    SmallCounter    ChildQty;
    const char*     pLabel;

    mTrie.AccessNode(StartIndex, LabelLen, pLabel, ChildQty, pChildren);

    /*
     * Whenever, we reach a child, we create groups of KET' trees along the current centroid path,
     * if the centroid path is long enough.
     */
    if (!ChildQty) {
        if (!StartCentroidLevel && !mCombinedKetsInfo[pVisitedNodes[StartCentroidLevel]]) {
            ++StartCentroidLevel; // The root node may have only one child. Then, it has no combined trees
        }

        if (Level > StartCentroidLevel + 1) {
            SmallCounter    LevelQty = Level - StartCentroidLevel - 1;
            NodeIndex       vLeafQty[LevelQty + 1];

            vLeafQty[0] = 0;
            for (SmallCounter i = 1, lev = StartCentroidLevel; i <= LevelQty; ++i, ++lev) {
                vLeafQty[i] = vLeafQty[i - 1] + mNodeInfo[pVisitedNodes[lev]].mLeafQty;
            }
            // This procedure will recursively create KET' trees starting from the smallest ones
            DoKetPrimeForInterval(vLeafQty, pVisitedNodes, StartCentroidLevel, LevelQty, kerr, bHasTransp, bReuseExistingNodes);
        }
        return;
    }

    NodeIndex       MaxQty = 0; 
    CAddNodeInfo*   pParentNodeInfo = &mNodeInfo[StartIndex];
    SmallCounter    ParentNodeCentroid = pParentNodeInfo->mCentroid;

    if (!ChildQty) {
        // No children, exit
        return;
    }

    if (ChildQty < 2 && Level) {
        /* 
         *  Don't do anything with a terminal node.
         *  The root label should always has one child and starts with Level == 0
         */
        return;
    }


    if (ChildQty && Level >= MAX_SMALL_COUNTER) {
        THROW_AND_PRINT_RUNTIME_ERROR("Internal error: the pattern is too long, we reached the maximal allowed level, but we still have children");
    }

    if (ChildQty && !pParentNodeInfo->mHasCentroid) {
        THROW_AND_PRINT_RUNTIME_ERROR("Internal error, node " << StartIndex << " does not have a centroid, child qty: " << int(ChildQty));
    }

    assert(ChildQty);
    if (ChildQty > 1) {
        /* 
         * If ChildQty > 1, we have to create insertion, deletion, substitution, and transposition trees from off-path trees. (an off-path subtree is 
         * a tree, which is not a centroid path)
         *
         * These trees are of the following three types:
         *  1) Real insertion, substitution, and transposition subtrees with the wildcard: KetSubstInsTrie.
         *  2) Real deletion subtrees without the wildcard: KetDelTrie.
         *  3) Subtrees obtained from trees of the first kind by replacing the wildchard with the next character on the centroid path. These subtrees,
         *     are merged with real deletion subtrees of the second kind. Then, they are attached to the path that leads from the beginning of the current
         *     centroid path to the current node (KetTriePrime).
         *  
         */

        /*
         * CASE :1 & 2 Real insertion, deletion, and substitution subtrees. 
         */
        vector<NodeIndex>   vRootNodes; // Some of these nodes will be disposed later, after trees are merged

        TTrie         KetDelTrie(TTrie::CreateDeletionTrie(StartIndex, pParentNodeInfo->mCentroid));

        if (mHasTransp) {
            //cout << "=== ORIG ============" << endl;
            //PrintTrie(TTrie(StartIndex), cout);
            TTrie         KetTranspTrie(TTrie::CreateTranspositionTrie(StartIndex, pParentNodeInfo->mCentroid));
            //cout << "=== ORIG ============" << endl;
            //PrintTrie(TTrie(StartIndex), cout);

            KetTranspTrie.AddToArrayIfNotEmpty(vRootNodes);

            //cout << "=== TRANSP ============" << endl;
            //PrintTrie(KetTranspTrie, cout);
            //cout << "=== @@@@@@ ============" << endl;

            KetDelTrie = KetDelTrie.MergeTries(KetTranspTrie, true); // Always reuse nodes at this point, later we create copies if needed
        }

        TTrie         KetInsTrie(TTrie::CreateInsertionTrie(StartIndex, pParentNodeInfo->mCentroid));

        TTrie         KetSubstTrie(TTrie::CreateSubstitutionTrie(StartIndex, pParentNodeInfo->mCentroid));

        KetDelTrie.AddToArrayIfNotEmpty(vRootNodes);
        KetInsTrie.AddToArrayIfNotEmpty(vRootNodes);
        KetSubstTrie.AddToArrayIfNotEmpty(vRootNodes);

        TTrie         KetSubstInsTrie = KetSubstTrie.MergeTries(KetInsTrie, true);  // Always reuse nodes at this point, later we create copies if needed

        KetSubstInsTrie.AddToArrayIfNotEmpty(vRootNodes);

        TTrie         KetPrimeTrie;
        

        /* 
         *  CASE 3:
         *  Now let's create KetPrimeTrie, by replacing a wildcard with next character along the centroid path.
         *  If the centroid path child is terminal, don't create a single KetPrimeTrie.
         */
        NodeIndex*      pCentroidChildren;
        SmallCounter    CentroidLabelLen;
        SmallCounter    CentroidChildQty;
        const char*     pCentroidLabel;

        mTrie.AccessNode(pChildren[pParentNodeInfo->mCentroid], CentroidLabelLen, pCentroidLabel, CentroidChildQty, pCentroidChildren);

        if (CentroidLabelLen && !KetSubstInsTrie.Empty()) {
            NodeIndex*      pChildrenKetRoot;
            SmallCounter    LabelLenKetRoot;
            SmallCounter    ChildQtyKetRoot;
            const char*     pLabelKetRoot;

            char NextChar = pCentroidLabel[0];

            KetSubstInsTrie.AccessNode(KetSubstInsTrie.GetRoot(), LabelLenKetRoot, pLabelKetRoot, ChildQtyKetRoot, pChildrenKetRoot);

            if (LabelLenKetRoot) {
                PrintTrie(KetSubstInsTrie); // Prints to cerr
                THROW_AND_PRINT_RUNTIME_ERROR("Invalid SubstInstTrie is constructed (checking step 1): child qty: " << int(ChildQtyKetRoot) << " LabelLen: " << int(LabelLenKetRoot) << " Lefel " << int(Level));
            }

            for (SmallCounter j = 0; j < ChildQtyKetRoot; ++j) {
                NodeIndex*      pChildrenKet;
                SmallCounter    LabelLenKet;
                SmallCounter    ChildQtyKet;
                const char*     pLabelKet;

                KetSubstInsTrie.AccessNode(pChildrenKetRoot[j], LabelLenKet, pLabelKet, ChildQtyKet, pChildrenKet);

                if (!LabelLenKet || pLabelKet[j] != WILDCHARD_CHAR) {
                    cerr << "KetSubstInsTrie: =========\n";
                    PrintTrie(KetSubstInsTrie); // Prints to cerr
                    cerr << "KetSubstTrie: =========\n";
                    PrintTrie(KetSubstTrie); // Prints to cerr
                    cerr << "KetInsTrie: =========\n";
                    PrintTrie(KetInsTrie); // Prints to cerr
                    cerr << "=====================\n";

                    THROW_AND_PRINT_RUNTIME_ERROR("Invalid SubstInstTrie is constructed (checking step 2), child " << int(j) << " LabelLen: " << int(LabelLenKet) << " First label char: " << (LabelLenKet ? string(1, pLabelKet[0]):string("<empty label>")) << " Lefel " << int(Level));
                }

                char pModLabelKet[LabelLenKet];
            
                memcpy(pModLabelKet, pLabelKet, LabelLenKet);
                pModLabelKet[0] = NextChar;

                NodeIndex Index = CNodeAllocator::AllocNode(LabelLenKet, pModLabelKet, ChildQtyKet, pChildrenKet);

                NodeIndex KetPrimeTrieIndex = TTrie::CreateTrieWithOneChild(Index);
                KetPrimeTrie = KetPrimeTrie.MergeTries(KetPrimeTrieIndex, true); // at this point always reuse existing nodes.
                KetPrimeTrie.AddToArrayIfNotEmpty(vRootNodes);
            }
        }

        KetPrimeTrie.AddToArrayIfNotEmpty(vRootNodes);
        KetPrimeTrie = KetPrimeTrie.MergeTries(KetDelTrie, true); // at this point always reuse existing nodes.
        KetPrimeTrie.AddToArrayIfNotEmpty(vRootNodes);


        if (!bReuseExistingNodes) {
            KetDelTrie = KetDelTrie.CloneTrie();
            KetSubstInsTrie = KetSubstInsTrie.CloneTrie();

            if (!KetPrimeTrie.Empty()) {
                KetPrimeTrie = KetPrimeTrie.CloneTrie();
            }
        }

        if (KetPrimeTrie.Empty()) {
            cerr << "KetPrimeTrie is erroneously empty" << endl;
            cerr << "=== Main Trie === " << StartIndex << endl;;
            PrintTrie(TTrie(StartIndex), cerr);
            cerr << "=== KET === \n";
            PrintTrie(KetDelTrie, cerr);
            cerr << "==== Subst \n";
            PrintTrie(KetSubstTrie, cerr);
            cerr << "==== Ins \n";
            PrintTrie(KetInsTrie, cerr);
            cerr << "==== SubstIns \n";
            PrintTrie(KetSubstInsTrie, cerr);
            cerr << "==== KET' \n";
            PrintTrie(KetPrimeTrie, cerr);
            cerr << "==========\n";
        }

        COneKetPrime    OnePrimeKet(KetPrimeTrie.GetRoot(), Level);

        mCombinedKetsInfo[StartIndex]       = KET_PRIME_ALLOCATOR::AllocKets(1, &OnePrimeKet);

        // Deletion trie may be empty, subst/ins trie - never!
        if (!KetDelTrie.Empty()) {
            pParentNodeInfo->mKetDelTrie          = KetDelTrie.GetRoot();
        }
        pParentNodeInfo->mKetSubstInsTrie     = KetSubstInsTrie.GetRoot();

        for (size_t i = 0; i < vRootNodes.size(); ++i) {
            if (vRootNodes[i] != pParentNodeInfo->mKetDelTrie && 
                vRootNodes[i] != pParentNodeInfo->mKetSubstInsTrie && 
                vRootNodes[i] != KetPrimeTrie.GetRoot()) {

                CNodeAllocator::FreeNode(vRootNodes[i]);
            }
        }

        /* 
         *  Recursively create errata trees.
         */

        if (kerr > 1) {
            CreateErrataTrees(KetSubstInsTrie, kerr - 1, bHasTransp);
            CreateErrataTrees(KetDelTrie, kerr - 1, bHasTransp);
            CreateErrataTrees(KetPrimeTrie, kerr - 1, bHasTransp);
        }
    }
    /* 
     *  Don't use pointer pParentNodeInfo after recursive calls, the may reallocate mNodeInfo.
     *  Then the pointer may become invalid.
     */
    pParentNodeInfo = NULL; 

    /* Recursive creation of subtrees */
    for (SmallCounter i = 0; i < ChildQty; ++i) {
        CreateErrataTreesStartWith(Level + 1, pVisitedNodes,
                                   i == ParentNodeCentroid ? StartCentroidLevel  : Level + 1,
                                   pChildren[i], kerr, bHasTransp, bReuseExistingNodes);
    }
}

void
CErrataTree::DoKetPrimeForInterval(NodeIndex* pLeafQty, NodeIndex* pVisitedNodes, SmallCounter StartCentroidLevel, SmallCounter LevelQty, 
                                   int kerr, bool bHasTransp, bool bReuseExistingNodes)
{
    if (LevelQty < 2) {
        return;
    } 
    // Do smaller trees first
    if (LevelQty > 2) {
        NodeIndex TotalQty = pLeafQty[LevelQty] - pLeafQty[0];
        NodeIndex HalfQty   = TotalQty / 2;

        // Find a place to divide the interval
        NodeIndex* p = lower_bound(pLeafQty + 1, pLeafQty + LevelQty + 1, HalfQty);

        SmallCounter Qty1 = p - pLeafQty;
        SmallCounter Qty2 = pLeafQty + LevelQty - p;
        if (Qty1 >= 2) {
            DoKetPrimeForInterval(pLeafQty, pVisitedNodes, StartCentroidLevel, Qty1, kerr, bHasTransp, bReuseExistingNodes);
        }
        if (Qty2 >= 2) {
            DoKetPrimeForInterval(pLeafQty + Qty1, pVisitedNodes, StartCentroidLevel + Qty1, Qty2, kerr, bHasTransp, bReuseExistingNodes);
        }
    }
    // Merge all nodes
    TTrie               KetPrimeComb;
    vector<NodeIndex>   vRootNodes;

    for (SmallCounter i = 0; i < LevelQty; ++i) {
        string                  Label;

        /* 
         *  1) Concatenate i labels leading from the top node at level StartCentroidLevel 
         *  to the node pVisitedNodes[StartCentroidLevel + LevelQty - 1].
         *  2) Then, attach this label to a KET' trie of node pVisitedNodes[StartCentroidLevel + i]
         *  3) Finally, all constructed trees are merged together and assigned to the node pVisitedNodes[StartCentroidLevel + LevelQty - 1]
         */
        mTrie.ConcatLabels(pVisitedNodes  + StartCentroidLevel, i, Label); 

        NodeIndex               Index = pVisitedNodes[StartCentroidLevel + i];
        CAddCombinedKets*       pCombKetsInfo = mCombinedKetsInfo[Index];
        const COneKetPrime*     pKetArr = pCombKetsInfo ? pCombKetsInfo->GetKetArr() : NULL;

        if (!pCombKetsInfo || !pKetArr) {
            THROW_AND_PRINT_RUNTIME_ERROR("Internal error: cannot find KET' for index: " << Index);
        }
        TTrie                   KetPrimeExt = TTrie(pKetArr[0].mKetPrime).AppendTrieToEdge(Label, bReuseExistingNodes);
#if 0
        cout <<" ------------\n";
        PrintTrie(KetPrimeExt, cout);
        cout <<" ------------\n";
#endif
        KetPrimeComb = KetPrimeComb.MergeTries(KetPrimeExt, bReuseExistingNodes);

        KetPrimeComb.AddToArrayIfNotEmpty(vRootNodes);

    }

    if (KetPrimeComb.Empty()) {
        THROW_AND_PRINT_RUNTIME_ERROR("Empty Ket combined, StartCentroidLevel = " << int(StartCentroidLevel) << " LevelQty: " << LevelQty << " Start node: " << pVisitedNodes[StartCentroidLevel]);
    }

    for (size_t i = 0; i < vRootNodes.size(); ++i) {
        if (vRootNodes[i] != KetPrimeComb.GetRoot()) {
            CNodeAllocator::FreeNode(vRootNodes[i]);
        }
    }

    if (!bReuseExistingNodes) {
        KetPrimeComb = KetPrimeComb.CloneTrie();
    }

    // Assign KetPrimeComb to the node pVisitedNodes[StartCentroidLevel + LevelQty - 1]
#if 0
    cout << "=== KET' combined " << KetPrimeComb.GetRoot() << "=== kerr = " << kerr << " ======\n";
    PrintTrie(KetPrimeComb, cout);
#endif

    // Recursively create errata trees.
    if (kerr > 1) {
        CreateErrataTrees(KetPrimeComb, kerr - 1, bHasTransp);
    }

    NodeIndex           AttachIndex = pVisitedNodes[StartCentroidLevel + LevelQty - 1];
    CAddCombinedKets*   pOldKets = mCombinedKetsInfo[AttachIndex];
    SmallCounter        OldQty = pOldKets->GetQty();
    assert(pOldKets && OldQty);

    COneKetPrime        AddKet(KetPrimeComb.GetRoot(), StartCentroidLevel);
    // pOldKets is disposed by AddKetReallocMem
    CAddCombinedKets*   pNewKets = KET_PRIME_ALLOCATOR::AddKetReallocMem(pOldKets, AddKet);
    mCombinedKetsInfo[AttachIndex] = pNewKets;
}

void
CErrataTree::RecordCentroidPaths(NodeIndex StartIndex)
{
    if (StartIndex >= mNodeInfo.size()) {
        THROW_AND_PRINT_RUNTIME_ERROR(" Invalid index: " << StartIndex);
    }
   
    NodeIndex*      pChildren;
    SmallCounter    LabelLen;
    SmallCounter    ChildQty;
    const char*     pLabel;

    mTrie.AccessNode(StartIndex, LabelLen, pLabel, ChildQty, pChildren);

    NodeIndex       MaxQty = 0; 
    CAddNodeInfo&   ParentNodeInfo = mNodeInfo[StartIndex];

    for (SmallCounter i = 0; i < ChildQty; ++i) {
        NodeIndex       Index = pChildren[i];
        CAddNodeInfo&   ChildNodeInfo = mNodeInfo[Index];

        if (ChildNodeInfo.mLeafQty > MaxQty) {
            ParentNodeInfo.mCentroid = i;
            ParentNodeInfo.mHasCentroid = true;
            MaxQty = ChildNodeInfo.mLeafQty;
        }
        RecordCentroidPaths(Index); // Do a recursive decomposition
    }
}

void
CErrataTree::FillChildQty(NodeIndex StartIndex)
{
    if (StartIndex >= mNodeInfo.size()) {
        THROW_AND_PRINT_RUNTIME_ERROR(" Invalid index: " << StartIndex);
    }

    NodeIndex qty = mNodeInfo[StartIndex].mLeafQty;

    if (!qty) {
        NodeIndex*     pChildren;
        SmallCounter    LabelLen;
        SmallCounter    ChildQty;
        const char*     pLabel;

        mTrie.AccessNode(StartIndex, LabelLen, pLabel, ChildQty, pChildren);

        for (SmallCounter i = 0; i < ChildQty; ++i) {
            NodeIndex Index = pChildren[i];

            FillChildQty(Index);
            qty += mNodeInfo[Index].mLeafQty;
        }

        if (!ChildQty) {
            qty = 1;
        }
    }

    mNodeInfo[StartIndex].mLeafQty = qty;
}
