#include <string.h>

#include <assert.h>
#include <vector>
#include <stdexcept>

#include "nodealloc.h"

std::vector<CNode*>     CNodeAllocator::mpNodePtr;
size_t                  CNodeAllocator::mMemUsed;
std::vector<NodeIndex>  CNodeAllocator::mFreeIndices;

CNode*
CNodeAllocator::AllocMemSimple(size_t NodeSize)
{
    CNode* p = reinterpret_cast<CNode *>(new char[NodeSize]);
    mMemUsed += NodeSize;
    return p;
}

CNode*
CNodeAllocator::AllocMem(size_t NodeSize, NodeIndex &Index)
{
    CNode* p = AllocMemSimple(NodeSize);

    if (mFreeIndices.empty()) {
        // We want to keep the array memory efficient. 
        // By default, vector optimizes the number of reallocation. 
        // Each time when reallocation is needed, it asks for the double space.
        // We, instead, just ask for additional RESERVE_THRESHOLD elements
        if (mpNodePtr.size() >= MAX_NODE_INDEX) {
            throw std::runtime_error("Maximal allowed index reached");
        }
        
        if (mpNodePtr.size() % RESERVE_THRESHOLD == 0) { 
            mpNodePtr. reserve(mpNodePtr.size() + RESERVE_THRESHOLD); 
        } 
        Index = static_cast<NodeIndex>(mpNodePtr.size());
        mpNodePtr.push_back(p);
    } else {
        // We want to recycle indices
        Index = mFreeIndices.back();
        mFreeIndices.pop_back();
        mpNodePtr[Index] = p;
    }

    return p;
}

void
CNodeAllocator::FreeMemSimple(CNode *pNode, size_t NodeSize)
{
    mMemUsed -=  CNode::NodeSize(pNode->mLabelLen, pNode->mChildQty);
    delete [] reinterpret_cast<char *>(pNode); 
}

void
CNodeAllocator::FreeMem(CNode *p, size_t NodeSize, NodeIndex Index)
{
    FreeMemSimple(p, NodeSize);
        
    // We want to keep the array memory efficient. 
    // By default, vector optimizes the number of reallocation. 
    // Each time when reallocation is needed, it asks for the double space.
    // We, instead, just ask for additional RESERVE_THRESHOLD elements
    if (mFreeIndices.size() % RESERVE_THRESHOLD == 0) { 
        mFreeIndices. reserve(mFreeIndices.size() + RESERVE_THRESHOLD); 
    } 
    mFreeIndices.push_back(Index); // we want to utilize indices
}
    

void 
CNodeAllocator::FreeNode(NodeIndex Index) 
{ 
    CNode*  pNode = mpNodePtr[Index];
    if (!pNode) {
        return;
    }
    FreeMem(pNode, CNode::NodeSize(pNode->mLabelLen, pNode->mChildQty), Index);
    mpNodePtr[Index] = NULL;
}

NodeIndex
CNodeAllocator::AllocNode(SmallCounter LabelLen, const char *pLabel, SmallCounter ChildQty, NodeIndex *pChildren) {
    size_t NodeSize = CNode::NodeSize(LabelLen, ChildQty);

    NodeIndex   Index = 0;
    CNode* p = AllocMem(NodeSize, Index); 

    p->mLabelLen = LabelLen;
    p->mChildQty = ChildQty;
    if (pLabel) {
        memcpy(p->mpLabel, pLabel, LabelLen);
    }

    if (pChildren) {
        memcpy(p->GetChildData(), pChildren, sizeof(*pChildren) * ChildQty);
    }

    return Index;
}

// 1. Allocates memory 
//    Function uses the (Alloc|Free)MemSimple instead of (Alloc|Free)Mem, because
//    it does not want to modify node index.
// 2. Updates pointer in array mpNodePtr.
// 3. Copies content to the new location smartly: if a new node has shorter label or 
// smaller number of childern this is taken in account.
// 4. Changes a number of child nodes
void
CNodeAllocator::ReallocNode(NodeIndex Index, SmallCounter LabelLen, SmallCounter ChildQty) {
    assert(Index < mpNodePtr.size());
    CNode*          pOld = mpNodePtr[Index];
    unsigned int    OldSize = CNode::NodeSize(pOld->mLabelLen, pOld->mChildQty);
    CNode*          pNew = AllocMemSimple(CNode::NodeSize(LabelLen, ChildQty));

    // Smart copy
    pNew->mChildQty = ChildQty;
    pNew->mLabelLen = LabelLen;
    strncpy(pNew->mpLabel, pOld->mpLabel, std::min(pOld->mLabelLen, LabelLen));

    NodeIndex *pChildren = pNew->GetChildData();

    if (pChildren) {
        memcpy(pChildren, pOld->GetChildData(), sizeof(*pChildren) * std::min(pOld->mChildQty, ChildQty));
    }
    
    mpNodePtr[Index] = pNew;

    FreeMemSimple(pOld, OldSize);
}
