#ifndef TESTUTIL_H
#define TESTUTIL_H

#include <string>
#include <vector>
#include <iomanip>
#include <set>

#include "comptrie.h"
#include "addtrieinfo.h"

#define IF_EMPTY(v) ((v) == NILL_NODE_INDEX ? (long long)-1 : (long long)v)

void GenRandString(int MinLen, int MaxLen, std::string &s);
void GenRandStringList(int MinLen, int MaxLen, bool bUnique, size_t ListSize, std::vector<std::string>& Result);

struct PrintVisitor {
    PrintVisitor(std::ostream &out) : mOut(out) {}

    virtual void operator() (NodeIndex CurrIndex, unsigned int CurrLevel, SmallCounter LabelLen, const char* pLabel, SmallCounter ChildQty) {
        mOut << std::setw(2 * CurrLevel) << "#" << std::setw(4) << CurrIndex << " D" << std::setw(2) << CurrLevel << " '" << std::string(pLabel, LabelLen) << "' --> " << int(ChildQty) << std::endl;
    }
    std::ostream &mOut;
};

struct ExtendedPrintVisitor : public PrintVisitor {
    ExtendedPrintVisitor(const std::vector<CAddNodeInfo>& NodeInfo, const std::vector<CAddCombinedKets*>& CombinedKetsInfo, std::ostream &out) : 
                            PrintVisitor(out), mNodeInfo(NodeInfo), mCombinedKetsInfo(CombinedKetsInfo) {}

    virtual void operator() (NodeIndex CurrIndex, unsigned int CurrLevel, SmallCounter LabelLen, const char* pLabel, SmallCounter ChildQty) {
        const CAddNodeInfo& NodeInfo = mNodeInfo[CurrIndex];

        // There can be fewer mCombinedKesInfo elements, then mNodeInfo elements
        const CAddCombinedKets*     pCombKets = (CurrIndex < mCombinedKetsInfo.size()) ? mCombinedKetsInfo[CurrIndex]: NULL;
        const COneKetPrime*         pKetArr = NULL;
        if (pCombKets) {
            pKetArr = pCombKets->GetKetArr();

            for (SmallCounter i = 0; pKetArr && i < pCombKets->GetQty(); ++i) {
                mKETPrimes.push_back(pKetArr[i].mKetPrime);
            }
        }
        mOut << std::setw(2 * CurrLevel) << "#" << std::setw(4) << CurrIndex 
             << " D" << std::setw(2) << CurrLevel << " '" << std::string(pLabel, LabelLen) 
             << "' --> " << int(ChildQty) 
             << " Centr: " << (NodeInfo.mHasCentroid ? int(NodeInfo.mCentroid):-1) 
             << " KET subst/ins: " << IF_EMPTY(NodeInfo.mKetSubstInsTrie)
             << " KET del: " << IF_EMPTY(NodeInfo.mKetDelTrie)
             << " KET': ";

        if (pKetArr) {
            for (SmallCounter i = 0; i < pCombKets->GetQty(); ++i) {
                if (i) { mOut << ","; }
                mOut << pKetArr[i].mKetPrime;
            }
        }

        mOut << ""
             << " LQty " << NodeInfo.mLeafQty 
             << std::endl;
        if (!NodeInfo.IsKetDelEmpty()) {
            mKETs.push_back(NodeInfo.mKetDelTrie);
        }
        if (!NodeInfo.IsKetSubstInsEmpty()) {
            mKETs.push_back(NodeInfo.mKetSubstInsTrie);
        }
    }
    const std::vector<CAddNodeInfo>&        mNodeInfo;
    const std::vector<CAddCombinedKets*>    mCombinedKetsInfo;
    std::vector<NodeIndex>                  mKETs;
    std::vector<NodeIndex>                  mKETPrimes;
};

template <typename NodeAllocator> 
void PrintTrieStartNode(const CCompTrie<NodeAllocator>& trie, PrintVisitor& vis, NodeIndex Start, unsigned int Level = 0)
{
    const char*         pLabel; 
    SmallCounter        LabelLen;
    SmallCounter        ChildQty;
    const NodeIndex*    pChildren;

    trie.GetNodeInfo(Start, ChildQty, LabelLen, pLabel, pChildren);

    vis(Start, Level, LabelLen, pLabel, ChildQty); 

    for (SmallCounter i = 0; i < ChildQty; ++i) {
        PrintTrieStartNode(trie, vis, pChildren[i], Level + 1);
    }
}

template <typename NodeAllocator> 
void PrintTrie(const CCompTrie<NodeAllocator>& trie, std::ostream &out = std::cerr) {
    PrintVisitor    Prnt(out);

    if (trie.Empty()) {
        return;
    }
    PrintTrieStartNode(trie, Prnt, trie.GetRoot());
}

template <typename NodeAllocator> 
void GetTrieStringsStartNode(const CCompTrie<NodeAllocator>& trie, std::vector<std::string> &res, NodeIndex Start, std::string Parent) {
    const char*         pLabel; 
    SmallCounter        LabelLen;
    SmallCounter        ChildQty;
    const NodeIndex*    pChildren;

    trie.GetNodeInfo(Start, ChildQty, LabelLen, pLabel, pChildren);

    if (LabelLen) {
        Parent.append(pLabel, LabelLen);
    }

    if (!ChildQty) {
        res.push_back(Parent);
    }

    for (SmallCounter i = 0; i < ChildQty; ++i) {
        GetTrieStringsStartNode(trie, res, pChildren[i], Parent);
    }
}

template <typename NodeAllocator> 
void GetTrieStrings(const CCompTrie<NodeAllocator>& trie, std::vector<std::string> &res) {
    if (trie.Empty()) {
        return;
    }
    GetTrieStringsStartNode(trie, res, trie.GetRoot(), "");
}

template <typename NodeAllocator> 
void GetTrieNodesStartNode(const CCompTrie<NodeAllocator>& trie, std::set<NodeIndex> &res, NodeIndex Start) 
{
    const char*         pLabel; 
    SmallCounter        LabelLen;
    SmallCounter        ChildQty;
    const NodeIndex*    pChildren;

    trie.GetNodeInfo(Start, ChildQty, LabelLen, pLabel, pChildren);

    res.insert(Start);

    for (SmallCounter i = 0; i < ChildQty; ++i) {
        GetTrieNodesStartNode(trie, res, pChildren[i]);
    }
}

template <typename NodeAllocator>
void GetTrieNodes(const CCompTrie<NodeAllocator>& trie, std::set<NodeIndex> &res) {
    GetTrieNodesStartNode(trie, res, trie.GetRoot());
}

template <typename NodeAllocator>
void PrintTrieStrings(const CCompTrie<NodeAllocator>& trie) {
    std::vector<std::string> StrList;

    GetTrieStrings(trie, StrList);

    std::cout << "--------- Trie strings ----------" << std::endl;
    for (int i = 0; i < StrList.size(); ++i) {
        std::cout << StrList[i] << std::endl;
    }
    std::cout << "---------------------------------" << std::endl;
}

struct CheckerVisitor {
    CheckerVisitor(std::ostream &out) : mMemUsed(0), mNodeCnt(0), mFail(false), mOut(out) {}
    void operator() (NodeIndex CurrIndex, unsigned int CurrLevel, SmallCounter LabelLen, const char* pLabel, SmallCounter ChildQty) {
        if (CurrLevel && !LabelLen && ChildQty) {
            mOut << "Node # " << CurrIndex << " at level # " << CurrLevel << " has empty label (child qty = " << int(ChildQty) << ")";
            mFail = true;
        }
        mMemUsed += CNode::NodeSize(LabelLen, ChildQty);
        ++mNodeCnt;
    }
    size_t          mMemUsed;
    size_t          mNodeCnt;
    bool            mFail;
    std::ostream    &mOut;
};

// Check for abnormalities: empty labels in non-root nodes.
// Calculates the exact amount of memory used.
template <typename NodeAllocator>
bool CheckTrie(const CCompTrie<NodeAllocator>& trie, size_t& MemUsed, size_t& NodeCnt) {
    CheckerVisitor vis(std::cerr);

    trie.VisitTrie(vis);

    MemUsed = vis.mMemUsed;
    NodeCnt = vis.mNodeCnt;

    return !vis.mFail;
}

#endif
