#include <assert.h>
#include <stdlib.h>

#include <cstdio>
#include <vector>

#include "fileutil.h"
#include "cmnutil.h"
#include "../Common/myexcept.h"

using namespace std;

off_t CFileWrapper::FileSize(const std::string & fileName)
{
    struct stat buf;
    if (stat(fileName.c_str(), &buf))
        return -1;

    return buf.st_size;
};

void CFileWrapper::Rename(const std::string & srcFileName, const std::string & dstFileName)
{
    std::string error;

    try {
        errno = 0;
        int res = rename(srcFileName.c_str(), dstFileName.c_str());

        if (res) {
            error = "Error renaming file: '" + srcFileName + " to '" + dstFileName + "' system error: '" + SysError(errno) + "'";
        }
    } catch (...) {
        throw;
    }

    if (!error.empty()) {
        THROW_RUNTIME_ERROR(error);
    }
}

void CFileWrapper::Remove(const std::string & fileName)
{
    std::string error;

    try {
        errno = 0;
        int res = unlink(fileName.c_str());

        if (res) {
            error = "Error removing file: '" + fileName + "' system error: '" + SysError(errno) + "'";
        }
    } catch (...) {
        throw;
    }

    if (!error.empty()) {
        THROW_RUNTIME_ERROR(error);
    }
}

void CFileWrapper::openCreateTemp(const std::string & currDir, mode_t mode)
{
    std::string error;

    if (opened()) {
        close();
    }

    try {
        std::string   templ = currDir + "/tmpXXXXXX";
        char          aBuf[templ.size() + 1];
        errno = 0;

        strcpy(aBuf, templ.c_str());
        mFd = mkstemp(aBuf);
        if (mFd < 0) {
            error = "Error creating tmp file in the dir '" + currDir + "' system error: '" + SysError(errno) + "'";
        } else {
            mFileName = aBuf;
        }
        if (error.empty()) {
            if (fchmod(mFd, mode)) {
                error = "Error while chmoding of tmp file: '" + mFileName + "' system error: '" + SysError(errno) + "'";
            }
        }
    } catch (...) {
        if (mFd >= 0) {
            ::close(mFd);
        }
        if (!mFileName.empty()) {
            RemoveNoExcept(mFileName);
            mFileName = "";
        }
        throw;
    }
    if (!error.empty()) {
        mFileName = "";
        THROW_RUNTIME_ERROR(error);
    }
}

void CFileWrapper::open(const std::string & fileName, int flags, mode_t mode)
{
    std::string error;

    if (opened()) {
        close();
    }

    try {
        errno = 0;
        mFd = ::open(fileName.c_str(), flags, mode);
        if (mFd < 0) {
            error = "Error opening file: '" + fileName + "' system error: '" + SysError(errno) + "'";
        }
    } catch (...) {
        throw;
    }
    if (!error.empty()) {
        THROW_RUNTIME_ERROR(error);
    }
    mFileName = fileName;
}

void CFileWrapper::close()
{
    if (!opened()) {
        return;
    }
    std::string error;

    try {
        errno = 0;
        if (::close(mFd)) {
            error = "Error closing file: '" + mFileName + "' system error: '" + SysError(errno) + "'";
        }
    } catch (...) {
        throw;
    }
    if (!error.empty()) {
        THROW_RUNTIME_ERROR(error);
    }
    mFd = -1;
    mFileName = "";
}

CFileWrapper::~CFileWrapper() {
    try {
        close();
    } catch (...) {
    }
}

size_t CFileWrapper::read(void * pBuf, size_t len)
{
    ssize_t res = 0;
    std::string error;

    if (!opened()) {
        THROW_RUNTIME_ERROR("read attempt on closed file");
    }
    try {
       errno = 0;
       res = ::read(mFd, pBuf, len); 

       if (res < 0 || errno) {
           error = "Error reading from file: '" + mFileName + "' system error: '" + SysError(errno) + "'";
           THROW_RUNTIME_ERROR(error);
       }
    } catch (...) {
        throw;
    }
    if (!error.empty()) {
        THROW_RUNTIME_ERROR(error);
    }

    return res;
}

size_t CFileWrapper::write(const void * pBuf, size_t len)
{
    ssize_t res = 0;
    std::string error;

    if (!opened()) {
        THROW_RUNTIME_ERROR("write attempt on closed file");
    }
    try{
       errno = 0;
       res = ::write(mFd, pBuf, len); 

       if (res < 0 || errno) {
           error = "Error writing to file: '" + mFileName + "' system error: '" + SysError(errno) + "'";
           THROW_RUNTIME_ERROR(error);
       }
    } catch (...) {
        throw;
    }
    if (!error.empty()) {
        THROW_RUNTIME_ERROR(error);
    }

    return res;
}

CBufferedFileReader::CBufferedFileReader( size_t bufSize) : mBufSize(bufSize), mBufRead(0), mBufPos(0)
{
    mpBuf = new char[bufSize];
}

void CBufferedFileReader::open(const std::string fileName) 
{
    if (mFile.opened()) {
        close();
    }
    mFile.open(fileName, O_RDONLY);
}

void CBufferedFileReader::close()
{
    mFile.close();
    mBufRead = mBufPos = 0;
}

size_t CBufferedFileReader::read(void * pBuf_, size_t len)
{
    char *   p = reinterpret_cast<char *>(pBuf_);
    size_t   res = 0;

    while (len) {

       if (mBufPos >= mBufRead) {
           mBufRead = mFile.read(mpBuf, mBufSize);
           mBufPos = 0;

           if (!mBufRead) {
               break;
           }
       }

       size_t   fromBuf = std::min(len, mBufRead - mBufPos);
       assert(fromBuf);

       memcpy(p, mpBuf + mBufPos, fromBuf);

       p += fromBuf;
       len -= fromBuf;
       mBufPos += fromBuf;
       res += fromBuf;
    }

    return res;
}

#include <iostream>
#include "testcmn.h"

int testBufferedFileReader(size_t fileSize, size_t bufSize)
{
    const char *                 pFname = "./ttttt1024";
    vector<unsigned char>   vBuf(fileSize);
    unsigned char *         pBuf = &vBuf[0];

    try {
        CFileWrapper        file;
        CBufferedFileReader bread(bufSize);

        file.open(pFname, O_RDWR | O_CREAT | O_TRUNC);
  
        for (unsigned int i = 0; i < fileSize; ++i) {
            pBuf[i] = rand();

            file.write(pBuf + i, 1);
        }

        file.close();
        // one-byte reading test
        bread.open(pFname);

        for (unsigned i = 0; i < fileSize; ++i) {
            unsigned char  c;
            size_t         res = bread.read(&c, 1);
            if (res != 1) {
                cerr << "Byte # " << i << " cannot be read" << FILE_POS_STR << endl;

                return 1;
            }                    
            if (pBuf[i] != c) {
                cerr << "Byte # " << i << " is read wrongly,expected " << pBuf[i] << " got " << c << FILE_POS_STR << endl;

                return 1;
            }
        }
        bread.close();
        bread.open(pFname);
        // random-size reading test
        
        vector<unsigned char>   vBuf1(fileSize);

        size_t              res, cnt = 0;

        do {
            size_t cnt1 = (rand() + 1) % (fileSize / 10 + 1);

            res = bread.read(&vBuf1[cnt], cnt1);
            cnt += res;

            assert(cnt <= fileSize);
        } while (res);

        for (unsigned i = 0; i < fileSize; ++i) {
            if (pBuf[i] != vBuf1[i]) {
                cerr << "Byte # " << i << " is read wrongly,expected " << pBuf[i] << " got " << vBuf1[i] << FILE_POS_STR << endl;

                return 1;
            }
        }
        bread.close();

        CFileWrapper::Remove(pFname);
    } catch (const std::runtime_error & e) {
        cout << "testBufferedFileReader got exception: " << e.what() << FILE_POS_STR << endl;
        return 1;
    } catch (...) {
        cout << "testBufferedFileReader got unknown exception: " << FILE_POS_STR << endl;

        return 1;
    }

    return 0;
}

int testFileWrapper()
{
    const char * pFname = "./ttttt1024";

    try {
        CFileWrapper   file;

        file.open(pFname, O_RDWR | O_CREAT | O_TRUNC);
  
        for (unsigned char c = 255; ; --c) {
            file.write(&c, 1);
            if (!c) {
                break;
            }
        }

        if (file.getFileName() != pFname) {
            cerr << "Wrong file name stored, expect " << pFname << " but got " << file.getFileName() << FILE_POS_STR << endl;
        }

        file.close();

        file.open(pFname, O_RDONLY);

        char buf[256];

        ssize_t sz = file.read(buf, 255);
          
        if (sz != 255) {
            cerr << "Read wrong number of bytes, expected 255 but got " << sz << FILE_POS_STR << endl;
            return 1;
        }

        sz = file.read(buf + 255, 1000);
          
        if (sz != 1) {
            cerr << "Read wrong number of bytes, expected 1 but got " << sz << FILE_POS_STR << endl;
            return 1;
        }

        for (int i = 0; i < 256; ++i)
        if ((unsigned char)buf[i] != 255 - i) {
            cerr << "Byte # " << i << " is read wrongly,expected " << (255 - i) << " got " << int(buf[i]) << FILE_POS_STR << endl;

            return 1;
        }

        if ((sz = CFileWrapper::FileSize(pFname)) != 256) {
            cerr << "Wrong file size, expect 256 but got " << sz << FILE_POS_STR << endl;
            return 1;
        }

        CFileWrapper::Remove(pFname);
    } catch (const std::runtime_error & e) {
        cout << "testFileWrapper got exception: " << e.what() << FILE_POS_STR << endl;
        return 1;
    } catch (...) {
        cout << "testFileWrapper got unknown exception: " << FILE_POS_STR << endl;

        return 1;
    }
    return 0;
}
