#ifndef FILE_UTIL_H
#define FILE_UTIL_H

#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#include <unistd.h>
#include <pthread.h>
#include <errno.h>

#include <string>

#include "../Common/myexcept.h"
#include "cmnutil.h"

/* 
  Class is designed to automatize file closing when the variable is destroyed.
  It also fires exception on error. All file operation might be done automatically
  to be thread safe.
*/

class CFileWrapper {
public:
    // Returns the size of the file on success, fires an exception in case of error
    static off_t FileSize (const std::string & fileName);
    // Fires an exception in case of error
    static void Remove(const std::string & fileName);
    static void Rename(const std::string & srcFileName, const std::string & dstFileName);
    static void RemoveNoExcept(const std::string & fileName) {
        try {
           Remove(fileName);
        } catch (...) {
        }
    }
    CFileWrapper() : mFd(-1) {}
    ~CFileWrapper();
    bool opened() const {
       return mFd >= 0;
    }
    static const int CREATE_NEW = O_RDWR | O_CREAT | O_TRUNC;
    void open(const std::string & fileName, int flags, mode_t mode = 0600);
    void openCreateTemp(const std::string & curDir, mode_t mode = 0600);
    size_t read(void * pBuf, size_t len); // fires exception only in case of error
    size_t write(const void * pBuf, size_t len); // fires exception only in case of error
    void close(); // fires exception in case of error
    const std::string getFileName() const {
       return mFileName;
    }
    int getFd() {
        if (!opened()) {
            THROW_RUNTIME_ERROR("getFd failed - file is not opened");
        }
        return mFd;
    }
    off_t getOffset() {
        if (!opened()) {
            THROW_RUNTIME_ERROR("getOffset failed - file is not opened");
        }
        off_t res = lseek(mFd, 0, SEEK_CUR);

        if (res < 0) {
            THROW_RUNTIME_ERROR("lseek failed: " + SysError());
        }
    }
    void setOffset(off_t pos) {
        if (!opened()) {
            THROW_RUNTIME_ERROR("setOffset failed - file is not opened");
        }
        off_t res = lseek(mFd, pos, SEEK_SET);

        if (res < 0) {
            std::stringstream str;
            str << "lseek for position: " << pos << " failed: " + SysError();
            THROW_RUNTIME_ERROR(str.str());
        }
    }
private:
    void open(const std::string & fileName, int flags, mode_t mode, bool modePres);
    std::string              mFileName;
    int                      mFd;
};

struct CBufferedFileReader {
    ~CBufferedFileReader() {
        delete [] mpBuf;
        try {
            close();
        } catch (...) {
        }
    }
    CBufferedFileReader(size_t bufSize);
    void open(const std::string fileName);
    void close();
    size_t read(void * pBuf, size_t len); // fires exception only in case of error
private:
    size_t        mBufSize;
    size_t        mBufRead;
    size_t        mBufPos;
    char *        mpBuf;
    CFileWrapper  mFile;
};

int testFileWrapper();
int testBufferedFileReader(size_t fileSize, size_t bufSize);

#endif
