#include "kdtree.h"
#include "format.h"

static unsigned int compsortlevel;
std::allocator<char> chara;

const char kdtree::termChar = '@';

bool  kdtreeNodeComp::operator ()(kdtree * a, kdtree * b)  {
    unsigned int limit =
        compsortlevel + min (a->getMaxFuzzy (), b->getMaxFuzzy ());
    for (unsigned int i = compsortlevel; i < limit; i++)
         {
        if (a->getIndexValue (i) < b->getIndexValue (i))
            return true;
        if (a->getIndexValue (i) > b->getIndexValue (i))
            return false;
        }
    return false;
}

unsigned int  kdtree::computeDist (VectorType key1[]) 
{
    return::computeKeyDistance (key1, this->key, MaxFuzzy);
}

kdtree::kdtree(CFastAllocator & kdAlloc, CFastAllocator & vecAlloc, 
                unsigned int MaxFuzzy, const char * stringData, 
                VectorType * vec_p):
        kdAlloc(kdAlloc), vecAlloc(vecAlloc),
        MaxFuzzy (MaxFuzzy), parent (NULL), loson (NULL), hison (NULL),
    insertmode (1), leftdepth (0), rightdepth (0), nodelevel (0),
level (0), keysum (0)
{
#ifdef USE_CHUNK
    if (stringData) {
       dataChunk.append(stringData);
       dataChunk.append(&termChar, 1);
       dataChunkCnt = 1;
    }
#else
    if (stringData) {
       data.push_back(stringData);
    }
#endif
    void * p = vecAlloc.allocChunk();
    key = new (p) VectorType[MaxFuzzy];
    if (vec_p != NULL) {
        memcpy (key, vec_p, MaxFuzzy * sizeof (VectorType));
    } else {
        memset (key, 0, MaxFuzzy * sizeof (VectorType));
    }
    keysum = computeKeySum ();
}

// Here params are shifted for purpose so as two overloaded constructor instances do
// not intersect
kdtree::kdtree(unsigned int MaxFuzzy, CFastAllocator & kdAlloc, CFastAllocator & vecAlloc, 
#ifdef USE_CHUNK
                  const std::string &dataChunkParam,
#else
                  ConstCharPtrs data,
#endif
                  VectorType * vec_p):
        kdAlloc(kdAlloc), vecAlloc(vecAlloc),
        MaxFuzzy (MaxFuzzy), 
#ifdef USE_CHUNK
        dataChunk(dataChunkParam),
#else
        data(data),
#endif
    parent (NULL), loson (NULL), hison (NULL), insertmode (1),
    leftdepth (0), rightdepth (0), nodelevel (0), level (0), keysum (0)
{
#ifdef USE_CHUNK
    dataChunkCnt = count(dataChunk.data(), dataChunk.data() + dataChunk.size(), termChar);
#endif
    void * p = vecAlloc.allocChunk();
    key = new (p) VectorType[MaxFuzzy];
    if (vec_p != NULL) {
        memcpy (key, vec_p, MaxFuzzy * sizeof (VectorType));
    } else {
        memset (key, 0, MaxFuzzy * sizeof (VectorType));
    }
    keysum = computeKeySum ();
}

kdtree::~kdtree () 
{
    /* 
     * Since most of the memory is allocated
     * using CFastAllocator, which will free
     * the memory as soon we call allocator 
     * destructor we MUST NOT use delete here, 
     * which assumes that the memory was allocated
     * directly on a heap, using oridinary 
     * non-placment new
     */
    if (hison) {
        hison->~kdtree();
    }
    if (loson) {
        loson->~kdtree();
    }
    /*
     * It also worth freeing memory for vectors
     * so as the memory could be reused. We cannot
     * free kdtree memory  here, because it is not 
     * always allocated via kdAlloc. 
     */
    vecAlloc.freeChunk(key);
}


/*
** Function dataAsString returns a string representation of a node
*/ 
string  kdtree::dataAsString () const 
{
    string res = "";
#ifdef USE_CHUNK
    res = dataChunk;
    replace(res.begin(), res.end(), termChar, '|');
    // delete the last '|'
    if (res.size()) {
        res.erase(res.size() - 1, 1);
    }
#else
    unsigned int i;
    for (i = 0; i < data.size (); i++) {
        string tmp = data[i];
        res += tmp; 
        res += string ((i < data.size () - 1) ? "|" : "");
    }
#endif
    return res;
}


/* getIndexValue: gets the value of the key at position level */ 
int  kdtree::getIndexValue (unsigned int inlevel) 
{
    return key[inlevel % MaxFuzzy];
}

unsigned int computeKeyDistance (VectorType key1[], VectorType key2[],
                    unsigned int MaxFuzzy) 
{
    int dist1 = 0;
    int dist2 = 0;

    for (unsigned int i = 0; i < MaxFuzzy; i++) {
        int d = abs (int (key1[i]) - int (key2[i]));
        if (key1[i] > key2[i]) 
            dist1 += d;
        else 
            dist2 += d;
    }

    return std::max(dist1, dist2);
}

void  kdtree::printKey () const
{
    printf ("[%s] Key  :",
             (parent !=
              NULL) ? ((parent->loson == this) ? "LO" : "HI") : "");
    for (unsigned int i = 0; i < MaxFuzzy; i++)
        printf ("%c%d", (i) ? '|' : ' ', key[i]);
    printf (" ");
} 

/* prints the content of a node */ 
void  kdtree::printNode ()  const
{
    for (unsigned int i = 0; i < nodelevel; i++)
        printf (" ");
    if (parent != NULL)
        printf ("Dist : %d ",::
                computeKeyDistance (parent->key, key, MaxFuzzy));
    printf ("Data : [%s]", dataAsString ().c_str ());
    printKey ();
    printf ("Depths : (%d, %d) ", leftdepth, rightdepth);
    printf ("level: %d ", level);
    printf ("\n");
}


/* prints the full k-d tree */ 
void  kdtree::printTree () const
{
    printNode ();
    if (loson != NULL)
        loson->printTree ();
    if (hison != NULL)
        hison->printTree ();
    return;
}


/* insertnode: inserts a node in the parent node */ 
void kdtree::insertNode (kdtree * nodetoadd, int where, int insertlevel) 
{
    bool leafnode = ((loson == NULL) && (hison == NULL));
    insertlevel %= MaxFuzzy;
    if (where == 0) {
        loson = nodetoadd;
        leftdepth = leftdepth + 1;
    } else {
        hison = nodetoadd;
        rightdepth = rightdepth + 1;
    }
    nodetoadd->parent = this;
    nodetoadd->nodelevel = nodelevel + 1;
    nodetoadd->level = insertlevel;
    kdtree * nodeparent = parent;
    kdtree * currNode = this;
    while (nodeparent != NULL) {
        if (leafnode == 1) {
            if (currNode == nodeparent->loson) {
                nodeparent->leftdepth =
                    max (currNode->leftdepth, currNode->rightdepth) + 1;
            } else {
                nodeparent->rightdepth =
                    max (currNode->leftdepth, currNode->rightdepth) + 1;
            }
        }
        currNode = nodeparent;
        nodeparent = nodeparent->parent;
    }
    return;
}


/*
** Function insertNodeIntoTree 
** Description	inserts a node into a tree using the k-f algorithm starting from the tree root
** returns: 1 for success, 
**          2 if node existed and data part was  updated, 
**         -1 if data already was in tree
**         -1 is also returned if insertmode is not 1 and the key exists
*/ 
int kdtree::insertNodeIntoTree (VectorType insKey[], const char * newData, kdtree ** res_pp) 
{
    kdtree * currNode = this;
    int where = 0;
    int insertlevel = 0;
    void * p = kdAlloc.allocChunk();
    kdtree * newNode = new (p) kdtree (kdAlloc, vecAlloc, MaxFuzzy, newData, insKey);
    while (true) {
        if (::computeKeyDistance (currNode->key, newNode->key, MaxFuzzy) == 0) {
            // Node is not going to be used. Free unused node in two steps:
            newNode->~kdtree();         // firstly, call destructor
            kdAlloc.freeChunk(newNode); // secondly, return memory to the allocator
            if (insertmode == 1) {
#ifdef USE_CHUNK
                    currNode->dataChunk.append(newData);
                    currNode->dataChunk.append(&termChar, 1);
                    currNode->dataChunkCnt++;
#else
                    unsigned int sz = currNode->data.size();
                    currNode->data.reserve(sz + 1);
                    currNode->data.push_back(newData);
#endif
                    
                    if (res_pp != NULL)
                        *res_pp = currNode;
                    return 2;
            } else {
                return -1;
            }
        }
        where = 0;
        if (newNode->getIndexValue(insertlevel) > currNode->getIndexValue (insertlevel))
            where = 1;
        if (where == 0) {
            if (currNode->loson == NULL) {
                currNode->insertNode (newNode, where, insertlevel + 1);
                if (res_pp)
                    *res_pp = newNode;
                return 1;
            } else {
                currNode = currNode->loson;
            }
        } else {
            if (currNode->hison == NULL) {
                currNode->insertNode (newNode, where, insertlevel + 1);
                if (res_pp)
                    *res_pp = newNode;
                return 1;
            } else
                currNode = currNode->hison;
        }
        insertlevel = (insertlevel + 1) % MaxFuzzy; 
    }
    return 1;
}

void kdtree::Traverse(VectorType key[], unsigned int d,
                vi &prob, vi &cnt)
{
    if (hison && loson) {
        if (prob.size() < level + 1) {
            prob.resize(level + 1);
            cnt.resize(level + 1);
        }
        cnt[level]++;

        int val1 = getIndexValue(level);
        int val2 = key[level % MaxFuzzy];

        if (abs(val1 -  val2) <= d) {
            prob[level]++;
        }
    }
    if (hison) {
        hison->Traverse(key, d, prob, cnt);
    }
    if (loson) {
        loson->Traverse(key, d, prob, cnt);
    }
}

/* Search: searches for a key within a given distance */ 
void kdtree::Search(kdtree * cmpkey, HyperCube * cube,
                    unsigned int maxDist, myagrep &a, 
#ifdef _TEST_KDTREE
                    NodeSet * matches, 
#endif  /*  */
                    StringList * strMatches, int *numComp) 
{
    unsigned int dist =::computeKeyDistance (key, cmpkey->key, MaxFuzzy);
    if (dist <= maxDist) {
        
#ifdef _TEST_KDTREE
            if (matches)
            matches->insert(this);
        
#endif  /*  */
#ifdef USE_CHUNK
        if (numComp) {
            (*numComp) += dataChunkCnt;
        }

        a.findPattern(CInterface::filterType, termChar,
                      dataChunk.data(), dataChunk.data() + dataChunk.size(), 
                      true, strMatches);
#else
        for (unsigned int i = 0; i < data.size (); i++) {
            if (numComp)
                (*numComp)++;
            if (a.checkRTRFilter(data[i])) {
                if (strMatches)
                    strMatches->push_back(data[i]);
            }
        }
#endif
    }
    if (hison == NULL && loson == NULL)
        return;
    if (loson != NULL) {
        VectorType hiKey = cube->hiKey[level];
        cube->hiKey[level] = min (cube->hiKey[level], key[level]);
        dist = cube->computeDistance (cmpkey->key);
        if (dist <= maxDist)
            loson->Search (cmpkey, cube, maxDist, a, 
#ifdef _TEST_KDTREE
                           matches, 
#endif  /*  */
                           strMatches, numComp);
        cube->hiKey[level] = hiKey;
    }
    if (hison != NULL) {
        VectorType loKey = cube->loKey[level];
        cube->loKey[level] = max (cube->loKey[level], key[level]);
        if (cube->loKey[level] < VectorTypeMax)
            cube->loKey[level]++;
        dist = cube->computeDistance (cmpkey->key);
        if (dist <= maxDist)
            hison->Search (cmpkey, cube, maxDist, a, 
#ifdef _TEST_KDTREE
                           matches, 
#endif  /*  */
                           strMatches, numComp);
        cube->loKey[level] = loKey;
    }
}

int  kdtree::DumpToFile (int fd, off_t offset) 
{
    size_t delta = 0;
    size_t buflen = BufLen();
    
// We store the data in the following order
// 0. buflen value
// 1. MaxFuzzy value
// 2. The key
// 3. offset of the left subtree, offset of the right subtree
// 4. data array:
//   BUT: In chunk mode we also store dataChunkCnt
    if (offset) {
        delta = (offset % getpagesize ());
    }
    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0
         || write (fd, "", 1) < 1) {
        return -1;
    }
    void *buf = mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ, MAP_SHARED, fd, offset - delta);
    if (!buf || buf == MAP_FAILED) {
        return -1;
    }
    char *sbuf = (char *) buf + delta;
    
    // 0.
    *((size_t *) sbuf) = buflen;
    sbuf += sizeof (size_t);
    
    // 1.
    *((unsigned int *) sbuf) = MaxFuzzy;
    sbuf += sizeof (unsigned int);

    
    // 2.
    memcpy (sbuf, key, sizeof (VectorType) * MaxFuzzy);
    sbuf += sizeof (VectorType) * MaxFuzzy;
    off_t * childOff = (off_t *) sbuf;
    
    // 3. at this moment we don't know offsets exactly using zeros
    childOff[0] = 0;
    childOff[1] = 0;
    sbuf += sizeof (off_t) * 2;

    
    // 4.
#ifdef USE_CHUNK
    *((int *)sbuf) = dataChunkCnt;
    sbuf += sizeof(dataChunkCnt);
    unsigned int len = dataChunk.size();
    memcpy(sbuf, dataChunk.data(), len);
    sbuf += len;
#else
    for (unsigned int i = 0; i < data.size (); i++) {
        unsigned int len = strlen(data[i]) + 1;
        memcpy (sbuf, data[i], len);
        sbuf += len;
    } 
#endif
    assert (sbuf - (char *) buf == int (delta + buflen));
    kdtree * nodes[2] = {loson, hison};
    unsigned int sub_sz = 0;
    for (unsigned int i = 0; i < 2; i++) {
        if (nodes[i]) {
            off_t s_sz = nodes[i]->DumpToFile (fd, offset + buflen + sub_sz);
            
            // 3.
            childOff[i] = offset + buflen + sub_sz;
            if (s_sz < 0) {
                munmap (buf, buflen + delta);
                return -1;
            }
            sub_sz += s_sz;
        }
    }
    //(void)msync(buf, buflen + delta, 0);
    if (munmap (buf, buflen + delta) < 0) {
        return -1;
    }
    // Returning the total number of bytes written
    return buflen + sub_sz;
}

bool kdtree::SearchInFileAtLevel (int fd, off_t offset, unsigned int level,
                                 HyperCube * cube, VectorType skey[],
                                 unsigned int maxDist, myagrep &a,
                                 StringList * resList, int *numComp) 
{
    unsigned int buflen;
    int delta = 0;
    unsigned int MaxFuzzy;
    if (lseek (fd, offset, SEEK_SET) < 0) {
        return false;
    }
    if (read (fd, &buflen, sizeof (buflen)) < (int) sizeof (buflen)) {
        return false;
    }
    if (lseek (fd, -sizeof (buflen), SEEK_CUR) < 0) {
        return false;
    }
    if (offset)
        delta = (offset % getpagesize ());
    
// We retrieve the data in the following order
// 0. buflen value
// 1. MaxFuzzy value
// 2. The key
// 3. offset of the left subtree, offset of the right subtree
// 4. data array:
//    BUT in chunk mode it is preceeded by string count
    void *buf = mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd, offset - delta);
    if (!buf || buf == MAP_FAILED) {
        return false;
    }
    char *sbuf = (char *) buf + delta;
    
    // 0.
    buflen = *((size_t *) sbuf);
    sbuf += sizeof (size_t);
    
    // 1.
    MaxFuzzy = *((unsigned int *) sbuf);
    sbuf += sizeof (unsigned int);
    if (!cube) {
        cube = new HyperCube (MaxFuzzy);
    } else {
        if (MaxFuzzy != cube->size()) {
            std::cerr << "Strange MaxFuzzy = " << MaxFuzzy << " buflen = " << buflen <<" at level " << level << " offset = " << offset << std::endl;
        }
        assert (MaxFuzzy == cube->size ());
    }
    VectorType key[MaxFuzzy];
    memcpy (key, sbuf, sizeof (VectorType) * MaxFuzzy);
    
    // 2.
    sbuf += sizeof (VectorType) * MaxFuzzy;
    
    // 3. 
    off_t * childOff = (off_t *) sbuf;
    sbuf += sizeof (off_t) * 2;
    
    // 4. We retrieve data list if and only if the distance is appropriate
    unsigned int dist =::computeKeyDistance (key, skey, MaxFuzzy);
    if (dist <= maxDist) {
        char *end = (char *) buf + delta + buflen;
        char *start = sbuf;
#ifdef USE_CHUNK
        int cnt = *((int *)start);
        start += sizeof(int);
        a.findPattern(CInterface::filterType, termChar, start, end, true, resList);
#else
        for (; sbuf < end; sbuf++) {
            if (*sbuf == 0 || sbuf == end - 1) {
                if (numComp)
                    (*numComp)++;
                if (a.checkRTRFilter(start)) {
                    if (resList != NULL)
                        resList->push_back (string (start, sbuf));
                }
                start = sbuf + 1;
            }
        }
#endif
    }
    for (int i = 0; i < 2; i++)
        if (childOff[i] != 0) {
            VectorType loKey = cube->loKey[level];
            VectorType hiKey = cube->hiKey[level];
            if (i != 0) {
                cube->loKey[level] = max (cube->loKey[level], key[level]);
                if (cube->loKey[level] < VectorTypeMax)
                    cube->loKey[level]++;
            } else {
                cube->hiKey[level] = min (cube->hiKey[level], key[level]);
            }
            dist = cube->computeDistance (skey);
            if (dist <= maxDist) {
                if (!SearchInFileAtLevel
                     (fd, childOff[i], (level + 1) % MaxFuzzy, cube, skey,
                      maxDist, a, resList, numComp))
                    goto error;
            }
            cube->loKey[level] = loKey;
            cube->hiKey[level] = hiKey;
        }
    if (!offset)
        delete cube;
    if (munmap (buf, buflen + delta) < 0)
        return -1;
    return true;
error:
    munmap (buf, buflen + delta);
    if (!offset)
        delete cube;
    return false;
}

int  kdtree::computeKeySum ()  const
{
    int isum = 0;
    for (unsigned int i = 0; i < MaxFuzzy; i++)
        isum += key[i];
    return isum;
}

void  kdtree::fillInorderTraversal (NodeSet * list) 
{
    if (loson)
        loson->fillInorderTraversal (list);
    list->push_back(this);
    if (hison)
        hison->fillInorderTraversal (list);
}


/*
** We have a binary tree, hence, balancing is simple: 
** after all nodes are sorted, we choose a median to be a new root.
** Nodes on the left and right median side are reccursively turned 
** into a balanced tree.
**
** Function returns new, balanced tree
*/ 
kdtree * kdtree::getBalancedCopy (CFastAllocator & newKdAlloc, CFastAllocator & newVecAlloc) 
{
    vector<kdtree *>    list, lolist, hilist;
    kdtreeNodeComp       cmp;

    fillInorderTraversal (&list);
    sort(list.begin(), list.end(), cmp);

    compsortlevel = 0;
    unsigned int nofnodes = list.size ();
    unsigned int median = nofnodes / 2;
    NodeSetIterator pos = list.begin ();
    kdtree * newroot = NULL;
    for (unsigned int i = 0; i < nofnodes; i++, pos++) {
        if (i < median)
            //lolist.insert (*pos);
            lolist.push_back(list[i]);
        if (i > median)
            //hilist.insert (*pos);
            hilist.push_back(list[i]);
        if (i == median) {
            void * p = newKdAlloc.allocChunk();
#ifdef USE_CHUNK
            newroot = new (p) kdtree (MaxFuzzy, newKdAlloc, newVecAlloc, list[i]->dataChunk, list[i]->key);
#else
            newroot = new (p) kdtree (MaxFuzzy, newKdAlloc, newVecAlloc, (*pos)->data, (*pos)->key);
#endif
        }
    }
    assert (newroot);
    newroot->insertnext (lolist, 1);
    newroot->insertnext (hilist, 1);
    return newroot;
}

void  kdtree::insertnext (const NodeSet & sortedlist, int sortlevel) 
{
    unsigned int nofnodes = sortedlist.size ();
    if (nofnodes == 0) {
        return;
    }
    //NodeSet sortedlist;
    NodeSet lolist;
    NodeSet hilist;
    NodeSetConstIterator pos;
    
    /* resort the list */ 
    compsortlevel = sortlevel;

    unsigned int median = nofnodes / 2;
    pos = sortedlist.begin ();
    for (unsigned int i = 0; i < nofnodes; i++, pos++) {
        if (i < median)
            lolist.push_back (*pos);
        if (i > median)
            hilist.push_back (*pos);
        if (i == median) {
#ifdef USE_CHUNK
            int prevPos = 0;
            for (unsigned int k = 0; k < sortedlist[i]->dataChunk.size(); ++k) {
                if (sortedlist[i]->dataChunk[k] == termChar) {
                    std::string str(sortedlist[i]->dataChunk.data() + prevPos, k - prevPos + 1);
                    insertNodeIntoTree ((*pos)->key, str.c_str(), NULL);
                    prevPos = k + 1;
                }
            }
#else
            for (unsigned j = 0; j < (*pos)->data.size (); j++) {
                insertNodeIntoTree ((*pos)->key, (*pos)->data[j], NULL);
            }
#endif
        }
    } 
    insertnext (lolist, sortlevel + 1);
    insertnext (hilist, sortlevel + 1);
} 

#ifdef _TEST_KDTREE
bool  testComputeDist () 
{
    const unsigned int MaxFuzzy = 2;
    HyperCube cube (MaxFuzzy);
    cube.loKey[0] = 10;
    cube.loKey[1] = 20;
    cube.hiKey[0] = 30;
    cube.hiKey[1] = 40;
    VectorType key[4][MaxFuzzy] = {{10, 40},
                                   {9, 41},
                                   {30, 19},
                                   {9, 40}};

    unsigned int  expVal[4] = { 0, 2, 1, 1 };
    int           nFailed = 0;

    for (int i = 0; i < 4; i++) {
        unsigned int res = cube.computeDistance (key[i]);
        if (res != expVal[i]) {
            printf("testComputeDist, subtest # %d failed, expected %d, but received %d\n",
                    i, expVal[i], res);
            nFailed++;
        }
    }
    return nFailed == 0;
}

bool  testSimpleSearch (int num, int maxDist, bool silent) 
{
    const unsigned   int MaxFuzzy = 2;
    VectorType       key[num][MaxFuzzy];
    kdtree           tree1(MaxFuzzy);
    StringList       dict;

    for (int i = 0; i < num; i++) {
        key[i][0] = i;
        key[i][1] = num - i;
        dict.push_back(Format ("string #%d", i));
        tree1.insertNodeIntoTree (key[i], dict.back().c_str(), NULL);
    } 
    kdtree * tree = tree1.getBalancedCopy ();
    if (!silent)
        printf ("===========================\n");
    if (!silent)
        tree->printTree ();
    VectorType key1[MaxFuzzy];
    for (int i = 0; i < num; i++) {
        for (int j = 0; j <= maxDist; j++) {
            key1[0] = i + j;
            key1[1] = num - i + j;
            for (int k = 0; k <= 2 * j; k++) {
                NodeSet res;
                NodeSetIterator pos;
                if (!silent)
                    printf ("Searching [%d,%d] dist=%d\n", key1[0],
                            key1[1], k);
                tree->Search (key1, k, NULL, &res, NULL);
                bool found = false;
                for (pos = res.begin (); pos != res.end (); pos++) {
                    found = true;
                    unsigned int dist = (*pos)->computeDist (key1);
                    if (!silent)
                        printf
                            ("Found [%d,%d] (searched [%d,%d]) dist =%d\n",
                             (*pos)->getIndexValue (0),
                             (*pos)->getIndexValue (1), key1[0], key1[1],
                             dist);
                    if (dist > k) {
                        if (!silent)
                            printf
                                ("Returned key [%d,%d] dist %d, search dist %d\n",
                                 (*pos)->getIndexValue (0),
                                 (*pos)->getIndexValue (1), dist, k);
                        delete tree;
                        return false;
                    }
                }
                bool fstat = (k == 2 * j);
                if (found != fstat) {
                    if (!silent)
                        printf
                            ("Key [%d,%d], wrong found status %d (expect. %d) search dist %d j=%d\n",
                             key1[0], key1[1], found, fstat, k, j);
                    delete tree;
                    return false;
                }
            }
        }
    }
    delete tree;
    return true;
}

bool 
testSearchRandom (int num, const int MaxFuzzy, unsigned int seed,
                  bool silent) 
{
    VectorType key[num][MaxFuzzy];
    string vals[num];
    kdtree tree1 (MaxFuzzy);
    unsigned int maxDist = max (5, MaxFuzzy / 5);
    srand (seed);
    for (int i = 0; i < num; i++) {
        for (int j = 0; j < MaxFuzzy; j++) {
            key[i][j] = rand () % (maxDist);
        } 
        vals[i] = Format ("string #%d", i);
        tree1.insertNodeIntoTree (key[i], &vals[i], NULL);
    } kdtree * tree = tree1.getBalancedCopy ();
    if (!silent)
        printf ("===========================\n");
    if (!silent)
        tree1.printTree ();
    if (!silent)
        tree->printTree ();
    VectorType key1[MaxFuzzy];
    for (int i = 0; i < num; i++) {
        if (i % 100 == 0)
            printf ("\b\b\b\b\b%d", i);
        fflush (stdout);
        for (int j = 0; j <= maxDist; j++) {
            unsigned int rest = j;
            for (int k = 0; k < MaxFuzzy; k++) {
                int dmin = 0;
                if (key[i][k] > rest / 2)
                    dmin = key[i][k] - rest / 2;
                key1[k] = dmin;
                if (rest > 0) {
                    if (k < MaxFuzzy - 1)
                        key1[k] += rand () % rest;
                    
                    else
                        key1[k] = key[i][k] + rest;
                    rest -= abs ((int) key[i][k] - key1[k]);
                }
            } unsigned int dst =
                computeKeyDistance (key[i], key1, MaxFuzzy);
            if (dst != j) {
                if (!silent)
                    printf
                        ("Internal error: Wrong algorithm creating random key at specified dist!\n");
                return false;
            }
            for (int k = 0; k <= j; k++) {
                NodeSet res;
                NodeSetIterator pos;
                
                    //if (!silent) printf ("Searching [%d,%d] dist=%d\n", key1[0], key1[1], k);
                    tree1.Search (key1, k, NULL, &res, NULL);
                bool found = false;
                for (pos = res.begin (); pos != res.end (); pos++) {
                    kdtree * res = *pos;
                    int dist = res->computeDist (key1);
                    int l = 0;
                    for (l; l < MaxFuzzy; l++)
                        if (key[i][l] != res->getIndexValue (l))
                            break;
                    if (l == MaxFuzzy) {
                        found = true;
                        if (k < j) {
                            if (!silent) {
                                printf
                                    ("Internal error: found entry when searched for distance %d < %d",
                                     k, j);
                            }
                            delete[]tree;
                            return false;
                        }
                    }
                    if (dist > k) {
                        if (!silent)
                            printf
                                ("Internal error: returned dist = %d, search dist %d\n",
                                 dist, k);
                        delete tree;
                        return false;
                    }
                }
                if (k == j && found == false) {
                    printf ("k == j but entry ");
                    kdtree tree2 (MaxFuzzy, "", key1);
                    tree2.printKey ();
                    printf ("not found\n");
                    delete tree;
                    return false;
                }
            }
        }
    }
    const char *treeName = "kdtest.img";
    int fd = open (treeName, O_RDWR | O_CREAT | O_TRUNC, DefaultAccess);
    if (fd < 0) {
        Perror ("Cannot create file");
        return false;
    }
    if (tree->DumpToFile(fd, 0) < 0) {
        Perror ("Cannot dump to file");
    }
    close(fd);
    delete tree;
    tree = NULL;
    if (!silent)
        printf ("Search step 2 ====================\n");
    fd = open (treeName, O_RDONLY);
    if (fd < 0) {
        Perror (Format ("Cannot open file %s", treeName).c_str ());
        return false;
    }
    for (int i = 0; i < num; i++) {
        if (i % 100 == 0)
            printf ("\b\b\b\b\b%d", i);
        fflush (stdout);
        for (int j = 0; j <= maxDist; j++) {
            unsigned int rest = j;
            for (int k = 0; k < MaxFuzzy; k++) {
                int dmin = 0;
                if (key[i][k] > rest / 2)
                    dmin = key[i][k] - rest / 2;
                key1[k] = dmin;
                if (rest > 0) {
                    if (k < MaxFuzzy - 1)
                        key1[k] += rand () % rest;
                    
                    else
                        key1[k] = key[i][k] + rest;
                    rest -= abs ((int) key[i][k] - key1[k]);
                }
            } unsigned int dst =
                computeKeyDistance (key[i], key1, MaxFuzzy);
            if (dst != j) {
                if (!silent)
                    printf
                        ("Internal error: Wrong algorithm creating random key at specified dist!\n");
                return false;
            }
            for (int k = 0; k <= j; k++) {
                StringList res;
                StringList::iterator pos;
                
                    //if (!silent) printf ("Searching [%d,%d] dist=%d\n", key1[0], key1[1], k);
                    if (!kdtree::SearchInFile (fd, key1, k, NULL, &res)) {
                    Perror ("searchInFile - failed");
                    return false;
                };
                bool found = false;
                for (pos = res.begin (); pos != res.end (); pos++) {
                    if (*pos == vals[i]) {
                        found = true;
                    }
                }
                if (k == j && found == false) {
                    if (!silent) {
                        printf
                            ("i=%d, k=%d, j=%d k == j but entry %s not found",
                             i, j, k, vals[i].c_str ());
                    }
                    return false;
                }
            }
        }
    }
    close (fd);
    return true;
}

bool  testSimpleFillPrint (int num) 
{
    const unsigned int MaxFuzzy = 2;
    VectorType     key[num][MaxFuzzy];
    kdtree         tree (MaxFuzzy);
    StringList     vals;

    for (int i = 0; i < num; i++) {
        key[i][0] = i;
        key[i][1] = num - i;
        vals.push_back(Format ("#%d @ 1", i));
        tree.insertNodeIntoTree (key[i], vals.back().c_str(), NULL);
        if (i % 2 == 0) {
            vals.push_back (Format ("#%d @ 2", i));
            tree.insertNodeIntoTree (key[i], vals.back().c_str(), NULL);
        }
        if (i % 3 == 0) {
            vals.push_back (Format ("#%d @ 3", i));
            tree.insertNodeIntoTree (key[i], vals.back().c_str(), NULL);
        }
    }
    int fd =
        open ("test_unbalanced.img", O_RDWR | O_CREAT | O_TRUNC, DefaultAccess);
    if (fd < 0) {
        Perror ("Cannot create file");
        return false;
    }
    if (tree.DumpToFile (fd, 0) < 0) {
        Perror ("Cannot dump to file");
    }
    close (fd);
    tree.printTree ();
    kdtree * newtree = tree.getBalancedCopy ();
    newtree->printTree ();
    fd = open ("test.img", O_RDWR | O_CREAT | O_TRUNC, DefaultAccess);
    if (fd < 0) {
        Perror ("Cannot create file");
        return false;
    }
    if (newtree->DumpToFile (fd, 0) < 0) {
        Perror ("Cannot dump to file");
    }
    close (fd);
    delete newtree;
    return true;
}


#include "cgram.h"
    bool testNgramConv () 
{
    const char *str = "abcdef";
    VectorType vec[6];
    VectorType vc[6] = {
    1, 0, 1, 2, 1, 1};
    StringToNgramVec (str, 1, 6, vec);
    PrintNgramVec (6, vec);
    for (int i = 0; i < 6; i++) {
        if (vec[i] != 1) {
            return false;
        }
    }
    StringToNgramVec (str, 2, 6, vec);
    PrintNgramVec (6, vec);
    for (int i = 0; i < 6; i++) {
        if (vec[i] != vc[i]) {
            return false;
        }
    }
    return true;
}
int  main (void) 
{
    int nTest = 0;
    int nFailed = 0;
    nTest++;
    nFailed += !testNgramConv ();
    nTest++;
    nFailed += !testComputeDist ();
    nTest++;
    nFailed += !testSimpleFillPrint (32);
    nTest++;
    nFailed += !testSimpleSearch (8, 3, true);
    nTest++;
    nFailed += !testSimpleSearch (1000, 10, true);
    nTest++;
    nFailed += !testSearchRandom (5000, 10, 0, true);
    printf ("# of tests %d, failed %d\n", nTest, nFailed);
    return 0;
}


#endif  
