#ifndef KDTREE_H
#define KDTREE_H

#include <list>
#include <vector>
#include <string>
#include <set>
#include <algorithm>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/mman.h>
#include <stdio.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdexcept>
#include "myagrep.h"
#include "interface.h"
#include "cgram.h"
#include <new>
#include <mem_alloc.h>
using namespace std;

typedef vector<double> vd;
typedef vector<int>    vi;

/*
** the basic length of a fuzzy vector (first the key sum, then the trigram frequencies)
*/ 
class kdtree;

struct kdtreeNodeComp  {
    bool operator () (kdtree * a, kdtree * b);
};

typedef vector<kdtree *> NodeSet;
typedef NodeSet::iterator NodeSetIterator;
typedef NodeSet::const_iterator NodeSetConstIterator;

class HyperCube  
{
    protected:
        unsigned int MaxFuzzy;
    public:
        HyperCube (unsigned int MaxFuzzy):MaxFuzzy (MaxFuzzy) {
        loKey = new VectorType[MaxFuzzy];
        hiKey = new VectorType[MaxFuzzy];
        for (unsigned int i = 0; i < MaxFuzzy; i++) {
            loKey[i] = 0, hiKey[i] = VectorTypeMax;
        }
    } 
    ~HyperCube () {
        delete [] loKey;
        delete [] hiKey;
    } 
    unsigned int computeDistance (const VectorType key[]) {
        unsigned res1 = 0;
        unsigned res2 = 0;
        for (unsigned int i = 0; i < MaxFuzzy; i++) {
            if (key[i] < loKey[i])
                res1 += loKey[i] - key[i];
            if (key[i] > hiKey[i])
                res2 += key[i] - hiKey[i];
        }
        return std::max(res1, res2);
    }
    unsigned int size () {
        return MaxFuzzy;
    }
    VectorType * loKey;
    VectorType * hiKey;
};

typedef vector<const char *> ConstCharPtrs;

unsigned int computeKeyDistance (VectorType key1[], VectorType key2[],
                                   unsigned int MaxFuzzy);
class kdtree  
{
  protected:
    unsigned int      MaxFuzzy;
    const static char termChar;
    CFastAllocator &  kdAlloc;
    CFastAllocator &  vecAlloc;
  public:
    ~kdtree ();
    kdtree(CFastAllocator & kdAlloc, CFastAllocator & vecAlloc, unsigned int MaxFuzzy, 
           const char * data, VectorType * vec_p);
    kdtree(unsigned int MaxFuzzy, CFastAllocator & kdAlloc, CFastAllocator & vecAlloc,
#ifdef USE_CHUNK
           const std::string &dataChunk, 
#else
           const ConstCharPtrs data, 
#endif
           VectorType * vec_p = NULL);
    int insertNodeIntoTree (VectorType insKey[], const char * stringData,
                              kdtree ** res_pp);
    int computeKeySum () const;
    unsigned int computeDist (VectorType key[]);
    int getIndexValue (unsigned int inlevel);
    void printNode () const;
    void printKey () const;
    void printTree () const;
    string dataAsString () const;
#ifdef USE_CHUNK
    unsigned int BufLen () {
        unsigned int i = 0, dataSize = dataChunk.size() + sizeof(dataChunkCnt);

        return sizeof (unsigned int) + sizeof(size_t) +
            sizeof (VectorType) * MaxFuzzy + dataSize + 2 * sizeof (off_t);
    }
#else
    unsigned int BufLen () {
        unsigned int i = 0, dataSize = 0;

        for (; i < data.size (); i++)
            //dataSize += data[i]->size () + 1;
            dataSize += strlen(data[i]) + 1;

        return sizeof (unsigned int) + sizeof(size_t) +
            sizeof (VectorType) * MaxFuzzy + dataSize + 2 * sizeof (off_t);
    }
    void Traverse(VectorType key[], unsigned int d,
                    vi &prob, vi &cnt);
#endif
    void Search (VectorType key[], unsigned int maxDist, myagrep &a, 
#ifdef _TEST_KDTREE
                   NodeSet * resList, 
#endif  /*  */
                   StringList * strList, int * numComp = NULL) {
        HyperCube cube (MaxFuzzy);
        kdtree cmptree (kdAlloc, vecAlloc, MaxFuzzy, NULL, key);
        Search (&cmptree, &cube, maxDist, a, 
#ifdef _TEST_KDTREE
                 resList, 
#endif  /*  */
                 strList, numComp);
    } 
    static bool SearchInFile (int fd, VectorType key[],
                                  unsigned int maxDist, myagrep &a,
                                  StringList * resList, int *numComp =
                                  NULL) {
        try {
            return SearchInFileAtLevel(fd, 0, 0, NULL, key, maxDist, a,
                                        resList, numComp);
        } catch (const std::runtime_error & e) {
            std::cerr << "Runtime exception: " << e.what() << std::endl;
        } catch (std::bad_alloc & e) {
            std::cerr << "Not enough memory exception!\n " << std::endl;
        } catch (...) {
            std::cerr << "Unknown exception!\n " << std::endl;
        }
    }
    void inititerate ();
    kdtree * iterate ();
    kdtree * getBalancedCopy (CFastAllocator & newKdAlloc, CFastAllocator & newVecAlloc);
    unsigned int getMaxFuzzy () {
        return MaxFuzzy;
    }
    int DumpToFile (int fd, off_t offset);
  protected:
    void insertNode(kdtree * obj_p, int where, int insertlevel);
    void Search(kdtree * cmpkey, HyperCube * cube, unsigned int maxDist, myagrep &a, 
#ifdef _TEST_KDTREE
                  NodeSet * resList, 
#endif  /*  */
                  StringList * strList, int *numComp);
    static bool SearchInFileAtLevel(int fd, off_t offset,
                                      unsigned int level,
                                      HyperCube * cube, VectorType key[],
                                      unsigned int maxDist, myagrep &a,
                                      StringList * resList, int *numComp);
    void fillInorderTraversal (NodeSet * list);
    void insertnext (const NodeSet & list, int sortlevel);
    VectorType *      key;
#ifdef USE_CHUNK
    string            dataChunk;
    int               dataChunkCnt;
#else
    ConstCharPtrs     data;
#endif
    kdtree *          parent;
    kdtree *          loson;
    kdtree *          hison;
    unsigned int      insertmode;
    unsigned int      leftdepth;
    unsigned int      rightdepth;
    unsigned int      nodelevel;
    unsigned int      level;
    unsigned int      keysum;
};


#endif
