#include "len1indexNoBDB.h"
#include <utility>
#include <sys/timeb.h>
#include "format.h"
#include "int_coder_inl.h"

const char * extDict = NULL;

DEFINE_CLASS_STATIC_COUNTER(CLen1IndexNoBDB, load_list)

bool CLen1IndexNoBDB::OpenDB (const char *name, bool create)
{
    if (opened)
        CloseDB ();

    /* Opening gram list file */
    string fnameGramSorted = string (name);
    if ((fdGram = open(fnameGramSorted.c_str(), ((!create) ? O_RDONLY : O_CREAT | O_TRUNC | O_RDWR), CInterface::defaultAccess)) <= 0) {
        Perror(Format("Error opening main n-gram file %s ", fnameGramSorted.c_str()));
        return false;
    }
    /* Opening file with inverted lists */
    string fnameGramLst = string (name) + ".lst";
    if ((fd = open (fnameGramLst.c_str (),
                    ((!create) ? O_RDONLY : O_CREAT | O_TRUNC | O_RDWR),
                    CInterface::defaultAccess)) <= 0) {
        Perror (Format ("Cannot open file %s", fnameGramLst.c_str ()));
        return false;
    }
    lstStart = NULL;
    mapSize = 0;
    if (!create) {

        gramMapSize  = FileSize(fnameGramSorted.c_str());
        if (gramMapSize < 0) {
            Perror("Cannot determine file size (gram sorted array)");
            goto exit_free;
        }
        gramMapSize = CInterface::getPageAlignedValue(gramMapSize);
        gramFileStart = (size_t *) mmap(NULL, gramMapSize, PROT_READ, MAP_SHARED, fdGram, 0);
        if (gramFileStart == MAP_FAILED) {
            Perror(Format("Map of sorted n-gram file '%s' failed", fnameGramSorted.c_str()));
            goto exit_free;
        }
        gramCnt = *gramFileStart;
        gramArray = (char *) (gramFileStart + 1);
        mapSize = FileSize (fnameGramLst.c_str());

        if (mapSize < 0) {
            Perror("Cannot determine file size (inverted list)");
            goto exit_free;
        }
        mapSize  = CInterface::getPageAlignedValue(mapSize);

        // Sometimes Linux does not map zero-sized files
        if (mapSize) {
            lstStart = (off_t *) mmap (NULL, mapSize, PROT_READ, MAP_SHARED, fd, 0);
            if (lstStart == MAP_FAILED) {
                Perror (Format("Map of inverted list file '%s' failed", fnameGramLst.c_str()));
                goto exit_free;
            }
        } else {
            lstStart = NULL;
        }
    }

    return opened = true;
exit_free:
    if (lstStart) munmap(lstStart, mapSize);
    if (gramFileStart) munmap(gramFileStart, gramMapSize);
    if (fd) close(fd);
    if (fdGram) close(fdGram);
    gramFileStart = NULL;
    lstStart = NULL;
    gramArray = NULL;
    gramCnt = 0;
    opened = false;
    return false;
}

bool CLen1IndexNoBDB::CloseDB ()
{
    bool res = true;
    if (!opened)
        return true;
    if (mapSize != 0) {
        if (munmap(lstStart, mapSize)) {
            Perror ("Cannot unmap (n-gram list)");
            res = false;
        } else {
            lstStart = NULL;
            mapSize = 0;
        }
    }
    if (gramMapSize != 0) {
        if (munmap(gramFileStart, gramMapSize)) {
            Perror("Cannot umamp (sorted n-gram array)");
            res = false;
        } else {
            gramMapSize = 0;
            gramFileStart = NULL;
            gramCnt = 0;
            gramArray = NULL;
        }
    }
    if (fd && close (fd)) {
        Perror ("Cannot close file");
        res = false;
    } else {
        fd = 0;
    }
    if (fdGram && close (fdGram)) {
        Perror ("Cannot close file");
        res = false;
    } else {
        fdGram = 0;
    }
    if (res) {opened = false;}
    return res;
}

void CLen1IndexNoBDB::ResetCounters()
{
    CLen1Index::ResetCounters();
    RESET_COUNTER(load_list);
}

bool CLen1IndexNoBDB::DumpDBCompress(const char *index_name, mOffset & offset_map)
{
    gramCnt = invgram.size();

    if (write(fdGram, &gramCnt, sizeof gramCnt) != sizeof gramCnt) {
        Perror("Cannot write number of n-grams to the file");
        return false;
    }
    off_t       offset = 0;
    char        keyStr[NgramLen + 2];
    CBitWriter  bufferedWriter(1024 * 1024);

    int gramNum = 1;
    double totalLsize = 0;
    double totalOffset = 0;
    for ( mPtrList::iterator it= invgram.begin (); it != invgram.end (); it++, gramNum++) {
        ngram    g(it->first);
        PtrList *plist = invgram[g];
        int      lsize = plist->size ();

        g.toStr(keyStr, NgramLen);

        if (write(fdGram, &keyStr , sizeof keyStr) != sizeof keyStr) {
            Perror("Cannot write to file (n-gram sorted array) (1)");
            return false;
        }
        // Let's write a compressed list
        try {
            if (lsize > 0) {
                vector<off_t> offsets(lsize);
                bufferedWriter.StartOutput(fd);

                PtrListIterator it = plist->begin ();
                for (int i = 0; i < lsize; i++, it++) {
                    offsets[i] = offset_map[*it];
                }
                sort(offsets.begin(), offsets.end());
#if 1
                //cout <<  " Lsize: " << lsize << " avg: " << (offsets[lsize - 1] / double(lsize)) << endl;
                mAvgListSize = (mAvgListSize * (gramNum - 1) + lsize) / gramNum;
                totalOffset += offsets[lsize - 1];
                totalLsize += lsize;
#endif
                off_t last = 0;
                for (int i = 0; i < lsize; ++i) {
                    if (!mUseDelta) {
                        VarByteEncode(1 + offsets[i] - last, bufferedWriter);
                    } else {
                        DeltaEncode(1 + offsets[i] - last, bufferedWriter);
                    }
                    last = offsets[i];
                }

                bufferedWriter.FinishOutput();

            }
        } catch (const std::runtime_error e) {
            cerr << "Error while writing lists: " << e.what() << endl;
            return false;
        }

        off_t newOffset = lseek(fd, 0, SEEK_CUR);

        if (newOffset < 0) {
            Perror("Cannot lseek");
            return false;
        }
        
        EntryDataCompress data(offset, lsize, newOffset - offset);

        offset = newOffset;

        if (write(fdGram, &data , sizeof data) != sizeof data) {
            Perror("Cannot write to file (n-gram sorted array) (2)");
            return false;
        }
    }
    mAvgGapSize = totalOffset / totalLsize;
    return true;
}

bool CLen1IndexNoBDB::DumpDBNoCompress(const char *index_name, mOffset & offset_map)
{
    gramCnt = invgram.size();

    if (write(fdGram, &gramCnt, sizeof gramCnt) != sizeof gramCnt) {
        Perror("Cannot write number of n-grams to the file");
        return false;
    }

    size_t listIndex = 0;
    char   keyStr[NgramLen + 2];

    for (mPtrList::iterator it = invgram.begin (); it != invgram.end (); it++) {
        ngram    g(it->first);
        PtrList *plist = invgram[g];
        int      lsize = plist->size ();

        g.toStr(keyStr, NgramLen);

        if (write(fdGram, &keyStr , sizeof keyStr) != sizeof keyStr) {
            Perror("Cannot write to file (n-gram sorted array) (1)");
            return false;
        }
        
        EntryData data(listIndex, lsize);

        if (write(fdGram, &data , sizeof data) != sizeof data) {
            Perror("Cannot write to file (n-gram sorted array) (2)");
            return false;
        }

        if (lsize > 0) {
            vector<off_t> offsets(lsize);

            PtrListIterator it = plist->begin ();
            for (int i = 0; i < lsize; i++, it++) {
                offsets[i] = offset_map[*it];
            }
            size_t wrSize = offsets.size() * sizeof offsets[0];
            if (write(fd, &offsets[0], wrSize) != wrSize) {
                Perror ("Cannot write to file (n-gram inverted lists)");
                return false;
            }
            listIndex += lsize;
        }
    }

    return true;
}

bool CLen1IndexNoBDB::DumpDB (const char *index_name, mOffset & offset_map)
{
    if (!opened) {
        cerr << index_name << " is not opened!" << endl;
        return false;
    }
    return (mHasCompress) ? DumpDBCompress(index_name, offset_map):DumpDBNoCompress(index_name, offset_map);
}

char *CLen1IndexNoBDB::gram_lower_bound(char * mins) {
    if (!opened) {
        cerr << "Using on closed index in " << __func__ << endl;
        return NULL;
    }
    // This code is taken from STL implemenetation
    char        *__first = gramArray;
    size_t      entrySize = EntrySize();
    size_t      __len = gramCnt;
    size_t      __half;
    char        *__middle;

    while (__len > 0) {
        __half = __len >> 1;
        __middle = __first;
        __middle += __half * entrySize;

        if (strncmp(__middle, mins, NgramLen + 2) < 0) {
            __first = __middle;
            __first += entrySize;
            __len = __len - __half - 1;
        } else {
            __len = __half;
        }
    }
    return __first;

}
 
bool CLen1IndexNoBDB::ReadListDBNoCompress(const char *str, int StrLen, int minpos,
                             int maxpos, OffsetList * ptr)
{
    if (minpos < 0) minpos = 0;
    bool first = true;
    OffsetList pl[2];
    int nc = 0;
    timeb t1, t2;

    assert (str);
    ptr->clear ();

    lstFetched++;

    START_COUNTER(load_list)
    if (NgramLen > StrLen) {
        // create a union of all lists
        // We will iterate through the map using minimum and maximum ngram value
        // and minimum/maximum values of position/length
        char minstr[NgramLen], maxstr[NgramLen];
        char mins[NgramLen + 2], maxs[NgramLen + 2];
        int i = 0;
        // filling common part
        for (; i < StrLen; i++)
            minstr[i] = maxstr[i] = str[i];
        for (; i < NgramLen; i++)
            minstr[i] = 32, maxstr[i] = 255;

        ngram gmin, gmax;
        gmin.pack (minstr, NgramLen, minpos);
        gmax.pack (maxstr, NgramLen, maxpos);
        gmin.toStr (mins, NgramLen);
        gmax.toStr (maxs, NgramLen);

        ftime (&t1);

        
        OffsetList  p;

        size_t      entrySize = EntrySize();
        char *lastPos = END();

        for (char *currPos = gram_lower_bound(mins);currPos != lastPos; currPos += entrySize) {
            if (strncmp (currPos, maxs, NgramLen + 2) > 0) {
                break;
            }
            // Now check whether pos is ok
            int        pos = ngram::PosFromStr(currPos, NgramLen);

            if (pos <= maxpos && pos >= minpos) {
                EntryData *pData = (EntryData *) (currPos + NgramLen + 2); 
#if 1

                // Getting in-list offsets
                off_t *tmpPtr = lstStart + pData->listIndex;

		        unsigned int sz = p.size ();
		        p.resize(sz + pData->cnt);

                for (int j = 0; j < pData->cnt; j++, tmpPtr++) {
		            p[sz + j] = *tmpPtr;
                    recFetched++;
                }
#endif
            }
        }

	    sort(p.begin (), p.end ());
        ftime (&t2);
        fetchTime += 1000 * (t2.time - t1.time) + (t2.millitm - t1.millitm);

        // Copying lists
        *ptr = p;
        STOP_COUNTER(load_list)
        return true;
    } else {
        typedef  DataAndPos<off_t>   PPOS;
        typedef  vector<PPOS>        offList;

        offList                      tmpList[2];

        nc = 0;
        
        // Create an intersection of all lists (< MaxListCount)
        for (int i = 0; i <= min (CLen1Index::MaxIntersectCount, StrLen - NgramLen); i++) {
            char   mins[NgramLen + 2], maxs[NgramLen + 2];
            ngram  gmin, gmax;

            gmin.pack (&str[i], NgramLen, minpos + i);
            gmax.pack (&str[i], NgramLen, maxpos + i);

            gmin.toStr (mins, NgramLen);
            gmax.toStr (maxs, NgramLen);

            ftime (&t1);

            offList p;

            size_t      entrySize = EntrySize();
            char *lastPos = END();

            for (char *currPos = gram_lower_bound(mins);currPos != lastPos; currPos += entrySize) {
                if (strncmp (currPos, maxs, NgramLen + 2) > 0) {
                    break;
                }
        
                // Now check whether pos is ok
                // Do not forget that we are incrementing acceptable pos value by variable i
                int pos = ngram::PosFromStr(currPos, NgramLen);
                if (pos <= maxpos + i && pos >= minpos + i) {
                    EntryData *pData = (EntryData *) (currPos + NgramLen + 2); 
#if 1

                    char tmp[NgramLen + 1];
                    strncpy(tmp, currPos, NgramLen);
                    tmp[NgramLen] = 0;
                    off_t *ptr = lstStart + pData->listIndex;
					unsigned int sz = p.size ();
					//p.resize (sz + cnt);
					p.reserve(sz + pData->cnt);

                    for (int j = 0; j < pData->cnt; j++, ptr++) {
						//p[sz + j] = *ptr;
                        p.push_back(PPOS(*ptr, pos - i));
                        recFetched++;
                    }
#endif
                }
            }
			sort (p.begin (), p.end ());
            ftime (&t2);
            fetchTime += 1000 * (t2.time - t1.time) + (t2.millitm - t1.millitm);
            
            if (first) {
                first = false;
                tmpList[1 - nc] = p;
                nc = 1 - nc;
            } else {
                ListIntersection (tmpList[nc], p, &tmpList[1 - nc]);
                nc = 1 - nc;
            }
            if (tmpList[nc].size() <= 1) {
                break;
            }
        }
        //*ptr = pl[nc];
        ptr->reserve(tmpList[nc].size());
        for (unsigned int k = 0; k < tmpList[nc].size(); ++k)
            ptr->push_back(tmpList[nc][k].mData);

        STOP_COUNTER(load_list)
        return true;

    }
    STOP_COUNTER(load_list)
    cerr << "We mustn't reach this point (internal error)\n";
    return false;
}
bool CLen1IndexNoBDB::ReadListDBCompress(const char *str, int StrLen, int minpos,
                             int maxpos, OffsetList * ptr) 
{
cout << "COMPRESS!" << endl;
    if (minpos < 0) minpos = 0;
    timeb t1, t2;

    assert (str);
    ptr->clear ();

    START_COUNTER(load_list)

    lstFetched++;
    if (NgramLen > StrLen) {
        // create a union of all lists
        // We will iterate through the map using minimum and maximum ngram value
        // and minimum/maximum values of position/length
        char minstr[NgramLen], maxstr[NgramLen];
        char mins[NgramLen + 2], maxs[NgramLen + 2];
        int i = 0;
        // filling common part
        for (; i < StrLen; i++)
            minstr[i] = maxstr[i] = str[i];
        for (; i < NgramLen; i++)
            minstr[i] = 32, maxstr[i] = 255;

        ngram gmin, gmax;
        gmin.pack (minstr, NgramLen, minpos);
        gmax.pack (maxstr, NgramLen, maxpos);
        gmin.toStr (mins, NgramLen);
        gmax.toStr (maxs, NgramLen);

        ftime (&t1);

        
        size_t      entrySize = EntrySize();
        char *lastPos = END();

        for (char *currPos = gram_lower_bound(mins);currPos != lastPos; currPos += entrySize) {
            if (strncmp (currPos, maxs, NgramLen + 2) > 0) {
                break;
            }
            // Now check whether pos is ok
            int        pos = ngram::PosFromStr(currPos, NgramLen);

            if (pos <= maxpos && pos >= minpos) {
                EntryDataCompress *pData = (EntryDataCompress *) (currPos + NgramLen + 2); 
#if 1
                CMemBitReader      reader(((char *)lstStart) + pData->offset, pData->size);

                // Getting compressed in-list offsets
                try {

		            unsigned int sz = ptr->size ();
		            ptr->resize(sz + pData->cnt);

                    off_t last = 0;
                    off_t t;
                    if (!mUseDelta) {
                        for (int j = 0; j < pData->cnt; j++) {
                            VarByteDecode(t, reader);
                            t--;
                            last += t;
		                    (*ptr)[sz + j] = last;
                            recFetched++;
                        }
                    } else {
                        for (int j = 0; j < pData->cnt; j++) {
                            DeltaDecode(t, reader);
                            t--;
                            last += t;
		                    (*ptr)[sz + j] = last;
                            recFetched++;
                        }
                    }
                } catch (const CReachedEof &) {
                    cerr << "unexpected EOF" << endl;
                    STOP_COUNTER(load_list)
                    return false;
                }
#endif
            }
        }

	sort(ptr->begin (), ptr->end ());
        ftime (&t2);
        fetchTime += 1000 * (t2.time - t1.time) + (t2.millitm - t1.millitm);

        STOP_COUNTER(load_list)
        return true;
    } else {
        OffsetList     CurrList, OtherList;
        OffsetList     *pTmpList[2] = {ptr, &OtherList};

        bool first = true;
        int nc = 0;
        
        // Create an intersection of all lists (< MaxListCount)
        for (int i = 0; i <= min (CLen1Index::MaxIntersectCount, StrLen - NgramLen); i++) {
            char   mins[NgramLen + 2], maxs[NgramLen + 2];
            ngram  gmin, gmax;

            gmin.pack (&str[i], NgramLen, minpos + i);
            gmax.pack (&str[i], NgramLen, maxpos + i);

            gmin.toStr (mins, NgramLen);
            gmax.toStr (maxs, NgramLen);

            ftime (&t1);

            OffsetList *pCurr = (first) ? ptr:&CurrList;
            pCurr->clear();

            size_t      entrySize = EntrySize();
            char *lastPos = END();

            for (char *currPos = gram_lower_bound(mins);currPos != lastPos; currPos += entrySize) {
                if (strncmp (currPos, maxs, NgramLen + 2) > 0) {
                    break;
                }
        
                // Now check whether pos is ok
                // Do not forget that we are incrementing acceptable pos value by variable i
                int pos = ngram::PosFromStr(currPos, NgramLen);
                if (pos <= maxpos + i && pos >= minpos + i) {
                    EntryDataCompress *pData = (EntryDataCompress *) (currPos + NgramLen + 2); 

#if 1
                    char tmp[NgramLen + 1];
                    strncpy(tmp, currPos, NgramLen);
                    tmp[NgramLen] = 0;


                    CMemBitReader      reader(((char *)lstStart) + pData->offset, pData->size);

                    // Getting compressed in-list offsets
                    try {
					    unsigned int sz = pCurr->size ();
					    pCurr->reserve(sz + pData->cnt);
                        off_t last = 0;
                        off_t t;
    
                        if (!mUseDelta) {
                            for (int j = 0; j < pData->cnt; j++) {
                                VarByteDecode(t, reader);
                                t--;
                                last += t;
						
                                pCurr->push_back(last);
                                recFetched++;
                            }
                        } else {
                            for (int j = 0; j < pData->cnt; j++) {
                                DeltaDecode(t, reader);
                                t--;
                                last += t;
						
                                pCurr->push_back(last);
                                recFetched++;
                            }
                        }
                    } catch (const CReachedEof &) {
                        cerr << "unexpected EOF" << endl;
                        STOP_COUNTER(load_list)
                        return false;
                    }
#endif
                }
            }
	    sort (pCurr->begin (), pCurr->end ());
            pCurr->erase(unique(pCurr->begin(), pCurr->end()), pCurr->end()); // Remove possibly repeating strings
            ftime (&t2);
            fetchTime += 1000 * (t2.time - t1.time) + (t2.millitm - t1.millitm);

            if (first) {
                first = false;
            } else {
                ListIntersection(*pCurr, *pTmpList[nc], pTmpList[1 - nc]);
                nc = 1 - nc;
            }
            if (pTmpList[nc]->size() <= 1) {
                break;
            }
        }
        if (pTmpList[nc] != ptr) {ptr->swap(*pTmpList[nc]);}

        STOP_COUNTER(load_list)
        return true;

    }
    cerr << "We mustn't reach this point (internal error)\n";
    STOP_COUNTER(load_list)
    return false;
}

bool CLen1IndexNoBDB::ReadListDB(const char *str, int StrLen, int minpos,
                             int maxpos, OffsetList * ptr)
{
cout << "READLISTDB: " << mHasCompress << endl;
    return (mHasCompress) ? ReadListDBCompress(str, StrLen, minpos, maxpos, ptr):
                            ReadListDBNoCompress(str, StrLen, minpos, maxpos, ptr);
}

void CLen1IndexNoBDB::PrintAddStat(int testNum, int kerr, ostream & ostr)
{
    if (testNum <= 0) return;
    ostr << "kerr = " << kerr << " " <<
            "Time to fetch lists (ms): " << GET_COUNTER_MS(load_list) /testNum  <<  " "
            "Time to check strings (ms): " << GET_COUNTER_MS(check_string)  /testNum << "\n";

    ostr << "Avg gap (offset in bytes): " << mAvgGapSize << " Avg list size: " << mAvgListSize << endl;
}



