#ifndef _COMMON_H
#define _COMMON_H

#include <sys/mman.h>
#include <sys/types.h>
#include <fcntl.h>
#include <sys/time.h>
#include <time.h>
#include <unistd.h>
#include <stdio.h>
#include <filesize.h>
#include <mcheck.h>
#include <math.h>
#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#include <string.h>

#include <iostream>
#include <iomanip>
#include <fstream>
#include <sstream>
#include <algorithm>

#include <vector>
#include <string>

#include "hashport.h"
#include "format.h"

using namespace std;


#include "afuzzy.h"
#include "utils.h"

#include <alphabet.h>

// For density statistics
#define MAX_RESULT_SIZE 20

typedef vector <string> StringList;

#define DECLARE_COUNTER(name)        mutable long long COUNTER##name;  \
                                     mutable timeval COUNTERt1##name;  \
                                     mutable struct timezone COUNTERtz##name;
#define DECLARE_STATIC_COUNTER(name) static long long COUNTER##name; \
                                     static timeval COUNTERt1##name;  \
                                     static struct timezone COUNTERtz##name;

#define DEFINE_CLASS_STATIC_COUNTER(className, name) \
                                     long long className :: COUNTER##name; \
                                     timeval className :: COUNTERt1##name; \
                                     struct timezone className :: COUNTERtz##name;

#define RESET_COUNTER(name)   COUNTER##name = 0;
#define START_COUNTER(name)          gettimeofday (&COUNTERt1##name, &COUNTERtz##name);
#define STOP_COUNTER(name)           {timeval t2; gettimeofday(&t2, &COUNTERtz##name); \
                                     COUNTER##name += (t2.tv_sec - COUNTERt1##name.tv_sec) * 1000000LL + \
                                     t2.tv_usec - COUNTERt1##name.tv_usec;}
// absolute counter value
#define GET_COUNTER(name)            (COUNTER##name)
// approximate in ms
#define GET_COUNTER_MS(name)         (COUNTER##name/1e3)
#define GET_COUNTER_SEC(name)        int(COUNTER##name / 1000000)
#define GET_COUNTER_USEC(name)       int(COUNTER##name % 1000000)

#define ENV_ABC "ABC"
#define ENV_KERR "MAX_KERR"
#define RANDOM_EDIT_PRECISION_VAR  100000

void FillExpansionList (string str,
                        bool hasTransp,
                        int kerr,
                        StringList * lst,
                        const char *ABC,
                        int minKerrLimited, bool limited, 
                        int start, bool showVariants);

struct CProbDesc {
    float mProbIns, mProbDel, mProbSubst, mProbTransp;
    CProbDesc(float ProbIns, float ProbDel, float ProbSubst, float ProbTransp) :
              mProbIns(ProbIns), mProbDel(ProbDel), mProbSubst(ProbSubst), mProbTransp(ProbTransp) {}
};

string ApplyRandomEdits(string OrigStr,
                        bool hasTransp,
                        int kerr,
                        const char *pABC,
                        unsigned ABCLen,
                        int start,
                        const CProbDesc& ProbDesc);
enum AlgoType {
    FILTER_NONE = 0,   // no filter is used
    FILTER_EXACT,  // exact multi-pattern algorithm is used to filter with delayed check
    FILTER_EXACT_IMMEDIATE, // exact multi-pattern (shift-or-based) algorithm is used to filter but the check is immediate
    FILTER_AGREP, // genuine agrep filter
    FILTER_COMBINED, // for k = 1 FILTER_AGREP otherwise FILTER_EXACT
};

#if 1
    inline void debugFunc(){
        cout << "here\n";
    }
#endif

template <typename CIndexType>
inline int MaxTestKerr(void) {
    int kerr = MAX_KERR;
    GetEnv(ENV_KERR, kerr);
    return min(kerr, int(MAX_KERR));
}

template <typename CIndexType>
inline bool SearchDbNotSupported(void) {
    return false;
}

class CInterface {
  protected:
    bool         mHasTransp;
    EAlphabet    mAlphaType;
    CInterface(bool hasTransp, EAlphabet AlphaType) : mHasTransp(hasTransp), mAlphaType(AlphaType) {}
  public:
    const static AlgoType filterType = FILTER_COMBINED;
    virtual bool Insert(const char *str) = 0;
    virtual bool InsertList(const StringList &);
    virtual bool Search(const char *str, int kerr, int *numComp,
                         StringList * resList, bool silent = true) = 0;
    virtual bool SearchInDB(const char *str, int kerr, int *numComp,
                             StringList * resList, bool silent = true) = 0;
    virtual bool DumpDB(const char *baseName) = 0;
    virtual bool OpenDB(const char *baseName, bool create = false) = 0;
    virtual bool CloseDB() = 0;
    virtual bool CreateInMemIndex(const char *inputFile, int insertCount, int *dcnt, bool silent = true);
    DECLARE_STATIC_COUNTER(check_string);
    virtual void DisplayStat() {
        cerr << "Stub for DisplayStat!" << endl;
    } virtual void Print() {
        cerr << "Stub for Print!" << endl;
    }
    static const int defaultHashRatio = 40; // this is a hint only
    // the actual fill ratio can be less
    static const int minHashRatio = 10; 
    static const int maxHashRatio = 90;
    static const int defaultAccess = 0700;
    template <class T>
    static
    T getPageAlignedValue(T val) {
        static size_t pg = getpagesize();
        return val + (pg - val % pg) % pg;
    }
    enum SearchType {
       SearchTypeMem,
       SearchTypeDisk,
       SearchTypeFlash
    };
    virtual void ResetCounters(){}
    virtual void PrintAddStat(SearchType type, int testNum, int kerr, ostream &){}
};

void toUpper (string & s);

/* This function gets the array of arguments 		*/
/* First argument is the query test collection     	*/
/* The next arguments are the names of the dictionary 	*/
/* test collections. Procedure performs tests for every */
/* test data collection. It creates an index in memory  */
/* performs test, outputs an average search time.       */
/* After that it dumps the content to the disk and      */
/* measure times clearing cash before every test        */
/* The output is n strings, where N is number of tests. */
/* files. */

template <typename CIndexType>
    int test(int argc, char *argv[], string baseName,
              StringList cldevList, int insertCount, bool silent,
              int testCutoff = 1000000)
{
    if (argc < 4) {
        cerr << "Not enough arguments\n";
        return 0;
    }
    try {
    timeval             t1, t2;
    struct timezone     _tz;
    int                 dcnt = 0;
    const int           maxKerr = MaxTestKerr<CIndexType>();
    bool                hasTransp = atol(argv[1]);
    string              s;
    GetEnv(ENV_ABC, s);
    EAlphabet           AlphaType = GetAlphabetType(s);
    if (AlphaType == eAlphaUnknown) {
        cerr << "Unknown alphabet: '" << s << "' check env var " << ENV_ABC << endl;
        return 0;
    }
    cout << "Alphabet type: " << s << "(" << AlphaType << ")" << endl;
    bool                noDumpDb = GetEnvSwitch("NO_DUMP_DB");

    bool                printPct = (GetEnvSwitch("NO_PRINT_PCT")) ? false:true;
// First of all we read the test files
    bool                showVariants = GetEnvSwitch("SHOW_VARIANTS");
    bool                skipFuncTest = GetEnvSwitch("SKIP_FUNC_TEST");
    bool                skipMem = GetEnvSwitch("SKIP_MEM");
    bool                skipDisk = GetEnvSwitch("SKIP_DISK") || SearchDbNotSupported<CIndexType>();
    
    int MaxFuncTestSize = MAX_FUNC_TEST_SIZE;
    GetEnv("MAX_FUNC_TEST_SIZE", MaxFuncTestSize);

    bool                skipFlash = true;
    string              flashDir;
    string              diskDir;
    string              resFile = "results";

    GetEnv("RES_FILE", resFile); // does not change if unset

    GetEnv("DISK_DIR", diskDir); // does not change if unset
    if (GetEnv("FLASH_DIR", flashDir) && !SearchDbNotSupported<CIndexType>()) {
        skipFlash = false;
    }
    if (noDumpDb) {
        skipFlash = skipDisk = true;
    }
    
    bool                noClearCache = GetEnvSwitch("NO_CLEAR_CACHE");
    bool                noPurgeFile = GetEnvSwitch("NO_PURGE_FILE");
    bool                printAddStat = GetEnvSwitch("PRINT_ADD_STAT");
    int                     maxDiskTest = INT_MAX;
    GetEnv("MAX_DISC_TEST", maxDiskTest);
    int                      maxFlashTest = INT_MAX;
    GetEnv("MAX_FLASH_TEST", maxFlashTest);
    bool                testInteract = GetEnvSwitch("TEST_INTERACT");

    if (!skipFuncTest) {
           
        StringList funcTestList;
        {
            ifstream TestFile1(argv[3]);

            if (!TestFile1) {
                Perror ("Cannot open the func test file");
                return 0;
            }

            while (getline(TestFile1, s)) {
                toUpper (s);
                funcTestList.push_back (s);
            }
        }

        CIndexType ind(hasTransp, AlphaType);
        if (dynamic_cast < CInterface * >(&ind) == NULL) {
            cerr <<
                "Invalid class type. It must be inherited from CInterface!"
                << endl;
            return 0;
        }
        cout<<"Creating index with baseName = " << baseName << endl;
        if (!ind.CreateInMemIndex (argv[3], insertCount, &dcnt, silent)) {
            cerr << "Cannot create index from " << argv[3] <<
                " functional test failed " << endl;
            return 0;
        }
        //ind.Print ();
        //cout<<" done"<<endl;
        // Now testing 
        int tcnt1 = 0, tcnt2 = 0;
        int succ1 = 0, succ2 = 0;

        if (!ind.DumpDB(baseName.c_str())) {
            Perror ("Cannot dump database");
            return 0;
        }

        if (!ind.CloseDB()) {
            Perror ("Cannot close database");
            return 0;
        }

        if (!ind.OpenDB (baseName.c_str ())) {
            Perror ("Cannot open database");
            return 0;
        }

        int funcTestNum = std::min(funcTestList.size(), (size_t) MaxFuncTestSize);
        int funcTestCnt = 0;
        int prevPct = 0;

        int k1 = 0, k2 = maxKerr;

        int kFuncTest = -1;
        GetEnv("FUNC_TEST_KERR", kFuncTest);

        if (kFuncTest > 0) { k1 = k2 = kFuncTest; }

        for (int kerr = k1; kerr <= k2; kerr++) {
            for (int i = 0; i < funcTestNum; i++) {
                string s0 = funcTestList[i];

                StringList lst;
                //cout<<" Filling expansion list for "<<s0<<" kerr = "<<kerr<<" ...";
                ++funcTestCnt;
                int pct = funcTestCnt * 100 / ((1 + k2 - k1) * funcTestNum);
                if (printPct) {
                    if (pct > prevPct || i == funcTestNum - 1) {
                        stringstream str;

                        prevPct = pct;
                        str << "Func test progress: " << pct << "% (kerr = " << kerr << ")";
                        string mess = str.str(); 
                        cout << mess;
                        for(unsigned int l = 0; l < mess.size() + 5; ++l)
                                cout << "\b\r";
                        cout.flush ();
                    }
                }
                FillExpansionList (s0, hasTransp, kerr, &lst, GetAlphabet(AlphaType), 
                                   kerr, kerr > 1, // Limited neighborhood for k > 1
                                   0,
                                   false);

#if 0
                cout<<" Fix exp size "<<lst.size ()<<endl;
#endif

                if (kerr > 1 && lst.size () > testCutoff) {
                    lst.erase (lst.begin () + testCutoff, lst.end ());
                }
                //cout<<" final size: "<<lst.size ()<<endl;
                sort(lst.begin(), lst.end());
                lst.erase(unique(lst.begin(), lst.end()), lst.end());

                for (StringList::iterator it = lst.begin ();
                     it != lst.end (); it++) {
                    const string & s = *it;
                    StringList res;


                    if (!skipMem) {
                        if (showVariants) {
                            cout << "Variant: " << s << " kerr = " << kerr << endl;
                        }
#if 0
                        if (s == "DECL'MOSE" && kerr == 2) {
                            debugFunc();
                        }
#endif

                        tcnt1++;
                        ind.Search(s.c_str (), kerr, NULL, &res, silent);
                        int f = 0;
                        for (int j = 0; j < res.size (); j++) {
                                if (res[j] == s0) {
                                f = 1;
                                break;
                                }
                        }
                        succ1 += f;

                        if (f == 0) {
                                cerr << "mem: '" << s0 << "' by '" << s <<
                                "' kerr = " << kerr << " unfound" << endl; 
                        }
                        res.clear ();
                    }
                    if (!skipDisk)  {
                        tcnt2++;
                        if (!ind.SearchInDB (s.c_str (), kerr, NULL, &res,
                                    silent)) {
                                Perror ("Error while searching!");
                                return 0;
                        };
                        int f = 0;
                        for (int j = 0; j < res.size (); j++) {
                            if (res[j] == s0) {
                                f = 1;
                                break;
                            }
                        }
                        if (f == 0) {
                            cerr << "disk: '" << s0 << "' by '" << s <<
                                "' kerr = " << kerr << " unfound" << endl;
                        }
                        succ2 += f;
                    }
                }
            }
        }
        if (!ind.CloseDB ()) {
            Perror ("Cannot close database");
            return 0;
        }
        bool OkFlag = (succ1 == tcnt1 && succ2 == tcnt2);
        cout << endl;
        cout << "Functional test result for " << (tcnt1 + tcnt2) << " queries :" <<
            ((OkFlag) ? "Ok" : "with errors") << endl;
        if (!OkFlag) {
            cout << "Memory search succeeded " << succ1 << " of " << tcnt1
                << endl;
            cout << "Disc search succeded " << succ2 << " of " << tcnt2 <<
                endl;
            return 0;
        }
        cout.flush ();
    }

    if (argc < 5)
        return 0;               // Stoped by functional test

    // Now we must have dictCnt * (2 + MAX_KERR) test files
    // Exactly dictCnt test sets. In every testSet, the first
    // MAX_KERR are the test files (kerr = 0, 1, ...)
    // The last one is dict
    int dictCnt = (argc - 5 + 1) / (2 + MAX_KERR);

    if (argc > 4 && (argc - 5 + 1) % (2 + MAX_KERR)) {
        cerr << "You must have " << (dictCnt * (2 + MAX_KERR) + 3) << " input arguments, that is " << 
                dictCnt << " test sets, each test set having " << (2 + MAX_KERR) << " files\n";
        return 0;
    }

    char curr_dir[PATH_MAX + 1];
    if  (!getcwd (curr_dir, PATH_MAX + 1)) {
        cerr << "Cannot get the current directory" << endl;
        return 0;
    }
    memset (&_tz, 0, sizeof _tz);

    string cdir(curr_dir);
    string methName;
    int    pos = cdir.find_last_of("/");
    if (pos == string::npos) {
        methName = cdir;
    } else {
        methName = cdir.substr(pos + 1);
    }

#if __GNUC__ > 2
    ofstream odata((string("../") + resFile).c_str(), ios_base::app);
#else
    ofstream odata((string("../") + resFile).c_str(), fstream::app);
#endif

    int Sigma = 0;

#ifdef ALPHABET_SIZE
    Sigma = ALPHABET_SIZE;
#endif

    GetEnv("ALPHABET_SIZE", Sigma);

    // Now processing different dictionary sets
    odata << methName << " hasTransp=" << hasTransp << " Sigma=" << Sigma << " kerr=" << MaxTestKerr<CIndexType>() << std::endl;

    for (int i = 0; i < dictCnt; i++) {
        StringList TestList[MAX_KERR + 1];
        int curDictNum = 5 + MAX_KERR + i * (2 + MAX_KERR);

#if 1
        cout << "curDictNum = " << curDictNum << " Dict file: " << argv[curDictNum] << endl;
#endif
        for (int k = 0; k <= MAX_KERR; ++k) {
            int curNum = 4 + k + i * (2 + MAX_KERR);
            ifstream TestFile(argv[curNum]);

#if 1
            cout << "Test file: " << argv[curNum] << endl;
#endif

            if (!TestFile) {
                std::string s = "Cannot open test file '";

                s += argv[curNum];
                s += "'";

                Perror (s);
                return 0;
            }

            while (getline(TestFile, s)) {
                toUpper (s);
                TestList[k].push_back (s);
            }
        }

        cout << "skipDisk " << skipDisk << 
                " skipFlash " << skipFlash << 
                " noClearCache " << noClearCache << 
                " noPurgeFile " << noPurgeFile << endl;

        CIndexType ind(hasTransp, AlphaType);
        if (dynamic_cast < CInterface * >(&ind) == NULL) {
            cerr <<
                "Invalid class type. It must be inherited from CInterface!"
                << endl;
            return 0;
        }
        string InputFile = argv[curDictNum];
        if (!ind.CreateInMemIndex(InputFile.c_str(), insertCount, &dcnt, silent)) {
            cerr << "Cannot create index from " << InputFile << endl;
            continue;
        }
        if (testInteract) {
            while (true) {
                string pat;
                int    kerr;
                cout << "Input pattern or Ctrl + D to continue, then enter kerr:" << endl;
                cin >> pat ;
                if (cin.eof() || pat.empty()) {
                    break;
                }
                cin >> kerr;
                if (cin.eof()) {
                    break;
                }
                StringList res;
                DECLARE_STATIC_COUNTER(SearchTime);
                RESET_COUNTER(SearchTime);
                START_COUNTER(SearchTime);
                int NumComp = 0;
                ind.Search(pat.c_str (), kerr, &NumComp, &res, true);
                STOP_COUNTER(SearchTime);

                for (int i = 0; i < res.size(); ++i) {
                    cout << res[i] << endl;
                }
                cout << "Found " << res.size() << " entry(ies), in " << GET_COUNTER_MS(SearchTime) << " ms " << 
                        "Number of comparisons: " << NumComp << endl;
            }
        }

        double time_mem[MAX_KERR + 1], dev_mem[MAX_KERR + 1];
        double time_disk[MAX_KERR + 1], dev_disk[MAX_KERR + 1];
        double time_flash[MAX_KERR + 1], dev_flash[MAX_KERR + 1];
        stringstream add_stat[MAX_KERR + 1];
        int num_comp_mem[MAX_KERR + 1], num_comp_disk[MAX_KERR + 1], num_comp_flash[MAX_KERR + 1];

        for (int kerr = 0; kerr <= MAX_KERR; kerr++) {
            time_mem[kerr] = time_disk[kerr] = time_flash[kerr] = -1;
            dev_mem[kerr]  = dev_disk[kerr]  = dev_flash[kerr] = -1;
            num_comp_mem[kerr] = num_comp_disk[kerr] = num_comp_flash[kerr] = -1;
        }

        int prevPct = 0;
        int cnt = 0;

        int TotalTestSize = 0;

        for (int kerr = 0; !skipMem && kerr <= maxKerr; kerr++) {
            TotalTestSize += TestList[kerr].size();
        }

        vector<size_t>             maxRes(MAX_KERR + 1);
        vector<size_t>             maxResNoPattern(MAX_KERR + 1);
        vector<double>             avgRes(MAX_KERR + 1);
        vector<double>             avgResNoPattern(MAX_KERR + 1);

        // In-memory search
        for (int kerr = 0; !skipMem && kerr <= maxKerr; kerr++) {
            size_t           testSize = TestList[kerr].size ();
            vector<double>   timeDat(testSize);
            vector<size_t>   cntRes;
            vector<size_t>   cntResNoPattern;

            long long compCnt = 0;
            time_mem[kerr] = 0;
            // Searching all test patterns
            int currTestSize = testSize;
            ind.ResetCounters();
            for (size_t k = 0; k < currTestSize; k++) {
                StringList      tmpRes;

                int _cnt = 0;
                gettimeofday (&t1, &_tz);
                const string &pattern = TestList[kerr][k];
                ind.Search(pattern.c_str (), kerr, &_cnt, &tmpRes, silent);
                gettimeofday (&t2, &_tz);
                compCnt = compCnt + _cnt;
                // Remove duplicates
                sort(tmpRes.begin(), tmpRes.end());
                tmpRes.erase(unique(tmpRes.begin(), tmpRes.end()), tmpRes.end());
                size_t ResultSize = tmpRes.size();
                avgRes[kerr] = (avgRes[kerr] * k + ResultSize) / double(k + 1);
                maxRes[kerr] = max(maxRes[kerr], ResultSize);
                if (ResultSize > MAX_RESULT_SIZE) ResultSize = MAX_RESULT_SIZE;
                if (cntRes.size() < ResultSize + 1) cntRes.resize(ResultSize + 1);
                cntRes[ResultSize]++;

                // Exclude pattern from result size
                StringList::iterator it = find(tmpRes.begin(), tmpRes.end(), pattern);

                if (it != tmpRes.end()) { tmpRes.erase(it); }

                size_t ResultSizeNoPattern = tmpRes.size();

                maxResNoPattern[kerr] = max(maxResNoPattern[kerr], ResultSizeNoPattern);
                avgResNoPattern[kerr] = (avgResNoPattern[kerr] * k + ResultSizeNoPattern) / double(k + 1);

                if (ResultSizeNoPattern > MAX_RESULT_SIZE) ResultSizeNoPattern = MAX_RESULT_SIZE;

                if (cntResNoPattern.size() < ResultSizeNoPattern + 1) cntResNoPattern.resize(ResultSizeNoPattern + 1);
                cntResNoPattern[ResultSizeNoPattern] ++; 

                timeDat[k] = ((t2.tv_sec - t1.tv_sec) * 1000000 + t2.tv_usec -
                     t1.tv_usec) / 1000.0;
                time_mem[kerr] += timeDat[k];
#if 0
                if (kerr > 2) {
                    cout << pattern << " time: " << timeDat[k] << " ms, kerr = " << kerr << endl;
                }
#endif
                ++cnt;
                int pct = cnt * 100 / TotalTestSize;
                if (printPct) {
                    if (pct > prevPct || k == currTestSize - 1) {
                        stringstream str;

                        prevPct = pct;
                        str << "In-memory test progress: " << pct << "% (kerr = " << kerr << ")";
                        string mess = str.str(); 
                        cout << mess;
                        for(unsigned int l = 0; l < mess.size() + 5; ++l)
                                cout << "\b\r";
                        cout.flush();
                    }
                }
            }
            time_mem[kerr] /= currTestSize;
            dev_mem[kerr] = 0;
            double avg = time_mem[kerr];
            for (size_t k = 0; k < currTestSize; k++) {
                double v = timeDat[k] - avg;
                dev_mem[kerr] += v * v;
            }
            if (printAddStat) {
                ind.PrintAddStat(CInterface::SearchTypeMem, currTestSize, kerr, add_stat[kerr]);
            }

            dev_mem[kerr] = sqrt (dev_mem[kerr] / currTestSize);
            num_comp_mem[kerr] = compCnt / currTestSize;
            add_stat[kerr] << "Kerr: " << kerr << endl;
            if (printAddStat) {
                int sum = 0;
                add_stat[kerr] << "Info for result set that might include the pattern:\n";
                add_stat[kerr] << " Avg result set size: " << avgRes[kerr] << " Max result set size: " << maxRes[kerr] << endl;
                for (int j = 1; j < cntRes.size(); ++j) {
                    add_stat[kerr] << j << ": ";
                    add_stat[kerr] << (100.0 * cntRes[j] / currTestSize) << "% ";
                    if (j >= 1) sum += cntRes[j];
                }
                add_stat[kerr] << " Total pct with neighbors: " << (sum * 100.0 / currTestSize) << " (" << currTestSize << " tests) " << endl;
                int sumNoPattern = 0;
                add_stat[kerr] << "Info for result set that NEVER includes the pattern:\n";
                add_stat[kerr] << " Avg result set size: " << avgResNoPattern[kerr] << " Max result set size: " << maxResNoPattern[kerr] << endl;
                for (int j = 1; j < cntResNoPattern.size(); ++j) {
                    add_stat[kerr] << j << ": ";
                    add_stat[kerr] << (100.0 * cntResNoPattern[j] / currTestSize) << "% ";
                    if (j >= 1) sumNoPattern += cntResNoPattern[j];
                }
                add_stat[kerr] << " Total pct with neighbors: " << (sumNoPattern * 100.0 / currTestSize) << " (" << currTestSize << " tests) " << endl;
            
            }
        }
        cout << endl;

        long long db_size = 0;

        if (!noDumpDb) {
            // Dump index to the disc. It is important no to do it before in-memory test,
            // because dump consumes memory.
            if (!ind.DumpDB(baseName.c_str())) {
                cerr << "Cannot dump DB\n";
                return 0;
            }


            if (!ind.CloseDB()) {
                cerr << "Cannot close DB\n";
                return 0;
            }
        
            // If we do not sync properly, later OS might sync
            // while we are searching and make searching slower
            for (int i = 0; i < 2; ++i) {
                (void)system("sync");
                sleep(2);
                (void)system("sync");
                sleep(2);
            }

            // Now, let's determine the size of the database
            for (size_t l = 0; l < cldevList.size (); l++) {
                off_t tmp_size = FileSize (cldevList[l].c_str ());
                if (tmp_size < 0) {
                    cerr << "Cannot determine file size (" << cldevList[l] <<
                        ")\n";
                } else
                    db_size += tmp_size;
            }
        }
        // Important note! only read cashes are purged correctly by the 
        // clearing driver. write cashes are not cleared at all on kernel
        // 2.6 and only about at half on 2.4. reading big file (but just once)
        // as big as the total memory size purges all writing cashes


        if (!diskDir.empty()) {
             // copy index to a new place change directory
             if (chdir(diskDir.c_str())) {
                 Perror(Format("Cannot change to '%s'", diskDir.c_str()));
                 return 0;
             }
             // Now copy
             for (size_t l = 0; l < cldevList.size (); l++) {
                 string cmd = Format("cp %s/%s .", curr_dir, cldevList[l].c_str());
                 if (system(cmd.c_str())) {
                     Perror(Format("Cannot copy %s/%s to %s", curr_dir, cldevList[l].c_str(), diskDir.c_str()));
                     return 0;
                 }
             }
        }
        char cdir[PATH_MAX + 1];
        if  (!getcwd (cdir, PATH_MAX + 1)) {
            cerr << "Cannot get the current directory" << endl;
            return 0;
        }
        cout << "Before disk current dir is: " << cdir << endl;
        // Read purge file after creating copy on disk!!! 
         
        long sum = 0;
        // Disc or flash search. Before each search we clear the file cash
        // timings are measured individually for every search procedure

        if (!skipDisk && !noClearCache && !noPurgeFile) {
            char rbuf[65536];
	    for (unsigned int ii = 0; ii < 1; ++ii) {

            int fdp = open(argv[2], O_RDONLY);
            if (fdp < 0) {
                Perror ("Cannot open purge file!");
                return 0;
            }
            errno = 0;
            int ll;
            while(ll = read(fdp, rbuf, sizeof rbuf) > 0) {
                sum += ll;
#if 0
                cout << "Reading purge file" << endl;
#endif
            }
            if (errno) {
                Perror("Error reading purge file!");
                return 0;
            }
            close(fdp);
	    }
        }

        prevPct = 0;
        cnt = 0;

        // Disk search

        TotalTestSize = 0;

        for (int kerr = 0; !skipMem && kerr <= maxKerr; kerr++) {
            TotalTestSize += std::min((int)TestList[kerr].size (), maxDiskTest);
        }

        for (int kerr = 0;  !skipDisk && kerr <= maxKerr; kerr++) {
            int           testSize = std::min((int)TestList[kerr].size (), maxDiskTest);

            vector<double>   timeDat(testSize);

            // Searching all test patterns
            double all_tm = 0;
            long long compCnt = 0;

            int currTestSize = testSize;
#if 0
            cout << "Curr test size: " << currTestSize << endl;
#endif

            ind.ResetCounters();
            size_t k = 0;
            for (; k < currTestSize; k++) {

                if (!noClearCache) {
                    FILE *f0 = fopen (DEV_CLDEV, "w");
                    if (!f0) {
                        Perror ("Cannot open cldev");
                        return 0;
                    }
                    // Let's clear all cashes
                    for (size_t l = 0; l < cldevList.size (); l++) {
                        if (fputs((string(cdir) + "/" + cldevList[l] +
                            "\n").c_str (), f0) < 0) {
                            Perror ("Cannot write to cldev");
                            return 0;
                        }
                    }
                    if (fclose (f0)) {
                        Perror ("Cannot close cldev");
                        return 0;
                    }
                }
                if (!ind.OpenDB (baseName.c_str ())) {
                    Perror ("Cannot open database");
                    return 0;
                }
                StringList   tmpRes;

                int _cnt = 0;
                gettimeofday (&t1, &_tz);
                if (!ind.SearchInDB (TestList[kerr][k].c_str (), kerr, &_cnt, &tmpRes,
                                silent)) {
                    Perror ("Error while searching!");
                    return 0;
                };
                gettimeofday (&t2, &_tz);
                compCnt = compCnt + _cnt;
                timeDat[k] =
                    ((t2.tv_sec - t1.tv_sec) * 1000000 + t2.tv_usec -
                     t1.tv_usec) / 1000.0;

                all_tm += timeDat[k];
                if (!ind.CloseDB ()) {
                    Perror ("Cannot close database");
                    return 0;
                }
#if 0
                cout << timeDat[k] << " kerrr = " << kerr << " cnt = " << _cnt << endl;
#endif
                ++cnt;
                int pct = cnt * 100 / TotalTestSize;
                if (printPct) {
                    if (pct > prevPct || k == currTestSize - 1) {
                        stringstream str;

                        prevPct = pct;
                        str << "Disc test progress: " << pct << "% (kerr = " << kerr << ")";
                        string mess = str.str(); 
                        cout << mess;
                        for(unsigned int l = 0; l < mess.size() + 5; ++l)
                                cout << "\b\r";
                        cout.flush();
                    }
                }
            }
            time_disk[kerr] = all_tm / currTestSize;
            num_comp_disk[kerr] = compCnt / currTestSize;
            double avg = time_disk[kerr];
            dev_disk[kerr] = 0;
            for (size_t k = 0; k < currTestSize; k++) {
                double v = timeDat[k] - avg;
                dev_disk[kerr] += v * v;
            }
            if (printAddStat) {
                ind.PrintAddStat(CInterface::SearchTypeDisk, currTestSize, kerr, add_stat[kerr]);
            }

            dev_disk[kerr] = sqrt (dev_disk[kerr] / currTestSize);
        }

        // Flash search

        if (!skipFlash) {
             // copy index to a new place change directory
             if (chdir(flashDir.c_str())) {
                 Perror(Format("Cannot change to '%s'", flashDir.c_str()));
                 return 0;
             }
             // Now copy
             for (size_t l = 0; l < cldevList.size (); l++) {
                 string cmd = Format("cp %s/%s .", curr_dir, cldevList[l].c_str());
                 if (system(cmd.c_str())) {
                     Perror(Format("Cannot copy %s/%s to %s", curr_dir, cldevList[l].c_str(), flashDir.c_str()));
                     return 0;
                 }
             }
        }
        if  (!getcwd (cdir, PATH_MAX + 1)) {
            cerr << "Cannot get the current directory" << endl;
            return 0;
        }
        cout << "Before flash current dir is: " << cdir << endl;
        // Read purge file after creating copy on flash disk!!! 
        if (!skipFlash && !noClearCache && !noPurgeFile) {
            char rbuf[65536];
	    for (unsigned int ii = 0; ii < 1; ++ii) {

            int fdp = open(argv[2], O_RDONLY);
            if (fdp < 0) {
                Perror ("Cannot open purge file!");
                return 0;
            }
            errno = 0;
            int ll;
            while(ll = read(fdp, rbuf, sizeof rbuf) > 0) {
                sum += ll;
#if 0
                cout << "Reading purge file" << endl;
#endif
            }
            if (errno) {
                Perror("Error reading purge file!");
                return 0;
            }
            close(fdp);
	    }
        }

        prevPct = 0;
        cnt = 0;

        TotalTestSize = 0;

        for (int kerr = 0; !skipMem && kerr <= maxKerr; kerr++) {
            TotalTestSize += std::min((int)TestList[kerr].size (), maxFlashTest);
        }

        for (int kerr = 0;  !skipFlash && kerr <= maxKerr; kerr++) {
            int           testSize = std::min((int)TestList[kerr].size (), maxFlashTest);
            vector<double>   timeDat(testSize);

            // Searching all test patterns
            double all_tm = 0;
            long long compCnt = 0;

            int currTestSize = testSize;
#if 0
            cout << "Curr test size: " << currTestSize << endl;
#endif

            ind.ResetCounters();
            size_t k = 0;
            for (; k < currTestSize; k++) {

                if (!noClearCache) {
                    FILE *f0 = fopen ("/dev/cldev", "w");
                    if (!f0) {
                        Perror ("Cannot open cldev");
                        return 0;
                    }
                    // Let's clear all cashes
                    for (size_t l = 0; l < cldevList.size (); l++) {
                        if (fputs((string(cdir) + "/" + cldevList[l] +
                            "\n").c_str (), f0) < 0) {
                            Perror ("Cannot write to cldev");
                            return 0;
                        }
                    }
                    if (fclose (f0)) {
                        Perror ("Cannot close cldev");
                        return 0;
                    }
                }
                if (!ind.OpenDB (baseName.c_str ())) {
                    Perror ("Cannot open database");
                    return 0;
                }
                StringList   tmpRes;

                int _cnt = 0;
                gettimeofday (&t1, &_tz);
                if (!ind.SearchInDB (TestList[kerr][k].c_str (), kerr, &_cnt, &tmpRes,
                                silent)) {
                    Perror ("Error while searching!");
                    return 0;
                };
                gettimeofday (&t2, &_tz);
                compCnt = compCnt + _cnt;
                timeDat[k] =
                    ((t2.tv_sec - t1.tv_sec) * 1000000 + t2.tv_usec -
                     t1.tv_usec) / 1000.0;

                all_tm += timeDat[k];
                if (!ind.CloseDB ()) {
                    Perror ("Cannot close database");
                    return 0;
                }
#if 0
                cout << timeDat[k] << " kerrr = " << kerr << " cnt = " << _cnt << endl;
#endif
                ++cnt;
                int pct = cnt * 100 / TotalTestSize;
                if (printPct) {
                    if (pct > prevPct || k == currTestSize - 1) {
                        stringstream str;

                        prevPct = pct;
                        str << "Flash test progress: " << pct << "% (kerr = " << kerr << ")";
                        string mess = str.str(); 
                        cout << mess;
                        for(unsigned int l = 0; l < mess.size() + 5; ++l)
                                cout << "\b\r";
                        cout.flush();
                    }
                }
            }
            time_flash[kerr] = all_tm / currTestSize;
            num_comp_flash[kerr] = compCnt / currTestSize;
            double avg = time_flash[kerr];
            dev_flash[kerr] = 0;
            for (size_t k = 0; k < currTestSize; k++) {
                double v = timeDat[k] - avg;
                dev_flash[kerr] += v * v;
            }
            if (printAddStat) {
                ind.PrintAddStat(CInterface::SearchTypeFlash, currTestSize, kerr, add_stat[kerr]);
            }

            dev_flash[kerr] = sqrt (dev_flash[kerr] / currTestSize);
        }
        
        if (chdir(curr_dir)) {
             Perror(Format("Cannot change to '%s'", curr_dir));
             return 0;
        }
        cout << endl;


        //ind.DisplayStat ();
	    std::stringstream result;

        result << dcnt << " " << (db_size / 1024) << " ";
        for (int i = 0; i <= MAX_KERR; i++)
            result << setprecision(5) << (skipMem ? -1:time_mem[i]) << " ";
        for (int i = 0; i <= MAX_KERR; i++)
            result << setprecision(5) << (skipDisk ? -1:time_disk[i]) << " ";
        for (int i = 0; i <= MAX_KERR; i++)
            result << setprecision(5) << (skipFlash ? -1:time_flash[i]) << " ";
        for (int i = 0; i <= MAX_KERR; i++)
            result << setprecision(5) << (skipMem ? -1:num_comp_mem[i]) << " ";
        for (int i = 0; i <= MAX_KERR; i++)
            result << setprecision(5) << (skipDisk ? -1:num_comp_disk[i]) << " ";
        for (int i = 0; i <= MAX_KERR; i++)
            result << setprecision(5) << (skipFlash ? -1:num_comp_flash[i]) << " ";
        for (int i = 0; i <= MAX_KERR; i++)
            result << setprecision(5) << (skipMem ? -1:dev_mem[i]) << " ";
        for (int i = 0; i <= MAX_KERR; i++)
            result << setprecision(5) << (skipDisk ? -1:dev_disk[i]) << " ";
        for (int i = 0; i <= MAX_KERR; i++)
            result << setprecision(5) << (skipFlash ? -1:dev_flash[i]) << " ";
        result << "\n";
        string tmp = result.str();
        cout << tmp;
        odata << tmp;

        tmp = "";

        for (int i = 0; i <= MAX_KERR; ++i)
        if (!add_stat[i].str().empty()) {
            tmp += "    " + add_stat[i].str() + "\n";
        }
        if (!tmp.empty()) {
            tmp = "Additional stat: \n" + tmp;
        }

        cout << tmp;
        odata << tmp;

        cout.flush();
        odata.flush();


    }
    }
    catch (const string & err) {
        cerr << "error occurred:" << err << endl;
        return 1;
    } 
    catch (const bad_alloc & alc) {
        cerr << "out of memory" << endl;
        return 1;
    } 
    catch (const exception & err) {
        cerr << "error occurred:" << err.what() << endl;
        return 1;
    } 
    catch (...) {
        cerr << "unknown error occurred:" << endl;
        return 1;
    }
    //  End of cycle by the test search sets
    return 1;
}


#endif
