#ifndef LEVENDFA_H
#define LEVENDFA_H

#include <vector>
#include <list>
#include <iostream>

#include "interface.h"

using namespace std;


class LevenDFA {
public:
    struct LevenNodeBase {
        unsigned char          mSym;
        bool                   mIsFull;
    };
private:
    friend class LeveNodeFull;
    // LevenNodeShort aligned
    char           m_ABC[256];
    unsigned int   m_abcMap[256];
    unsigned int   m_abcSize;
    unsigned int   m_kerr;
    bool           m_exDist;
    bool           m_hasTransp;
    unsigned int   m_nodeSizeShort;
    unsigned int   m_nodeSizeFull;
    struct LevenNodeShort : public LevenNodeBase {
        unsigned char  mChildSym1, mChildSym2/*, mChildSym3, mChildSym4*/;
        LevenNodeBase  *mpChild1, *mpChild2/*, *mpChild3, *mpChild4*/;
    };
    struct LevenNodeFull : public LevenNodeBase {
        LevenNodeBase   *mpChildren[0]; // Memory will be allocated if needed
    };

    LevenNodeBase **GetParentPtr(LevenNodeBase *pNode, unsigned char c) {
        if (!pNode->mIsFull) {
            LevenNodeShort *p = (LevenNodeShort *)pNode;
            if (p->mChildSym1 == c) return &p->mpChild1;
            if (p->mChildSym2 == c) return &p->mpChild2;
            /*
            if (p->mChildSym3 == c) return &p->mpChild3;
            if (p->mChildSym4 == c) return &p->mpChild4;
            */
            return NULL;
        } else {
            LevenNodeFull *p = (LevenNodeFull *)pNode;
            int pos = GetPos(c);
            return (p->mpChildren[pos]) ? &(p->mpChildren[pos]):NULL;
        }
    }
    // might create new node if short is not enough
    LevenNodeBase *insertChild(LevenNodeBase *pNode, unsigned char c, LevenNodeBase ***pppChild)
    {
        LevenNodeShort *pNewNode = getNewNodeShort(c);

        LevenNodeBase *pNodeToIns = pNode;

        if (!pNode->mIsFull) {
            LevenNodeShort *pNodeShort = (LevenNodeShort *)pNode;
            if (!pNodeShort->mpChild1) {
                *pppChild = &(pNodeShort->mpChild1);
                **pppChild = pNewNode;
                pNodeShort->mChildSym1 = c;
                return pNode;
            } else if (!pNodeShort->mpChild2) {
                *pppChild = &(pNodeShort->mpChild2);
                **pppChild = pNewNode;
                pNodeShort->mChildSym2 = c;
                return pNode;
            } /*else if (!pNodeShort->mpChild3) {
                *pppChild = &(pNodeShort->mpChild3);
                **pppChild = pNewNode;
                pNodeShort->mChildSym3 = c;
                return pNode;
            } else if (!pNodeShort->mpChild4) {
                *pppChild = &(pNodeShort->mpChild4);
                **pppChild = pNewNode;
                pNodeShort->mChildSym4 = c;
                return pNode;
            } */ else {
                // Not enough cells in short node, create full one
                pNodeToIns = createFull(pNodeShort); 
            }
        }
        LevenNodeFull *p = (LevenNodeFull *)pNodeToIns;
        int pos = GetPos(c); 
        *pppChild = &(p->mpChildren[pos]);
        p->mpChildren[pos] = pNewNode;
        return pNodeToIns;
    }
    LevenNodeFull  *createFull(const LevenNodeShort *pShort) {
        LevenNodeFull  *pFull = getNewNodeFull(pShort->mSym);
        if (pShort->mpChild1) {
            int pos = GetPos(pShort->mChildSym1);
            pFull->mpChildren[pos] = pShort->mpChild1;
        }
        if (pShort->mpChild2) {
            int pos = GetPos(pShort->mChildSym2);
            pFull->mpChildren[pos] = pShort->mpChild2;
        }
        /*
        if (pShort->mpChild3) {
            int pos = GetPos(pShort->mChildSym3);
            pFull->mpChildren[pos] = pShort->mpChild3;
        }
        if (pShort->mpChild4) {
            int pos = GetPos(pShort->mChildSym4);
            pFull->mpChildren[pos] = pShort->mpChild4;
        }
        */
        return pFull;
    }
    LevenNodeBase *m_pRoot;
    LevenNodeBase *m_pCurr;

    list<LevenNodeFull *>   m_freeNodesFull;
    list<LevenNodeShort *>  m_freeNodesShort;
    list<void *>            m_memBlocks;

    LevenNodeFull *getNewNodeFull(unsigned char c);
    LevenNodeShort *getNewNodeShort(unsigned char c);

    void freeShortNode(LevenNodeShort *p) {
        memset(p, 0, sizeof(LevenNodeShort));
        m_freeNodesShort.insert(m_freeNodesShort.end(), p);
    }
    void allocateNodesFull(int N);
    void allocateNodesShort(int N);
    void GenNeighborhood(LevenNodeBase **ppStartNode, int start, string & str, int kerr, bool hasTransp);
    inline unsigned int GetPos(unsigned char c) {
        return m_abcMap[c];
    }
    inline void InsertString(LevenNodeBase **ppParentNode, const string & str, int start, bool kerr) {
        int sz = str.size();
        while (start < sz) {
            unsigned char c = str[start];
            LevenNodeBase **ppTmp = GetParentPtr(*ppParentNode, c);
            if (!ppTmp) {
                LevenNodeBase *pModNode = insertChild(*ppParentNode, c, &ppTmp);
                assert(*ppTmp);
                // Hack to make check for exact distance
                // after this asignment final flag for this node will
                // never be set. Clearly, if a string is inserted into
                // DFA trie with non-zero kerr (that is smaller distance
                // from the original string) it will never be inserted
                // with zero kerr (that is with with larger distance 
                // from the original string)
                if (c == EOS) {
                   if (m_exDist && kerr) {
                       (*ppTmp)->mSym = 0;
                   }
                }
                if (pModNode != *ppParentNode) {
                    freeShortNode((LevenNodeShort *)*ppParentNode); // free old node
                    *ppParentNode = pModNode; // modify pointer in the parent node
                }
            }
            ppParentNode = ppTmp;
            ++start;
        }
    }
public:
    const static char EOS = 1;
    LevenDFA(const char *pABC, const string & str, int kerr, bool hasTransp, bool exDist = false);
    ~LevenDFA();
    void Start() {
        m_pCurr = m_pRoot;
    }
    bool Next(unsigned char c, bool &final) {
        if (m_pCurr) {
            LevenNodeBase **pp = GetParentPtr(m_pCurr, c);
            m_pCurr = (pp) ? *pp:NULL;
        }
        final = false;
        if (!m_pCurr) return false;
        final = (m_pCurr->mSym == EOS);
        return true;
    }
    LevenNodeBase *GetState() const {
        return m_pCurr;
    }
    void SetState(LevenNodeBase *pCurr) {
        m_pCurr = pCurr;
    }
    int  getKerr() const { return m_kerr; }
    bool hasTransp() const { return m_hasTransp; }
};

#endif
