#include <list>
#include "afuzzy.h"
#include "diskindex.h"

void FIndex::init_mem()
{
    p_count = twop(MaxDigit);
    header = new block_info[p_count];
};

void FIndex::free_mem()
{
    delete header;
    header = NULL;
    p_count = MaxDigit = 0;
} 

FIndex::FIndex(bool hasTransp): mHasTransp(hasTransp), opened(false), header(NULL), p_count(0), MaxDigit(0)
{
    hsign.clear_mem();
} 

FIndex::~FIndex()
{
    if (opened)
        Close();
    free_mem();
};

bool FIndex::Open(const char *name)
{
    if (opened) {
        Error("Already opened");
        return false;
    }
    fd = open(name, O_RDWR | O_BINARY);
    if (fd < 0) {
        Error("Cannot open DB");
        return false;
    }
    if (read(fd, (void *) &MaxDigit, sizeof(MaxDigit)) != sizeof(MaxDigit)) {
        Error("Cannot read header (MaxDigit size) from the DB file");
        return false;
    }
    init_mem();
    if (!hsign.read(fd)) {
        Error("Cannot read header (hash map) from the DB file");
        return false;
    }
    int32 size = p_count * sizeof(block_info);
    if (read(fd, (void *) header, size) != size) {
        Error("Cannot read header (hash table) from the DB file");
        return false;
    }
    errno = 0;
    opened = true;
    return true;
};

bool FIndex::Close()
{
    if (!opened) {
        Error("Already closed");
        return false;
    }
    int32 size = p_count * sizeof(block_info);
    if (lseek(fd, 0, SEEK_SET) < 0) {
        Error("Cannot lseek to the beginnig of the DB file!");
        return false;
    }
    if (write(fd, (void *) &MaxDigit, sizeof(MaxDigit)) !=
        sizeof(MaxDigit)) {
        Error("Cannot write header (MaxDigit) to the DB file");
        return false;
    }
    if (!hsign.write(fd)) {
        Error("Cannot write header (hash map) to the DB file");
        return false;
    }
    if (write(fd, (void *) header, size) < size) {
        Error("Cannot write header (hash table) to the DB file");
        return false;
    }
    if (close(fd)) {
        Error("Cannot close DB");
        return false;
    }
    opened = false;
    free_mem();
    return true;
}

bool FIndex::Defragment(const char *name, const char *tmpname)
{
    if (opened) {
        Error("Already opened");
        return false;
    }
    if (Open(name) == false) {
        Error("Cannot open the source DB file");
        return false;
    }
    const char *temp = tmpname;
    FIndex ind(mHasTransp);
    FBlock b;
    if (!ind.Create(temp, hsign, MaxDigit)) {
        Error("Cannot create temporary DB");
        return false;
    }
    // ind is now an open empty DB
    for (int32 i = 0; i < p_count; i++)
        for (int32 addr = header[i].block_addr; addr;) {
            ind.header[i].block_count = header[i].block_count;
            if (!b.Read(fd, addr)) {
                Error("Cannot read a block from the source DB");
                return false;
            }
            int32 addr_new = lseek(ind.fd, 0, SEEK_END);
            if (addr_new < 0) {
                Error("Cannot lseek in the dest DB file");
                return false;
            }
            if (!ind.header[i].block_addr)
                ind.header[i].block_addr = addr_new;
            addr = b.GetNextBlock();
            if (addr)
                b.SetNextBlock(addr_new + BlockSize);

            // Next block is surely to be placed just after this block
            if (!b.Write(ind.fd, addr_new)) {
                Error("Cannot write to the dest DB file");
                return false;
            }
        };
    if (!Close()) {
        Error("Cannot close the source DB file");
        return false;
    }
    if (!ind.Close()) {
        Error("Cannot close the dest DB file");
        return false;
    }
    if (unlink(name) || rename(temp, name)) {
        Error("Cannot overwrite the source DB file by the temporary file");
    }
    return true;
}

void FIndex::gen_seq()
{
    for (int32 j = 0; j < p_count; j++) {
        header[j].block_addr = 0;   // This block is still empty
        header[j].block_count = 0;
    };
};

bool FIndex::Create(const char *name, HMap & h, int _MaxDigit)
{
    if (opened) {
        Error("DB is opened");
        return false;
    }
    hsign = h;
    MaxDigit = _MaxDigit;
    init_mem();
    fd = open(name, O_RDWR | O_CREAT | O_TRUNC | O_BINARY, 0770);
    if (fd < 0) {
        Error("Cannot open DB");
        return false;
    }
    int32 size = p_count * sizeof(block_info);
    gen_seq();
    hsign = h;
    if (write(fd, (void *) &MaxDigit, sizeof(MaxDigit)) !=
        sizeof(MaxDigit)) {
        Error("Cannot write header (MaxDigit) to the DB file (MaxDigit)");
        return false;
    }
    if (!hsign.write(fd)) {
        Error("Cannot write header (hash map) to the DB file ");
        return false;
    }
    if (write(fd, (void *) header, size) != size) {
        Error("Cannot write header (hash table) to the DB file");
        return false;
    }
    opened = true;
    return true;
};

bool FIndex::PrintStat()
{
    if (!opened) {
        Error("Not opened");
        return false;
    }
    for (int32 i = 0; i < p_count; i++) {
        Sign sgn = Sign(hsign, MaxDigit, i);

        //header[i].center;
        int val = sgn.abs();
        printf("Count:%ld |%03d|\n", header[i].block_count, val);
    }
} 

bool
FIndex::FastCreate(const char *dfile, const char *dbname, HMap & h,
                       int _MaxDigit, bool silent)
{
    if (opened) {
        Error("Already opened");
        return false;
    }
    if (!Create(dbname, h, _MaxDigit))
        return false;           // creating header disk image
    FILE *f = fopen(dfile, "r");
    if (!f) {
        Error("Cannot open data file");
        return false;
    }
    char str[1024];
    FBlock blk_arr[p_count];    // Clear by default
    if (fseek(f, 0, SEEK_SET)) {
        Error("Cannot fseek");
        return false;
    }
    for (int32 j = 0; fgets(str, 1022, f); j++) {
        int32 len = strlen(str);
        if (len && str[len - 1] == '\n') {
            str[len - 1] = 0;
            len--;
        }
        Sign val(hsign, MaxDigit, str); // Getting signature
        bool found = false;
        int32 i = (val);
        {
            found = true;
            if (blk_arr[i].CanAppend(len))  // There is some space in the block
            {
                if (!blk_arr[i].AppendRecord(str, &val)) {
                    Error
                        ("Internal error - cannot append string in the block!");
                    return false;
                }
                if (!header[i].block_addr)  //First record in the first block in the ith node
                {
                    off_t addr = lseek(fd, 0, SEEK_END);
                    if (addr < 0) {
                        Error("Cannot lseek to the end of the DB file");
                        return false;
                    }
                    if (!blk_arr[i].Write(fd, addr)) {
                        Error("Cannot write a block to the DB file");
                        return false;
                    }
                    header[i].block_addr = addr;
                    header[i].block_count++;
                }
                // else - nothing to do now
                // 'cause string is already saved in the block
            }                   // end CanAppend
            else                // else CanAppend
            {

                // We need one more block to save this
                FBlock b;
                if (!b.AppendRecord(str, &val)) {
                    Error
                        ("Internal error - cannot append string in the block!");
                    return false;
                }
                off_t addr = lseek(fd, 0, SEEK_END);
                if (addr < 0) {
                    Error("Cannot lseek to the end of the DB file");
                    return false;
                }
                if (!b.Write(fd, addr)) {
                    Error("Cannot write to the DB file");
                    return false;
                }
                if (!header[i].block_addr)  //First record in the first block in the ith node
                    header[i].block_addr = addr;

                else {
                    blk_arr[i].SetNextBlock(addr);
                    if (!blk_arr[i].Write(fd)) {
                        Error("Cannot write to the DB file");
                        return false;
                    }
                };
                blk_arr[i] = b; // operator of copy
                header[i].block_count++;
            }
        };
        if (!found) {
            Error("Internal error - cannot find the center!");
            fprintf(stderr, "N dig = %d\n", val.abs());
        }
        if (!silent && j % 100 == 0) {
            printf("\b\b\b\b\b\b\b\b%07ld", j);
            fflush(stdout);
        }
    };

    // Now writing finally blocks to the disk and ...
    for (int i = 0; i < p_count; i++)
        if (header[i].block_addr)
            if (!blk_arr[i].Write(fd)) {
                Error("Cannot write to the DB file");
                return false;
            }
    // .. and putting down the header by Close()    
    if (!Close()) {
        Error("Cannot close");
        return false;
    }
    if (!silent)
        puts("");
    return true;
};

bool FIndex::Insert(const char *str)
{
    if (!opened) {
        Error("Not opened");
        return false;
    }
    Sign val(hsign, MaxDigit, str); // Getting signature
    int32 i = (val);
    {
        int32 addr = header[i].block_addr;
        FBlock b;
        if (!addr) {
            addr = lseek(fd, 0, SEEK_END);
            if (addr < 0) {
                Error("Cannot lseek to the end of the DB file");
                return false;
            }
            b.AppendRecord(str, &val);
            if (!b.Write(fd, addr)) {
                Error("Cannot write to the DB file");
                return false;
            }
            header[i].block_addr = addr;
            header[i].block_count++;
            return true;
        };
        if (b.Read(fd, addr) == false) {
            Error("Cannot read from the DB file");
            return false;
        }
        if (b.AppendRecord(str, &val)) {
            if (!b.Write(fd)) {
                Error("Cannot write to the DB file");
                return false;
            }
            return true;
        }
        // Cannot allocate in the block
        else if (b.CanAppend(strlen(str))) {
            Error("Internal error - cannot append a record");
            return false;
        }

        else {

            // Now read to the final block
            for (; addr = b.GetNextBlock();) {
                if (b.Read(fd, addr) == false) {
                    Error("Cannot read block from the DB file");
                    return false;
                }
                if (b.AppendRecord(str, &val)) {
                    if (!b.Write(fd)) {
                        Error("Cannot write a block to the DB file");
                        return false;
                    }
                    return true;
                };
            };
            FBlock bn;
            int32 addrn = lseek(fd, 0, SEEK_END);
            if (addrn < 0) {
                Error("Cannot lseek to the end of the DB file");
                return false;
            }
            bn.AppendRecord(str, &val);
            b.SetNextBlock(addrn);
            header[i].block_count++;
            if (!bn.Write(fd, addrn) || !b.Write(fd)) {
                Error("Cannot write to the DB file");
                return false;
            }
            return true;
        };
    };
    Error("Internal error - cannot find the center!");
    fprintf(stderr, "N dig = %d\n", val.abs());
    return false;
};

bool FIndex::SearchBestMatch(const char *str, bool silent, int32 * pcnt)
{
    if (!opened) {
        Error("Not opened");
        return false;
    }
    FBlock b;
    std::list < std::string > str_list;
    Sign sgn(hsign, MaxDigit, str);
    int minerr = MaxLen + 1, cnt = 0;
    afuzzy a(mHasTransp);
    a.init(str, minerr);
    for (int r = 0; r <= MaxDigit && r <= 2 * minerr; r++) {
        int s, e;
        if (!r)
            s = e = int32(sgn);

        else
            s = 0, e = p_count - 1;
        for (int i = s; i <= e; i++) {
            bool WorthScanFlag = false;
            int dst = (sgn - Sign(hsign, MaxDigit, i));
            if (dst != r)
                continue;
            int nerr = (r + 1) / 2; // The potenial number of errors
            for (int err_change = 0; err_change <= nerr; err_change++) {
                if (dst <= (nerr - err_change) + 2 * err_change
                    && abs(sgn.abs() -
                           Sign(hsign, MaxDigit,
                                i).abs()) <= (nerr - err_change)) {
                    WorthScanFlag = true;
                    break;
                }
            }                   // checked WorthScanFlag
            if (!WorthScanFlag)
                continue;
            for (int32 addr = header[i].block_addr; addr;
                 addr = b.GetNextBlock()) {
                bool res = b.Read(fd, addr);
                cnt++;
                if (!res) {
                    Error("Cannot read a block from the DB file");
                    return false;
                }
                unsigned int d;
                for (char *ptr = b.ToFirst(); ptr; ptr = b.ToNext()) {
                    if (a.checkRTR(ptr, &d)) {
                        if (d < minerr) {
                            minerr = d;
                            a.init(str, minerr);
                            str_list.clear();
                        }
                        str_list.insert(str_list.end(), string(ptr));
                    }
                }
            };                  // end for by block
        };                      // end by i
    };                          // end for by r
    if (!silent) {
        for (std::list < std::string >::iterator it = str_list.begin();
             it != str_list.end(); it++)
            printf("%s\n", (*it).c_str());
        printf("Best match distance: %d\n", minerr);
        printf(">Kb(s) read %d\n>Number of blocks read%d \n",
               int32(cnt) * BlockSize / 1024, cnt);
    }
    if (pcnt)
        *pcnt = cnt;
    return true;
};

bool FIndex::Search(const char *str, unsigned int nerr, int *numComp,
                    StringList * resList, bool silent, int32 * pcnt)
{
    if (!opened) {
        Error("Not opened");
        return false;
    }
    Sign sgn(hsign, MaxDigit, str);
    int32 cnt = 0;
    FBlock b;
    afuzzy a(mHasTransp);
    a.init(str, nerr);
    if (!nerr) {
        int i = int32(sgn);
        for (int32 addr = header[i].block_addr; addr;
             addr = b.GetNextBlock()) {
            bool res = b.Read(fd, addr);
            cnt++;
            if (!res) {
                Error("Error read");
                return false;
            }
            for (char *ptr = b.ToFirst(); ptr; ptr = b.ToNext()) {
                if (numComp)
                    (*numComp)++;
                if (!strcmp(ptr, str)) {
                    if (!silent)
                        printf("%s\n", ptr);
                    if (resList != NULL)
                        resList->push_back(string(ptr));
                }
            }
            ;
        };
        if (!silent)
            printf(">Kb(s) read %d\n>Number of blocks read%d \n",
                   int32(cnt) * BlockSize / 1024, cnt);
        if (pcnt)
            *pcnt = cnt;
        return true;
    }
    for (int32 i = 0; i < p_count; i++) {
        bool WorthScanFlag = false;
        int dst = (sgn - Sign(hsign, MaxDigit, i));
        int nerr_e = nerr;

/*
  if (nerr>=2) nerr_e--; // All mismatches of type "change" are unlikely
                        // so that we can speed up the process considerably
                        // without really losing search quality
*/
        for (int err_change = 0; err_change <= nerr_e; err_change++) {
            if (dst <= (nerr - err_change) + 2 * err_change
                && abs(sgn.abs() - Sign(hsign, MaxDigit, i).abs()) <=
                (nerr - err_change)) {
                WorthScanFlag = true;
                break;
            }
        }

        //WorthScanFlag = (dst <= 2*nerr - (nerr>1) ? 1:0);
        if (WorthScanFlag) {
            for (int32 addr = header[i].block_addr; addr;
                 addr = b.GetNextBlock()) {
                bool res = b.Read(fd, addr);
                cnt++;
                if (!res) {
                    Error("Cannot read a block from the DB file");
                    return false;
                }
                for (char *ptr = b.ToFirst(); ptr; ptr = b.ToNext()) {

                    //if (g.findfirst(ptr))
                    if (numComp)
                        (*numComp)++;
                    if (a.checkRTR(ptr)) {
                        if (!silent)
                            puts(ptr);
                        if (resList != NULL)
                            resList->push_back(string(ptr));
                    }
                }
                ;
            };
        };
    };
    if (!silent)
        printf(">Kb(s) read %d\n>Number of blocks read%d \n",
               int32(cnt) * BlockSize / 1024, cnt);
    if (pcnt)
        *pcnt = cnt;
    return true;
};

bool FIndex::SearchSubstr(const char *str, unsigned int nerr, bool silent,
                          int32 * pcnt)
{
    if (!opened) {
        Error("Not opened");
        return false;
    }
    Sign sgn(hsign, MaxDigit, str);
    int32 cnt = 0;
    FBlock b;
    afuzzy a(mHasTransp);
    a.init(str, nerr);
    if (!nerr) {
        for (int32 i = 0; i < p_count; i++)
            if (!sgn.Diff(Sign(hsign, MaxDigit, i))) {
                for (int32 addr = header[i].block_addr; addr;
                     addr = b.GetNextBlock()) {
                    bool res = b.Read(fd, addr);
                    cnt++;
                    if (!res) {
                        Error("Cannot read a block from the DB file");
                        return false;
                    }
                    for (char *ptr = b.ToFirst(); ptr; ptr = b.ToNext())
                        if (a.checkSUB(ptr))
                            if (!silent)
                                puts(ptr);
                    ;
                };
            }
    }

    else {
        for (int32 i = 0; i < p_count; i++)
            if (sgn.Diff(Sign(hsign, MaxDigit, i)) <= nerr) {
                for (int32 addr = header[i].block_addr; addr;
                     addr = b.GetNextBlock()) {
                    bool res = b.Read(fd, addr);
                    cnt++;
                    if (!res) {
                        Error("Cannot read a block from the DB file");
                        return false;
                    }
                    for (char *ptr = b.ToFirst(); ptr; ptr = b.ToNext())
                        //if (g.findfirst(ptr))
                        if (a.checkSUB(ptr))
                            if (!silent)
                                puts(ptr);
                    ;
                };
            }
    }                           // end else
    if (!silent)
        printf(">Kb(s) read %d\n>Number of blocks read%d \n",
               int32(cnt) * BlockSize / 1024, cnt);
    if (pcnt)
        *pcnt = cnt;
    return true;
};
