#include "node.h"

TrieNode::TrieNode ():FirstLeaf (NULL)
{
};

TrieNode::~TrieNode ()
{
    for (TrieLeaf * tf = FirstLeaf; tf;) {
        TrieLeaf *n = tf->Next ();
        delete tf;
        tf = n;
    }
};


void
TrieNode::UpdateUpperLink (TrieLeaf * uplnk)
{
    for (TrieLeaf * tf = FirstLeaf; tf; tf = tf->Next ())
        tf->UpdateUpperLink (uplnk);
}

TrieLeaf *TrieNode::Insert (std::string Key, TrieLeaf * UpperLink,
                            TrieNode * SubTree)
{
    if (!FirstLeaf) {
        FirstLeaf = new TrieLeaf (Key, UpperLink, SubTree);
        return FirstLeaf;
    }
    TrieLeaf *tfBest = NULL;
    int       comLen = 0;
    int       len;
    for (TrieLeaf * tf = FirstLeaf; tf; tf = tf->Next ()) {
        if ((len = tf->CheckCommon (Key)) > comLen) {
            comLen = len;
            tfBest = tf;
        }
        if (!tf->Next() && !tfBest) {
            return tf->Append (Key, UpperLink, SubTree);
        }
    }
    assert(tfBest);
    return tfBest->Insert (Key, SubTree);
};

bool TrieNode::Search (const std::string & s, int kerr, int *numComp,
                       StringList * resList, bool hasTransp, bool silent) const
{
    string ss = s + EOS;
    checkDistContext c (kerr, ss.c_str ());

    return SearchAtLevel (ss, "", c, numComp, resList, hasTransp, silent);
};

bool TrieNode::SearchExact (const std::string & s, int *numComp,
                            StringList * resList, bool silent) const
{
    std::string ss = s;
    ss += EOS;
    bool found;
    return SearchAtLevelExact (ss, "", &found, numComp, resList, silent);
};

bool TrieNode::SearchAtLevel (const std::string & s, const std::string & st,
                              const checkDistContext & c, int *numComp,
                              StringList * resList,
                              bool hasTransp, bool silent) const
{
    for (TrieLeaf * t = FirstLeaf; t; t = t->Next ()) {
        checkDistContext c1 = c;
        std::string sk = st + t->Key ();
        const TrieNode *SubTree = t->Sub ();
        if (numComp && !SubTree)
            (*numComp)++;
        if (c1.checkDist (sk.c_str (), SubTree == NULL, hasTransp))    // if we compare string suffixes comparison is not complete just 
        {
            if (SubTree) {
                if (!SubTree->
                    SearchAtLevel (s, sk, c1, numComp, resList, hasTransp, silent))
                    return false;
            } else {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL)
                    resList->push_back (sk.substr (0, sk.size () - 1));
            }
        }
    }
    return true;
};

bool TrieNode::SearchAtLevelExact (const std::string & s, const std::string & st,
                                   bool * pfound, int *numComp,
                                   StringList * resList,
                                   bool silent) const
{
    bool found = false;
    for (TrieLeaf * t = FirstLeaf; t && !found; t = t->Next ()) {
        std::string sk = st + t->Key ();
        const TrieNode *SubTree = t->Sub ();
        if (numComp && !SubTree)
            (*numComp)++;
        if (SubTree && s.length() >= sk.length()) {
            std::string s1 = std::string (s, 0, sk.length ());
            if (s1 == sk) {
                //cout<<"### "<<s1<<" "<<sk<<"\n";
                if (!SubTree->
                    SearchAtLevelExact (s, sk, &found, numComp, resList,
                                        silent))
                    return false;
            }
        } else if (s == sk) {
            found = true;
            if (!silent)
                cout << sk << '\n';
            if (resList != NULL)
                resList->push_back (sk.substr (0, sk.length () - 1));
        }
    }
    *pfound = found;
    return true;
};

#ifdef _DEBUG
void TrieNode::Print (std::string s, int level)
{
    for (TrieLeaf * t = FirstLeaf; t; t = t->Next ()) {
        for (int i = 0; i < level; i++)
            cout << " ";
        cout << "Node address =" << this << "\n";
        t->Print (s, level + 1);
    }
};
#endif

int TrieNode::DumpToFile (int fd, off_t offset)
{

// We store the data in the following order
// 1. the buffer size, the number of leaves
// 2. the leaf keys themselvs, with zero offsets
// 3. leaf subtrees - updating offsets

    int cnt = 0;
    int buflen = 0;
    int sub_sz = 0;
    buflen += sizeof (cnt) + sizeof (buflen);
    for (TrieLeaf * t = FirstLeaf; t; t = t->Next (), cnt++) {
        buflen += sizeof (off_t) /* <- pointer size */ +
            1 /* string record size (must be<255) */  +
            (t->Key ()).length () + 1;
    }
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    if (lseek (fd, offset + buflen - 1, SEEK_SET) < 0
        || write (fd, "", 1) < 1) {
        return -1;
    }
    void *buf =
        mmap (NULL, buflen + delta, PROT_WRITE | PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return -1;
    char *sbuf = (char *) buf + delta;
    // Step 1.
    *(int *) (void *) sbuf = buflen;
    sbuf += sizeof (buflen);
    *(int *) (void *) sbuf = cnt;
    sbuf += sizeof (cnt);
    // Step 2.
    off_t *ptrs[cnt];
    int i = 0;
    for (TrieLeaf * t = FirstLeaf; t; t = t->Next (), i++) {
        assert (i < cnt);
        int len = (t->Key ()).length ();
        // String length
        *((u_int8_t *) (void *) sbuf) = len + 1;
        sbuf++;
        // Zero offset
        ptrs[i] = (off_t *) ((void *) sbuf);
        *ptrs[i] = 0;
        // Leaf key
        sbuf += sizeof (off_t);
        memcpy (sbuf, (t->Key ()).c_str (), len + 1);
        sbuf += len + 1;
    }
    // Step 3.
    i = 0;
    for (TrieLeaf * t = FirstLeaf; t; t = t->Next (), i++) {
        assert (i < cnt);
        // Diving into subtree
        int s_sz =
            (t->Sub ())? t->Sub ()->DumpToFile (fd,
                                                offset + buflen +
                                                sub_sz) : 0;
        if (s_sz > 0)
            *ptrs[i] = offset + buflen + sub_sz;
        if (s_sz < 0) {
            munmap (buf, buflen + delta);
            return -1;
        }
        sub_sz += s_sz;
    }

    assert (sbuf - (char *) buf == delta + buflen);

    if (munmap (buf, buflen + delta) < 0)
        return -1;
    // Returning the total number of bytes written
    return buflen + sub_sz;

}

bool TrieNode::SearchInFileAtLevelExact (int fd, off_t offset,
                                    const std::string & s, const std::string & st,
                                    int *numComp, StringList * resList,
                                    bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    for (int i = 0; i < num; i++) {
        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        std::string sk = st + std::string (sbuf);
        if (numComp && !sub_off) {
            (*numComp)++;
        }
        if (sub_off) {
            if (sk.size() <= s.size() && !strncmp(s.c_str(), sk.c_str(), sk.size())) {
                if (!SearchInFileAtLevelExact(fd, sub_off, s, sk, numComp, resList, silent))
                    return false;
            }
        } else {
            if (sk == s) {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL)
                    resList->push_back (sk.substr (0, sk.length () - 1));
            }
        }
        sbuf += slen;
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    return true;
}


bool TrieNode::SearchInFileAtLevel (int fd, off_t offset,
                                    const std::string & s, const std::string & st,
                                    const checkDistContext & c,
                                    int *numComp, StringList * resList,
                                    bool hasTransp, bool silent)
{
    int buflen;
    if (lseek (fd, offset, SEEK_SET) < 0
        || read (fd, &buflen, sizeof (buflen)) < sizeof (buflen))
        return false;
    int delta = 0;
    if (offset)
        delta = (offset % getpagesize ());
    void *buf =
        mmap (NULL, buflen + delta, PROT_READ, MAP_SHARED, fd,
              offset - delta);
    if (!buf || buf == MAP_FAILED)
        return false;
    char *sbuf = (char *) buf + sizeof (buflen) + delta;
    int num = *((int *) (void *) sbuf);
    sbuf += sizeof (num);

    int pd = 0, nd = 0, plen = 1;
    // Now reading leafs 
    // The sequence of storing leafs data:
    // 1. string length +1 (the size of string buffer including trailing `\0'
    // 2. offset
    // 3. string buffer
    for (int i = 0; i < num; i++) {
        checkDistContext c1 = c;

        int slen = *(u_int8_t *) sbuf;
        sbuf++;
        off_t sub_off = *(off_t *) sbuf;
        sbuf += sizeof (off_t); // The part of the string stored in the leaf  
        std::string sk = st + std::string (sbuf);
        if (numComp && !sub_off)
            (*numComp)++;
        if (c1.checkDist (sk.c_str (), sub_off == 0, hasTransp)) { // if we compare string suffixes comparison is not complete just 
            if (sub_off) {
                if (!SearchInFileAtLevel
                    (fd, sub_off, s, sk, c1, numComp, resList, hasTransp, silent))
                    return false;
            } else {
                if (!silent)
                    cout << sk << '\n';
                if (resList != NULL)
                    resList->push_back (sk.substr (0, sk.length () - 1));
            }
        }
        sbuf += slen;
    }
    if (munmap (buf, buflen + delta) < 0)
        return false;
    return true;
}

bool TrieNode::SearchInFile (int fd, const std::string & s, int kerr,
                             int *numComp, StringList * resList,
                             bool hasTransp, bool silent)
{
    string ss = s + EOS;
    checkDistContext c (kerr, ss.c_str ());

    return (kerr) ? SearchInFileAtLevel (fd, 0, ss, "", c, numComp, resList, hasTransp, silent):
                    SearchInFileAtLevelExact (fd, 0, ss, "", numComp, resList, silent);
}
