#include <algorithm>
#include "index.h"

#define USE_ALTERN_GENERATION 1
using namespace std;

CTrieIndexDFADouble::CTrieIndexDFADouble (bool hasTransp, EAlphabet AlphaType)
            : CInterface(hasTransp, AlphaType), opened (false), fdFwd(0), fdRev(0)
{
}

CTrieIndexDFADouble::~CTrieIndexDFADouble ()
{
    if (opened)
        CloseDB ();
}

bool CTrieIndexDFADouble::Insert (const char *str)
{
    try {
        string s(str);

        nodeFwd.Insert(s);
        reverse(s.begin(), s.end());

        nodeRev.Insert(s);

    } catch (const string & e) {
        cerr << "Error inserting of '" << (string(str)) << "': " << e << endl;
        throw;
    }
    return true;
}

bool CTrieIndexDFADouble::Search(const char *str, int kerr, int *numComp,
                         StringList * resList, bool silent)
{
    START_COUNTER(stime);
    string s(str);


    if (kerr != 0) {
        assert(kerr <= MAX_ERR);
        assert(MAX_ERR < 3);
        int switchLen = 0;

        if (kerr < 2) {
            switchLen = s.size() / 2;
        } else {
            switchLen = (s.size())/ 2 + 0;
        }
        int switchLen1 = s.size() - switchLen;
        if (!USE_ALTERN_GENERATION) {
            switchLen -= mHasTransp;
            switchLen1 -= mHasTransp;
        }
        if (switchLen <= 0) {
            nodeFwd.Search(GetAlphabet(mAlphaType), s, kerr, numComp, resList, mHasTransp, silent);
            return true;
        }

        StringList     tmpList, tmpList1;

        if (kerr == 1) {
            // Process a specific transposition case
            if (mHasTransp && USE_ALTERN_GENERATION) {
                swap(s[switchLen - 1], s[switchLen]); // switchLen > 0

                nodeFwd.SearchExact(s, numComp, &tmpList, silent);

                swap(s[switchLen - 1], s[switchLen]); // switchLen > 0
            }
            // Here we search as if we have no transposition
            // First forward dict search
            // kerr1 == 0, kerr2 <= 1
            // Note that result sets do not intersect
            nodeFwd.SearchSplit(GetAlphabet(mAlphaType), s, 0, true, 1, false, 
                                numComp, switchLen, &tmpList, mHasTransp, silent);
            
            // Next reverse
            // kerr1 == 0, kerr2 == 1
            reverse(s.begin(), s.end());
            nodeRev.SearchSplit(GetAlphabet(mAlphaType), s, 0, true, 1, true, 
                                numComp, switchLen1, &tmpList1, mHasTransp, silent);
            START_COUNTER(lproc);
            for (int i = 0; i < tmpList1.size(); ++i) {
               string &s = tmpList1[i];
               reverse(s.begin(), s.end());
               tmpList.push_back(s);
            }
            sort(tmpList.begin(), tmpList.end());
            tmpList.erase(unique(tmpList.begin(), tmpList.end()), tmpList.end());
            if (resList) {
                *resList = tmpList;
            }
            STOP_COUNTER(lproc);
        } else if (kerr == 2) {
            if (mHasTransp && USE_ALTERN_GENERATION) {
                // Process a specific transposition case, error is 1, because transposition
                // already accounts for one edit operation
                swap(s[switchLen - 1], s[switchLen]); // switchLen > 0
                // First forward dict search
                // kerr1 == 0, kerr2 <= 1
                // Note that result sets do not intersect
                nodeFwd.SearchSplit(GetAlphabet(mAlphaType), s, 0, true, 1, false, 
                                numComp, switchLen, &tmpList, mHasTransp, silent);
                // Next reverse
                // kerr1 == 0, kerr2 == 1
                reverse(s.begin(), s.end());
                nodeRev.SearchSplit(GetAlphabet(mAlphaType), s, 0, true, 1, true, 
                                numComp, switchLen1, &tmpList1, mHasTransp, silent);
                START_COUNTER(lproc);
                for (int i = 0; i < tmpList1.size(); ++i) {
                   string &s = tmpList1[i];
                   reverse(s.begin(), s.end());
                   tmpList.push_back(s);
                }
                STOP_COUNTER(lproc);
                reverse(s.begin(), s.end());
                swap(s[switchLen - 1], s[switchLen]); // switchLen > 0
            }
            // Three cases here: (0, 2), (1, 1), (0, 2) (reverse)
            // Note that result sets do not intersect
            // First kerr1 == 0, kerr2 <= 2
            nodeFwd.SearchSplit(GetAlphabet(mAlphaType), s, 0, false, 2, false, 
                            numComp, switchLen, &tmpList, mHasTransp, silent);
            
            // Next kerr1 == 1, kerr2 <= 1, still forward dict
            nodeFwd.SearchSplit(GetAlphabet(mAlphaType), s, 1, true, 1, false, 
                            numComp, switchLen, &tmpList, mHasTransp, silent);
            
            // Finaly search in reverse dict for kerr1 == 0, kerr2 == 2
            tmpList1.clear();
            reverse(s.begin(), s.end());
            nodeRev.SearchSplit(GetAlphabet(mAlphaType), s, 0, true, 2, true, 
                            numComp, switchLen1, &tmpList1, mHasTransp, silent);
            START_COUNTER(lproc);
            for (int i = 0; i < tmpList1.size(); ++i) {
               string &s = tmpList1[i];
               reverse(s.begin(), s.end());
               tmpList.push_back(s);
            }
            sort(tmpList.begin(), tmpList.end());
            tmpList.erase(unique(tmpList.begin(), tmpList.end()), tmpList.end());
            if (resList) {
                *resList = tmpList;
            }
            STOP_COUNTER(lproc);
        }
    } else {
        nodeFwd.SearchExact(s, numComp, resList, silent);
    }
    STOP_COUNTER(stime);
    return true;
}

bool CTrieIndexDFADouble::SearchInDB (const char *str, int kerr, int *numComp,
                             StringList * resList, 
                             bool silent)
{
    if (!opened) {
        cerr << "Database is not opened!" << endl;
    }

    std::string s(str);

    if (kerr != 0) {
        assert(kerr <= MAX_ERR);
        assert(MAX_ERR < 3);
        int switchLen = s.size() / 2;
        int switchLen1 = s.size() - switchLen;
        if (!USE_ALTERN_GENERATION) {
            switchLen -= mHasTransp;
            switchLen1 -= mHasTransp;
        }
        if (switchLen <= 0) {
            if (!CTrieDFA::SearchInFile(GetAlphabet(mAlphaType), fdFwd, s, kerr, numComp, resList, mHasTransp, silent)) return false;
            return true;
        }

        StringList     tmpList, tmpList1;

        if (kerr == 1) {
            // Process a specific transposition case
            if (mHasTransp && USE_ALTERN_GENERATION) {
                swap(s[switchLen - 1], s[switchLen]); // switchLen > 0

                if (!CTrieDFA::SearchInFileExact(fdFwd, s, numComp, &tmpList, silent)) return false;

                swap(s[switchLen - 1], s[switchLen]); // switchLen > 0
            }
            // Here we search as if we have no transposition
            // First forward dict search
            // kerr1 == 0, kerr2 <= 1
            // Note that result sets do not intersect
            if (!CTrieDFA::SearchInFileSplit(GetAlphabet(mAlphaType), fdFwd, s, 0, true, 1, false, 
                                numComp, switchLen, &tmpList, mHasTransp, silent)) return false;
            
            // Next reverse
            // kerr1 == 0, kerr2 == 1
            reverse(s.begin(), s.end());
            if (!CTrieDFA::SearchInFileSplit(GetAlphabet(mAlphaType), fdRev, s, 0, true, 1, true, 
                                numComp, switchLen1, &tmpList1, mHasTransp, silent)) return false;
            for (int i = 0; i < tmpList1.size(); ++i) {
               string &s = tmpList1[i];
               reverse(s.begin(), s.end());
               tmpList.push_back(s);
            }
            sort(tmpList.begin(), tmpList.end());
            tmpList.erase(unique(tmpList.begin(), tmpList.end()), tmpList.end());
            if (resList) {
                *resList = tmpList;
            }
        } else if (kerr == 2) {
            if (mHasTransp && USE_ALTERN_GENERATION) {
                // Process a specific transposition case, error is 1, because transposition
                // already accounts for one edit operation
                swap(s[switchLen - 1], s[switchLen]); // switchLen > 0
                // First forward dict search
                // kerr1 == 0, kerr2 <= 1
                // Note that result sets do not intersect
                if (!CTrieDFA::SearchInFileSplit(GetAlphabet(mAlphaType), fdFwd, s, 0, true, 1, false, 
                                numComp, switchLen, &tmpList, mHasTransp, silent)) return false;
                // Next reverse
                // kerr1 == 0, kerr2 == 1
                reverse(s.begin(), s.end());
                if (!CTrieDFA::SearchInFileSplit(GetAlphabet(mAlphaType), fdRev, s, 0, true, 1, true, 
                                numComp, switchLen1, &tmpList1, mHasTransp, silent)) return false;
                for (int i = 0; i < tmpList1.size(); ++i) {
                   string &s = tmpList1[i];
                   reverse(s.begin(), s.end());
                   tmpList.push_back(s);
                }
                reverse(s.begin(), s.end());
                swap(s[switchLen - 1], s[switchLen]); // switchLen > 0
            }
            // Three cases here: (0, 2), (1, 1), (0, 2) (reverse)
            // Note that result sets do not intersect
            // First kerr1 == 0, kerr2 <= 2
            if (!CTrieDFA::SearchInFileSplit(GetAlphabet(mAlphaType), fdFwd, s, 0, false, 2, false, 
                            numComp, switchLen, &tmpList, mHasTransp, silent)) return false;
            
            // Next kerr1 == 1, kerr2 <= 1, still forward dict
            if (!CTrieDFA::SearchInFileSplit(GetAlphabet(mAlphaType), fdFwd, s, 1, true, 1, false, 
                            numComp, switchLen, &tmpList, mHasTransp, silent)) return false;
            
            // Finaly search in reverse dict for kerr1 == 0, kerr2 == 2
            tmpList1.clear();
            reverse(s.begin(), s.end());
            if (!CTrieDFA::SearchInFileSplit(GetAlphabet(mAlphaType), fdRev, s, 0, true, 2, true, 
                            numComp, switchLen1, &tmpList1, mHasTransp, silent)) return false;
            for (int i = 0; i < tmpList1.size(); ++i) {
               string &s = tmpList1[i];
               reverse(s.begin(), s.end());
               tmpList.push_back(s);
            }
            sort(tmpList.begin(), tmpList.end());
            tmpList.erase(unique(tmpList.begin(), tmpList.end()), tmpList.end());
            if (resList) {
                *resList = tmpList;
            }
        }
    } else {
            if (!CTrieDFA::SearchInFileExact(fdFwd, s, numComp, resList, silent)) return false;
    }
    return true;
}

bool CTrieIndexDFADouble::OpenDB (const char *file_name, bool create)
{
    string fwdName = string(file_name) + "_fwd.img";
    string revName = string(file_name) + "_rev.img";

    if (opened)
        CloseDB ();
    if (!create) {
        fdFwd = open (fwdName.c_str(), O_RDONLY, 0);
        if (fdFwd <= 0) {
            Perror ("Error opening file");
            return false;
        }
        fdRev = open (revName.c_str(), O_RDONLY, 0);
        if (fdRev <= 0) {
            Perror ("Error opening file");
            return false;
        }
    } else {
        fdFwd = open (fwdName.c_str(), O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
        if (fdFwd <= 0) {
            Perror ("Error opening file");
            return false;
        }
        fdRev = open (revName.c_str(), O_RDWR | O_TRUNC | O_CREAT, defaultAccess);
        if (fdRev <= 0) {
            Perror ("Error opening file");
            return false;
        }
    }
    return (opened = true);
}

bool CTrieIndexDFADouble::CloseDB ()
{
    if (!opened)
        return true;
    if (close(fdFwd) || close(fdRev)) {
        return false;
    }
    opened = false;
    return true;
}


bool CTrieIndexDFADouble::DumpDB (const char *file_name)
{
    bool stat;
    if (opened)
        CloseDB ();
    if (!OpenDB (file_name, true)) {
        return false;
    }
    //node.Print();
    stat = nodeFwd.DumpToFile(fdFwd, 0) && nodeRev.DumpToFile(fdRev, 0);
    return stat && CloseDB ();
}

void CTrieIndexDFADouble::PrintAddStat(SearchType type, int testNum, int kerr, ostream & ostr)
{
    if (type == SearchTypeMem) {
         double tm = GET_COUNTER_MS(stime) / testNum;
         double tmrev = GET_COUNTER_MS(lproc) / testNum;
         ostr << "\nTotal stat: k = "  << kerr << " time: " << tm << " List reverse: " << tmrev << endl;
         ostr << "\nFwd stat: " ;
         nodeFwd.PrintAddStat(testNum, kerr, ostr);
         ostr << "\nRev stat: " ;
         nodeRev.PrintAddStat(testNum, kerr, ostr);
    }
}

#ifdef _DEBUG
void CTrieIndexDFADouble::Print ()
{
    node.Print ();
}
#endif
