#include <fstream>
#include <iostream>
#include <string>
#include <map>
#include <vector>
#include <algorithm>
#include <set>
#include "../Interface/hashport.h"

#include <string.h>
#include <stdlib.h>
#include <errno.h>
#include <time.h>
#include <assert.h>
#include "../Lang/langcase.h"


//#define SIMPLE_NGRAM_MODEL
inline unsigned int gramPos(unsigned int pos, unsigned int len) {
#ifdef SIMPLE_NGRAM_MODEL
    if (!pos) {
        return 0;
    } else if (pos < len - 1) {
        return 1;
    } 
    return 2;
#else
    return pos;
#endif
};

template <class T> class ProbGenerator {
    typedef std::map<T, unsigned int>  EventMapType;
    typedef std::vector<T>             EventListType;

    EventMapType                       m_eventMap;
    EventListType                      m_eventList;
    std::vector<unsigned int>          m_eventCount;
    std::vector<unsigned int>          m_eventSum;
    unsigned int                       m_totalCount;
  public:
    ProbGenerator();
    void RegisterEvent(const T &);
    const T * GetRandomEvent ();
};

template <class T> ProbGenerator <T>::ProbGenerator():m_totalCount(0)
{
#if 1
    srand(time (NULL));
#else
    srand(0);
#endif
}

template <class T> 
void ProbGenerator<T>::RegisterEvent (const T & event)
{
    typename EventMapType::iterator i = m_eventMap.find (event);

    if (i == m_eventMap.end ()) {
        m_eventList.push_back (event);
        m_eventCount.push_back (1);
    } else {
        m_eventCount[i->second]++;
    }
    m_totalCount++;
}

template <class T> const T * ProbGenerator <T>::GetRandomEvent ()
{
    if (m_totalCount == 0)
        return NULL;

    unsigned int sum = 0;
    if (!m_eventSum.size()) {
        m_eventSum.resize(m_eventCount.size());
        for (unsigned int i = 0; i < m_eventCount.size (); i++) {
            sum += m_eventCount[i];
            m_eventSum[i] = sum; 
        }
        assert(sum == m_totalCount);
    }

    unsigned int                    val = (RAND_MAX * rand () + rand ()) % m_totalCount;
    typename std::vector<unsigned int>::const_iterator it = std::lower_bound(m_eventSum.begin(), m_eventSum.end(), val);

    assert (it != m_eventSum.end());

    return &m_eventList[it - m_eventSum.begin()];

}

void Perror (const char *str = NULL)
{
    std::cerr << ((str != NULL) ? std::string (str) +
             ":" : "") << strerror (errno) << std::endl;
}

void Usage (const char *prgName)
{
    std::cerr << "Usage:" << std::endl;
    std::cerr << prgName << " <template file> <prev size - model parameter> <count>" << std::endl;
}

typedef ProbGenerator<unsigned char> PGUC;

typedef std::hash_map<std::string, PGUC>  Stat;

unsigned int prevSize;

const unsigned int WLEN = 64;

int main (int argc, char *argv[])
{
    if (argc != 4) {
        Usage (argv[0]);
        return -1;
    }

    std::ifstream f (argv[1]);
    if (!f) {
        Perror ("Cannot open file");
        return -1;
    }

    prevSize = atol(argv[2]);

    if (!prevSize || prevSize > 10) {
         std::cerr << "Incorrect prevSize '" << prevSize << "' allowed values 1..10\n";
         return -1;
    }

    ProbGenerator<unsigned int >      wordSizes;
    static Stat                       wordStat[WLEN][WLEN];
    static bool                       hasWordStat[WLEN][WLEN];

    ProbGenerator<unsigned char>      charStat;
    std::string                       s;

    while (f >> s) {
        unsigned int len = s.size ();
        wordSizes.RegisterEvent(len);

        std::string prev = "";
        
        for (unsigned int i = 0; i < len; i++) {
            unsigned char c = toupper_eng_rus(s[i]);

            charStat.RegisterEvent(c);
            wordStat[len][gramPos(i, len)][prev].RegisterEvent(c);
            hasWordStat[len][gramPos(i, len)] = 1;

            
            if (i + 1 > prevSize) {
                prev = prev.substr (1);
            }
            prev += c;
        }

    }

    unsigned int genCnt = 0, modelCnt = 0, repCnt = 0;

    std::hash_set<std::string>                 repInd;

    for (unsigned int i = 0; i < atol (argv[3]); i++) {
        const unsigned int *len_p = wordSizes.GetRandomEvent ();
#if 0
    std::cerr << i << std::endl;
#endif

        if (len_p == NULL) {
		std::cerr << "Empty stat information (word sizes)\n";
        }

        std::string prev = "";
        unsigned int len = *len_p;
        Stat::iterator i2;

        std::string s = "";
        for (unsigned int j = 0; j < len; j++) {
            bool ok = hasWordStat[len][gramPos(j, len)];
            Stat *    pSt = NULL;
            if (ok) {
                pSt = &wordStat[len][gramPos(j, len)];
                i2 = pSt->find(prev);
                if (i2 == pSt->end()) {
                    ok = false;
                }
            } else {
                ok = false;
            }

            const unsigned char *c_p = NULL;
            if (ok) {
                c_p = i2->second.GetRandomEvent ();
                modelCnt++;
            } else {
                genCnt++;
		std::cerr << "using general probability not model: prev " <<
                    prev << " i " << j << std::endl;
                c_p = charStat.GetRandomEvent ();
            }

            
            if (j + 1 > prevSize)
                prev = prev.substr (1);
            prev += *c_p;
            
            s += *c_p;
        }
        if (repInd.find (s) == repInd.end ()) {
            repInd.insert (s);
	    std::cout << s << std::endl;
        } else {
            i--, repCnt++;
        }
    }
    std::cerr << "Model cnt = " << modelCnt << ", general cnt = " << genCnt <<
        " repetition cnt " << repCnt << std::endl;

    return 0;
}
